Require Import Coqlib.
Require Import Maps.
Require Import Values.
Require Import Integers.
Require Import ZArith.
Require Import VTDefs.
Require Import LemmaMaths.
Require Import Psatz.
Require Import LemmaVTMap.
Require Import LemmaInterference.
Require Import LemmaOtherWorkload.
Require Import LemmaWorkloadEnlargeSelf.


Open Scope Z_scope.


Section EXTENDED_WHOLE.

  (* whole periods that fall within [ts, te] *)
  Function workload_extended_whole T C ts te :=
    if (zlt (te - ts) T) then 0 else
      let a := if (zeq (ts mod T) 0) then ts else
                 ts / T * T + T in
      let e := te / T * T in
      (e - a) / T * C.

  Fixpoint workload_extended_whole_total Nnat conf ts te :=
    match Nnat with
    | O => 0
    | S n =>
      match (ZMap.get (Z.of_nat n) conf) with
      | mkPrioConfigValid T C =>
        workload_extended_whole_total n conf ts te +
        workload_extended_whole T C ts te
      end
    end.

  Lemma workload_extended_whole_nonneg:
    forall T C t1 t2,
      0 <= t1 ->
      0 < T ->
      0 < C ->
      0 <= workload_extended_whole T C t1 t2.
  Proof.
    intros * Ht HT HC.
    unfold workload_extended_whole.
    destruct (zlt (t2 - t1) T); try omega.
    destruct (zeq (t1 mod T) 0) as [Hmod | Hmod].
    {
      apply div_mul_nonneg; try omega.
      assert (H1: t1 / T * T <= t2 / T * T).
      {
        apply div_mul_le; omega.
      }
      rewrite mod_0_divide_mul in H1; omega.
    }
    {
      apply div_mul_nonneg; try omega.      
      assert (H1: (t1+T) / T * T <= t2 / T * T).
      {
        apply div_mul_le; omega.
      }
      rewrite mod_not_0_divide_mul in H1; omega.
    }
  Qed.
  
  Lemma workload_extended_whole_empty_slot:
    forall T C t,
      0 < T ->
      workload_extended_whole T C t t = 0.
  Proof.
    intros.
    unfold workload_extended_whole.
    rewrite zlt_true; try omega.
  Qed.

  Lemma workload_extended_whole_right_align:
    forall T C v ts,
      0 <= v ->
      0 <= ts ->
      0 < T ->
      ts mod T <> 0 ->
      workload_extended_whole T C v ts = workload_extended_whole T C v (ts /T * T).
  Proof.
    intros * Hv Ht HT Hmod. 
    unfold workload_extended_whole.
    assert (H1: ts - ts / T * T >= 1).
    { apply z_gt_mod; auto; try omega. }
    assert (H2: (ts / T * T) mod T = 0).
    { apply Z_mod_mult. }
    
    destruct (zlt (ts - v) T).
    {
      rewrite zlt_true; omega.
    }
    {
      assert (H10: ts < ts / T * T + T).
      { apply x_lt_floor_plus; omega. }
        
      destruct (zlt (ts / T * T - v) T).
      {
        assert (v mod T <> 0).
        { apply within_boundary_mod_not_0 with (c:=ts/T*T); try omega. }        
        rewrite zeq_false; auto.

        assert (H11: v / T * T + T = ts / T * T).
        { apply div_mul_round_up; try omega. }
        rewrite H11.
        replace (ts / T * T - ts / T * T) with 0 by omega.
        auto.
      }
      {
        assert (H12: ts / T * T / T * T = ts / T * T).
        { rewrite mod_0_divide_mul; auto; try omega. }
        rewrite H12.
        auto.
      }
    }
  Qed.

  Lemma workload_extended_whole_right_step:
    forall T C v ts,
      0 <= v <= ts ->
      1 < T ->
      ts mod T = 0 ->
      workload_extended_whole T C v (ts + T) = workload_extended_whole T C v ts + C.
  Proof.
    intros * Hv HT Hmod.
    unfold workload_extended_whole.
    rewrite zlt_false; try omega.
    assert (H1: (ts + T) mod T = 0).
    { apply mod_0_preserve; omega. }
    rewrite mod_0_divide_mul; try omega.    

    destruct (zlt (ts - v) T).
    {
      destruct (zeq v ts) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite Hmod.
        rewrite zeq_true.
        replace (ts + T - ts) with T by omega.
        rewrite Z_div_same_full; try omega.
      }
      {
        assert (Hvmod: v mod T <> 0).
        { apply within_boundary_mod_not_0 with (c:=ts); try omega. }
        rewrite zeq_false; auto.
        
        assert (H2: v / T * T + T = ts).
        {
          apply div_mul_round_up; omega.
        }
        rewrite H2.
        replace (ts + T - ts) with T by omega.
        rewrite Z_div_same_full; try omega.
      }
    }
    {
      assert (H2: ts / T * T = ts).
      { rewrite mod_0_divide_mul; try omega. }
      rewrite H2.

      replace (ts + T - (if zeq (v mod T) 0 then v else v / T * T + T)) with
          (T + (ts - (if zeq (v mod T) 0 then v else v / T * T + T))) by omega.
      rewrite Nadd_div_exact; try omega.
      - rewrite Z_div_same_full; try omega.
        lia.
      - destruct (zeq (v mod T) 0); try omega.        
        assert (v /T * T + T <= ts - T).
        {
          apply mod_0_bound_increase; try omega.
          destruct (zeq (v + T) ts); try omega.
          rewrite <- e in *.
          assert ((v + T) mod T <> 0).
          { apply mod_not_0_preserve; omega. }
          omega.
        }
        omega.
      - apply Z_mod_same_full.
    }
  Qed.
  
End EXTENDED_WHOLE.


Section SCHED_TYPE.

  Inductive esched_type: Type :=
  | ESched_whole                    (* whole periods that must be finished by deadline *)
  | ESched_overlap                  (* tasks that contributes to workload_remain in the next period *)
  | ESched_other.

  Lemma esched_type_eq_dec:
    forall x y : esched_type, {x = y} + {x <> y}.
  Proof.
    intros.
    destruct x; destruct y; auto; (right; congruence).
  Qed.

  Lemma esched_type_eq_true:
    forall (A: Type) (ty: esched_type) (a b: A),
      (if esched_type_eq_dec ty ty then a else b) = a.
  Proof.
    intros.
    case (esched_type_eq_dec ty ty); intros.
    auto.
    elim n; auto.
  Qed.

  Lemma esched_type_eq_false:
    forall (A: Type) (x y: esched_type) (a b: A),
      x <> y ->
      (if esched_type_eq_dec x y then a else b) = b.
  Proof.
    intros.
    case (esched_type_eq_dec x y); intros.
    elim H; auto.
    auto.
  Qed.

  Function get_esched_type (fromid T toid t v ts te: Z) :esched_type :=
    if (zle v (t / T * T)) then
      if (zlt (t / T * T) ts) then
        (if (zle (t / T * T + T) ts) then ESched_whole else
           (if (zeq fromid toid) then ESched_other else
              (if (priEntry_lt (PriEntry (t/ T*T +T) fromid) (PriEntry te toid)) then
                 ESched_overlap else ESched_other)))
      else ESched_other
    else ESched_other.

  Fixpoint count_etype_occurrence_aux N idle tcounts nrange conf toid v ts te ty :=
    match nrange with
    | O => 0
    | S n =>
      let i := highest_not_finished_map idle (tcounts + Z.of_nat n)
                                        (TAC_pool_at_t (tcounts + Z.of_nat n) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat n) conf) in
      if (zeq i idle) then count_etype_occurrence_aux N idle tcounts n conf toid v ts te ty else
        match (ZMap.get i conf) with
        | mkPrioConfigValid T C =>
          let cur_ty := get_esched_type i T toid (tcounts + Z.of_nat n) v ts te in
          if (esched_type_eq_dec cur_ty ty) then
            count_etype_occurrence_aux N idle tcounts n conf toid v ts te ty + 1 else
            count_etype_occurrence_aux N idle tcounts n conf toid v ts te ty
        end
    end.

  Function count_etype_occurrence N idle tcounts tcounte conf toid v ts te ty :=
    count_etype_occurrence_aux N idle tcounts (Z.to_nat (tcounte - tcounts)) conf toid v ts te ty.
  
  Fixpoint count_single_task_etype_aux N idle fromid tcount nrange conf toid v ts te ty: Z :=
    match nrange with
    | O => 0
    | S n =>
      let i := highest_not_finished_map idle (tcount + Z.of_nat n)
                                        (TAC_pool_at_t (tcount + Z.of_nat n) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (tcount + Z.of_nat n) conf) in
      if (zeq i fromid) then 
        match (ZMap.get i conf) with
        | mkPrioConfigValid T C =>
          let cur_ty := get_esched_type i T toid (tcount + Z.of_nat n) v ts te in
          if (esched_type_eq_dec cur_ty ty) then
            count_single_task_etype_aux N idle fromid tcount n conf toid v ts te ty + 1 else
            count_single_task_etype_aux N idle fromid tcount n conf toid v ts te ty
        end
          else count_single_task_etype_aux N idle fromid tcount n conf toid v ts te ty
    end.

  Function count_single_task_etype_occurrence N idle fromid tcounts tcounte conf toid v ts te ty: Z :=
    count_single_task_etype_aux N idle fromid tcounts (Z.to_nat (tcounte - tcounts)) conf toid v ts te ty.

  Fixpoint count_etype_occurrence2_aux (Nid: nat) (N idle tcounts tcounte: Z) conf (toid v ts te: Z) ty: Z :=
    match Nid with
    | O => 0
    | S n => count_etype_occurrence2_aux n N idle tcounts tcounte conf toid v ts te ty +
            count_single_task_etype_occurrence N N (Z.of_nat n) tcounts tcounte conf toid v ts te ty
    end.  

  Lemma count_single_task_etype_occurance_0_when_empty_duration:
    forall N1 N2 N3 t conf toid v ts te ty,
      count_single_task_etype_occurrence N1 N2 N3 t t conf toid v ts te ty = 0.
  Proof.
    unfold count_single_task_etype_occurrence.
    intros.
    replace (t - t) with 0 by omega.
    simpl.
    auto.
  Qed.
                                                                                   
  Lemma count_etype_occurance2_aux_0_when_empty_duration:
    forall Nid N1 N2 t conf toid v ts te ty,
      count_etype_occurrence2_aux Nid N1 N2 t t conf toid v ts te ty = 0.
  Proof.
    intros.
    induction Nid.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHNid.
      rewrite count_single_task_etype_occurance_0_when_empty_duration; auto.
    }
  Qed.
  
  Lemma count_single_task_etype_occurrence_stagnates_when_not_scheduled:
    forall N fromid tcounts tcounte conf toid v ts te ty,
      tcounts <= tcounte ->
      highest_not_finished_map N tcounte (TAC_pool_at_t tcounte (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) tcounte conf) <> fromid ->
      count_single_task_etype_occurrence N N fromid tcounts (tcounte + 1) conf toid v ts te ty =      
      count_single_task_etype_occurrence N N fromid tcounts tcounte conf toid v ts te ty.
  Proof.
    intros * Hte Hid Hidle.
    unfold count_single_task_etype_occurrence.
    remember (Z.to_nat (tcounte - tcounts)) as range.
    assert (H1: tcounte = tcounts + Z.of_nat range).
    {
      rewrite Heqrange.
      rewrite Z2Nat.id; omega.
    }
    assert (H2: (Z.to_nat (tcounte + 1 - tcounts)) = S range).
    {
      rewrite Heqrange.
      rewrite <- Z2Nat.inj_succ; try omega.
      replace (tcounte + 1 - tcounts) with (tcounte - tcounts + 1) by omega.
      unfold Z.succ.
      auto.
    }
    rewrite H2.
    simpl.
    rewrite <- H1.
    rewrite zeq_false; auto.
  Qed.
  
  Lemma count_single_task_etype_occurrence_incr_when_scheduled_and_match:
    forall N tcounts tcounte conf toid v ts te ty i T C,
      tcounts <= tcounte ->
      i = highest_not_finished_map N tcounte (TAC_pool_at_t tcounte (Z.to_nat N) conf) conf
                                   (real_priqueue (Z.to_nat N) tcounte conf) ->
      0 <= i < N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      get_esched_type i T toid tcounte v ts te = ty ->      
      count_single_task_etype_occurrence N N i tcounts (tcounte + 1) conf toid v ts te ty =      
      count_single_task_etype_occurrence N N i tcounts tcounte conf toid v ts te ty + 1.
  Proof.
    intros * Hts Hhigh Hi Hconf Htype.
    unfold count_single_task_etype_occurrence.
    remember (Z.to_nat (tcounte - tcounts)) as range.
    assert (H1: tcounte = tcounts + Z.of_nat range).
    {
      rewrite Heqrange.
      rewrite Z2Nat.id; omega.
    }
    assert (H2: (Z.to_nat (tcounte + 1 - tcounts)) = S range).
    {
      rewrite Heqrange.
      rewrite <- Z2Nat.inj_succ; try omega.
      replace (tcounte + 1 - tcounts) with (tcounte - tcounts + 1) by omega.
      unfold Z.succ.
      auto.
    }
    rewrite H2.
    simpl.
    rewrite <- H1.
    rewrite <- Hhigh.
    rewrite zeq_true.
    rewrite Hconf.
    rewrite Htype.
    rewrite esched_type_eq_true; auto.
  Qed.

  Lemma count_single_task_etype_occurrence_stagnates_when_scheduled_but_mismatch:
    forall N tcounts tcounte conf toid v ts te ty i T C,
      tcounts <= tcounte ->
      i = highest_not_finished_map N tcounte (TAC_pool_at_t tcounte (Z.to_nat N) conf) conf
                                   (real_priqueue (Z.to_nat N) tcounte conf) ->
      0 <= i < N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      get_esched_type i T toid tcounte v ts te <> ty ->      
      count_single_task_etype_occurrence N N i tcounts (tcounte + 1) conf toid v ts te ty =      
      count_single_task_etype_occurrence N N i tcounts tcounte conf toid v ts te ty.
  Proof.
    intros * Hts Hhigh Hi Hconf Htype.
    unfold count_single_task_etype_occurrence.
    remember (Z.to_nat (tcounte - tcounts)) as range.
    assert (H1: tcounte = tcounts + Z.of_nat range).
    {
      rewrite Heqrange.
      rewrite Z2Nat.id; omega.
    }
    assert (H2: (Z.to_nat (tcounte + 1 - tcounts)) = S range).
    {
      rewrite Heqrange.
      rewrite <- Z2Nat.inj_succ; try omega.
      replace (tcounte + 1 - tcounts) with (tcounte - tcounts + 1) by omega.
      unfold Z.succ.
      auto.
    }
    rewrite H2.
    simpl.
    rewrite <- H1.
    rewrite <- Hhigh.
    rewrite zeq_true.
    rewrite Hconf.
    rewrite esched_type_eq_false; auto.
  Qed.

  Definition ecount_single_slot N idle fromid tcounts conf toid v ts te ty :Z :=
      (if zeq
         (highest_not_finished_map idle tcounts (TAC_pool_at_t tcounts (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) tcounts conf)) fromid
    then
     match
       ZMap.get
         (highest_not_finished_map idle tcounts (TAC_pool_at_t tcounts (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) tcounts conf)) conf
     with
     | mkPrioConfigValid T _ =>
         if esched_type_eq_dec
              (get_esched_type
                 (highest_not_finished_map idle tcounts 
                    (TAC_pool_at_t tcounts (Z.to_nat N) conf) conf 
                    (real_priqueue (Z.to_nat N) tcounts conf)) T toid tcounts v ts te) ty
         then 1
         else 0
     end
       else 0).
  
  Lemma count_single_task_etype_occurrence_aux_left_step:
    forall N idle fromid tcounts range conf toid v ts te ty,
      0 < Z.of_nat range ->
      count_single_task_etype_aux N idle fromid tcounts range conf toid v ts te ty =
      count_single_task_etype_aux N idle fromid (tcounts + 1) (pred range) conf toid v ts te ty +
      ecount_single_slot N idle fromid tcounts conf toid v ts te ty.
  Proof.
    intros * Hrange.
    revert tcounts.
    induction range.
    {
      simpl in Hrange.      
      intros.
      omega.
    }
    {
      intros.
      destruct (zeq (Z.of_nat range) 0) as [Heq | Hneq].
      {
        assert (H2: Z.to_nat (Z.of_nat range) = Z.to_nat 0).
        { rewrite Heq. auto. }
        simpl in H2.
        rewrite Nat2Z.id in H2.
        
        rewrite H2 in *.
        simpl.
        replace (tcounts + 0) with tcounts by omega.
        auto.
      }
      {
        simpl.
        destruct (zeq
         (highest_not_finished_map idle (tcounts + Z.of_nat range)
            (TAC_pool_at_t (tcounts + Z.of_nat range) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat range) conf)) fromid) as [Hsched|Hother].
        {
          destruct (ZMap.get
       (highest_not_finished_map idle (tcounts + Z.of_nat range)
          (TAC_pool_at_t (tcounts + Z.of_nat range) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat range) conf)) conf) as [T C] eqn:Hconf.
          destruct (esched_type_eq_dec
         (get_esched_type
            (highest_not_finished_map idle (tcounts + Z.of_nat range)
               (TAC_pool_at_t (tcounts + Z.of_nat range) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat range) conf)) T toid
            (tcounts + Z.of_nat range) v ts te) ty) as [Hmatch | Hmismatch].
          {
            rewrite IHrange; try xomega.
            assert (forall a b c, a + 1 = b -> a + c + 1 = b + c). intros. omega.
            apply H.
            remember (pred range) as n.
            assert (Hr: range = S n).
            {
              rewrite Heqn.
              rewrite NPeano.Nat.succ_pred; auto.
              omega.
            }
            rewrite Hr.
            simpl.

            replace (tcounts + 1 + Z.of_nat n) with (tcounts + Z.of_nat range).
            rewrite Hsched.
            rewrite zeq_true.
            rewrite Hsched in *.
            rewrite Hconf.            
            rewrite Hmatch.
            rewrite esched_type_eq_true.
            auto.
            - xomega.
          }
          {
            rewrite IHrange; try xomega.
            assert (forall a b c, a = b -> a + c = b + c). intros. omega.
            apply H.
            remember (pred range) as n.
            assert (Hr: range = S n).
            {
              rewrite Heqn.
              rewrite NPeano.Nat.succ_pred; auto.
              omega.
            }
            rewrite Hr.
            simpl.

            replace (tcounts + 1 + Z.of_nat n) with (tcounts + Z.of_nat range).
            rewrite Hsched.
            rewrite zeq_true.
            rewrite Hsched in *.
            rewrite Hconf.            
            rewrite esched_type_eq_false; auto.
            - xomega.
          }
        }
        {
          rewrite IHrange; try xomega.
          assert (forall a b c, a = b -> a + c = b + c). intros. omega.
          apply H.
          remember (pred range) as n.
          assert (Hr: range = S n).
          {
            rewrite Heqn.
            rewrite NPeano.Nat.succ_pred; auto.
            omega.
          }
          rewrite Hr.
          simpl.

          replace (tcounts + 1 + Z.of_nat n) with (tcounts + Z.of_nat range).          
          rewrite zeq_false; auto.
          - xomega.
        }
      }
    }
  Qed.
  
  Lemma count_single_task_etype_occurrence_composition:
    forall N idle fromid t1 t2 t3 conf toid v ts te ty,
      t1 <= t2 <= t3 ->
      count_single_task_etype_occurrence N idle fromid t1 t2 conf toid v ts te ty +
      count_single_task_etype_occurrence N idle fromid t2 t3 conf toid v ts te ty =
      count_single_task_etype_occurrence N idle fromid t1 t3 conf toid v ts te ty.      
  Proof.
    intros * Ht.
    remember (Z.to_nat (t2 -t1)) as range.
    assert (H1: t2 = t1 + Z.of_nat range).
    {
      rewrite Heqrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite H1 in *. clear H1 Heqrange.

    induction range.
    {
      simpl.
      replace (t1 + 0) with t1 by omega.
      rewrite count_single_task_etype_occurance_0_when_empty_duration.
      omega.
    }
    {
      rewrite <- IHrange; try xomega. clear IHrange.
      unfold count_single_task_etype_occurrence.      
      
      replace (t1 + Z.of_nat (S range) - t1) with (Z.of_nat (S range)) by omega.
      replace (t1 + Z.of_nat range - t1) with (Z.of_nat range) by omega.
      repeat rewrite Nat2Z.id.        

      assert (H1: t1 + Z.of_nat (S range) = t1 + Z.of_nat range + 1). xomega.
      rewrite H1. clear H1.

      assert (H2: Z.to_nat (t3 - (t1 + Z.of_nat range + 1)) =
                  pred (Z.to_nat (t3 - (t1 + Z.of_nat range)))).
      {
        rewrite <- Z2Nat.inj_pred.
        rewrite <- Z.sub_1_r.
        replace (t3 - (t1 + Z.of_nat range + 1)) with (t3 - (t1 + Z.of_nat range) - 1) by omega.
        auto.
      }
      rewrite H2.
      
      assert (H3: count_single_task_etype_aux N idle fromid (t1 + Z.of_nat range) 
                                             (Z.to_nat (t3 - (t1 + Z.of_nat range))) conf toid v ts te ty =
             count_single_task_etype_aux N idle fromid (t1 + Z.of_nat range + 1) 
                                        (pred (Z.to_nat (t3 - (t1 + Z.of_nat range)))) conf toid v ts te ty +
             ecount_single_slot N idle fromid (t1 + Z.of_nat range) conf toid v ts te ty).
      {
        apply count_single_task_etype_occurrence_aux_left_step; auto.
        - rewrite Z2Nat.id; xomega.
      }
      rewrite H3. clear H3.

      assert (H: forall a b c d, a = b + c -> a + d = b + (d + c)). intros. omega.
      apply H.
      simpl.
      unfold ecount_single_slot.

      destruct (zeq
         (highest_not_finished_map idle (t1 + Z.of_nat range)
            (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) fromid) as [Heq|Hneq]; try omega.
      
      destruct (ZMap.get
       (highest_not_finished_map idle (t1 + Z.of_nat range)
          (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) conf).
      destruct (esched_type_eq_dec
         (get_esched_type
            (highest_not_finished_map idle (t1 + Z.of_nat range)
               (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (t1 + Z.of_nat range) conf)) z toid
            (t1 + Z.of_nat range) v ts te) ty); omega.
    }
  Qed.

  Lemma count_etype_occurance2_aux_stagnates_when_slot_idle:
    forall Nid N tcounts tcounte conf toid v ts te ty,
      tcounts <= tcounte ->
      Z.of_nat Nid <= N ->
      highest_not_finished_map N tcounte 
            (TAC_pool_at_t tcounte (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) tcounte conf) = N ->
      count_etype_occurrence2_aux Nid N N tcounts (tcounte + 1) conf toid v ts te ty =      
      count_etype_occurrence2_aux Nid N N tcounts tcounte conf toid v ts te ty.
  Proof.
    intros * Hte HN Hidle.
    induction Nid.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: Z.of_nat Nid < Z.of_nat (S Nid)). xomega.
      assert (HN': Z.of_nat Nid <= N). xomega.
      rewrite <- (IHNid HN').

      rewrite count_single_task_etype_occurrence_stagnates_when_not_scheduled; auto; try omega.
    }
  Qed.

  Lemma count_etype_occurance2_aux_stagnates_when_not_scheduled:
    forall Nid N tcounts tcounte conf toid v ts te ty,
      tcounts <= tcounte ->
      Z.of_nat Nid <= highest_not_finished_map N tcounte 
            (TAC_pool_at_t tcounte (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) tcounte conf) ->
      count_etype_occurrence2_aux Nid N N tcounts (tcounte + 1) conf toid v ts te ty =      
      count_etype_occurrence2_aux Nid N N tcounts tcounte conf toid v ts te ty.
  Proof.
    intros * Hte HN Hidle.
    induction Nid.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: Z.of_nat Nid < Z.of_nat (S Nid)). xomega.
      rewrite <- IHNid; try omega.

      rewrite count_single_task_etype_occurrence_stagnates_when_not_scheduled; auto; try omega.
    }
  Qed.

  Lemma count_etype_occurance2_aux_increase_when_scheduled_task_match:
    forall Nid N tcounts tcounte conf toid v ts te ty i T C,
      tcounts <= tcounte ->
      i < Z.of_nat Nid <= N ->     
      i = highest_not_finished_map N tcounte 
            (TAC_pool_at_t tcounte (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) tcounte conf) ->
      0 <= i <= N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      get_esched_type i T toid tcounte v ts te = ty ->
      count_etype_occurrence2_aux Nid N N tcounts (tcounte + 1) conf toid v ts te ty =      
      count_etype_occurrence2_aux Nid N N tcounts tcounte conf toid v ts te ty + 1.
  Proof.
    intros * Hte HN Hsched Hi Hconfi Hmatch.
    induction Nid.
    {
      simpl in HN.
      omega.
    }
    {
      simpl.
      destruct (zeq (Z.of_nat Nid) i) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite count_etype_occurance2_aux_stagnates_when_not_scheduled; auto; try omega.
        rewrite count_single_task_etype_occurrence_incr_when_scheduled_and_match with (T:=T) (C:=C); auto; try omega.
      }
      {
        rewrite count_single_task_etype_occurrence_stagnates_when_not_scheduled; auto; try omega.
        rewrite IHNid; auto; try xomega.
      }
    }
  Qed.
  
  Lemma count_etype_occurance2_aux_increase_when_scheduled_task_mismatch:
    forall Nid N tcounts tcounte conf toid v ts te ty i T C,
      tcounts <= tcounte ->
      i < Z.of_nat Nid <= N ->     
      i = highest_not_finished_map N tcounte 
            (TAC_pool_at_t tcounte (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) tcounte conf) ->
      0 <= i <= N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      get_esched_type i T toid tcounte v ts te <> ty ->
      count_etype_occurrence2_aux Nid N N tcounts (tcounte + 1) conf toid v ts te ty =      
      count_etype_occurrence2_aux Nid N N tcounts tcounte conf toid v ts te ty.
  Proof.
    intros * Hte HN Hsched Hi Hconfi Hmatch.
    induction Nid.
    {
      simpl in HN.
      omega.
    }
    {
      simpl.
      destruct (zeq (Z.of_nat Nid) i) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite count_etype_occurance2_aux_stagnates_when_not_scheduled; auto; try omega.
        rewrite count_single_task_etype_occurrence_stagnates_when_scheduled_but_mismatch with (T:=T) (C:=C); auto; try omega.
      }
      {
        rewrite count_single_task_etype_occurrence_stagnates_when_not_scheduled; auto; try omega.
        rewrite IHNid; auto; try xomega.
      }
    }
  Qed.

  Lemma count_etype_occurrence2_aux_eq:
    forall N tcounts nrange conf toid v ts te ty,
      0 < N ->
      count_etype_occurrence_aux N N tcounts nrange conf toid v ts te ty =
      count_etype_occurrence2_aux (Z.to_nat N) N N tcounts (tcounts + Z.of_nat nrange) conf toid v ts te ty.
  Proof.
    intros * HN.
    induction nrange.
    {
      simpl.
      replace (tcounts + 0) with tcounts by omega.
      rewrite count_etype_occurance2_aux_0_when_empty_duration.
      auto.
    }
    {
      assert (H1: tcounts + Z.of_nat (S nrange) = tcounts + Z.of_nat nrange + 1). xomega.
      assert (H2: Z.of_nat (Z.to_nat N) = N). { rewrite Z2Nat.id; omega. }
      rewrite H1. clear H1.
      simpl.
      destruct (zeq
         (highest_not_finished_map N (tcounts + Z.of_nat nrange)
            (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf)) N) as [Hidle | Hsched].
      {
        rewrite IHnrange.
        rewrite count_etype_occurance2_aux_stagnates_when_slot_idle; auto; try omega.
      }
      {
        remember (highest_not_finished_map N (tcounts + Z.of_nat nrange)
             (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
             (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf)) as i.        
        assert (H3: 0 <= i <= N).
        {
          rewrite Heqi.
          apply highest_not_finished_map_res_bound; auto.
          - remember (Z.to_nat N) as nN.
            replace N with (Z.of_nat nN).
            apply real_priqueue_sound.
        }
        destruct (ZMap.get i conf) as [T C] eqn:Hconf.        
        destruct (esched_type_eq_dec (get_esched_type i T toid (tcounts + Z.of_nat nrange) v ts te) ty) as [Hmatch|Hmismatch].
        {
          rewrite IHnrange.
          rewrite count_etype_occurance2_aux_increase_when_scheduled_task_match with (i:=i) (T:=T) (C:=C); auto; try xomega.
        }
        {
          rewrite IHnrange.
          rewrite count_etype_occurance2_aux_increase_when_scheduled_task_mismatch with (i:=i) (T:=T) (C:=C); auto; try xomega.
        }
      }
    }
  Qed.
  
  Lemma count_etype_occurrence2_eq:
    forall N tcounts tcounte conf toid v ts te ty,
      0 < N ->
      tcounts <= tcounte ->
      count_etype_occurrence N N tcounts tcounte conf toid v ts te ty =
      count_etype_occurrence2_aux (Z.to_nat N) N N tcounts tcounte conf toid v ts te ty.      
  Proof.
    intros * HN Ht.
    unfold count_etype_occurrence.
    remember (Z.to_nat (tcounte - tcounts)) as range.
    replace tcounte with (tcounts + Z.of_nat range).
    - rewrite count_etype_occurrence2_aux_eq; auto.
    - rewrite Heqrange.
      rewrite Z2Nat.id; omega.
  Qed.

  Lemma sched_etype_sum_aux_inc_when_schedule_not_idle:
    forall N conf id v ts te tcounts nrange,
      0 <= highest_not_finished_map N (tcounts + Z.of_nat nrange)
                                  (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf) < N ->
      count_etype_occurrence_aux N N tcounts (S nrange) conf id v ts te ESched_whole +
      count_etype_occurrence_aux N N tcounts (S nrange) conf id v ts te ESched_overlap +
      count_etype_occurrence_aux N N tcounts (S nrange) conf id v ts te ESched_other =
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_whole +
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_overlap +
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_other + 1.
  Proof.
    intros * Hreal.
    simpl.
    repeat rewrite zeq_false; try omega.
    destruct (ZMap.get
       (highest_not_finished_map N (tcounts + Z.of_nat nrange)
          (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf)) conf) as [T0 C0] eqn:Hconf0.
    destruct ((get_esched_type
            (highest_not_finished_map N (tcounts + Z.of_nat nrange)
               (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf)) T0 id
            (tcounts + Z.of_nat nrange) v ts te)) eqn:Hty.
    {
      (* ESched_whole *)
      rewrite esched_type_eq_true.
      repeat rewrite esched_type_eq_false; try congruence.
      omega.
    }
    {
      (* ESched_overlap *)
      rewrite esched_type_eq_true.
      repeat rewrite esched_type_eq_false; try congruence.
      omega.
    }
    {
      rewrite esched_type_eq_true.
      repeat rewrite esched_type_eq_false; try congruence.
      omega.
    }
  Qed.

  Lemma sched_etype_sum_aux_stagnates_when_schedule_idle:
    forall N conf id v ts te tcounts nrange,
      highest_not_finished_map N (tcounts + Z.of_nat nrange)
                                  (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf) = N ->
      count_etype_occurrence_aux N N tcounts (S nrange) conf id v ts te ESched_whole +
      count_etype_occurrence_aux N N tcounts (S nrange) conf id v ts te ESched_overlap +
      count_etype_occurrence_aux N N tcounts (S nrange) conf id v ts te ESched_other =
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_whole +
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_overlap +
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_other.
  Proof.
    intros * Hreal.    
    simpl.
    rewrite Hreal.
    repeat rewrite zeq_true.
    auto.
  Qed.

  Lemma sched_etype_sum_aux_bounded_by_duration:
    forall N conf id v ts te tcounts nrange,
      0 < N ->
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_whole +
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_overlap +
      count_etype_occurrence_aux N N tcounts nrange conf id v ts te ESched_other <= Z.of_nat nrange.
  Proof.
    intros * HN.
    induction nrange.
    {
      simpl.
      omega.
    }
    {
      assert (H1: 0 <= highest_not_finished_map N (tcounts + Z.of_nat nrange)
                                  (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                  (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf) <= N).
      {
        apply highest_not_finished_map_res_bound; auto.
        - remember (Z.to_nat N) as Nnat.
          replace N with (Z.of_nat Nnat).
          apply real_priqueue_sound.
          rewrite HeqNnat.
          rewrite Z2Nat.id; omega.
      }

      destruct (zeq (highest_not_finished_map N (tcounts + Z.of_nat nrange)
                                  (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
                                  (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf)) N).
      {
        rewrite sched_etype_sum_aux_stagnates_when_schedule_idle; auto.
        xomega.
      }
      {
        rewrite sched_etype_sum_aux_inc_when_schedule_not_idle; try xomega.
      }
    }
  Qed.

  Lemma count_etype_occurrence_aux_nonneg:
    forall N conf id v ts te tcounts nrange ty,
      0 <= count_etype_occurrence_aux N N tcounts nrange conf id v ts te ty.
  Proof.
    intros.
    induction nrange.
    {
      simpl.
      omega.
    }
    {
      simpl.
      destruct (zeq
         (highest_not_finished_map N (tcounts + Z.of_nat nrange)
            (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf)) N); auto.
      destruct (ZMap.get
       (highest_not_finished_map N (tcounts + Z.of_nat nrange)
          (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
          (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf)) conf) as [T C].
      destruct (esched_type_eq_dec
         (get_esched_type
            (highest_not_finished_map N (tcounts + Z.of_nat nrange)
               (TAC_pool_at_t (tcounts + Z.of_nat nrange) (Z.to_nat N) conf) conf
               (real_priqueue (Z.to_nat N) (tcounts + Z.of_nat nrange) conf)) T id
            (tcounts + Z.of_nat nrange) v ts te) ty); try omega.
    }
  Qed.
  
End SCHED_TYPE.


Section BURST.

  Fixpoint last_idle_or_long_overlap_aux N idle conf id deadline nrange: option Z :=
    match nrange with
    | O => None
    | S n =>
      let i := highest_not_finished_map idle (Z.of_nat n) (TAC_pool_at_t (Z.of_nat n) (Z.to_nat N) conf) conf
                                        (real_priqueue (Z.to_nat N) (Z.of_nat n) conf) in      
      if (zeq idle i) then Some (Z.of_nat n) else
        match (ZMap.get i conf) with
        | mkPrioConfigValid T C =>          
          if (priEntry_lt (PriEntry deadline id) (PriEntry ((Z.of_nat n) / T * T + T) i))
               then Some (Z.of_nat n) else
            last_idle_or_long_overlap_aux N idle conf id deadline n
        end
    end.

  (* Starting from burst, every slot schedules a task that has a priority higher than
     or equal to (deadline, id) *)
  Function start_of_burst N conf ts id deadline :=
    match (last_idle_or_long_overlap_aux N N conf id deadline (Z.to_nat ts)) with
    | Some v => v + 1
    | None => 0
    end.

  Lemma last_idle_or_long_overlap_aux_result_range:
    forall N idle conf id te nrange v,
      last_idle_or_long_overlap_aux N idle conf id te nrange = Some v ->
      0 <= v < Z.of_nat nrange.
  Proof.
    intros * Hval.
    induction nrange.
    {
      simpl in Hval.
      congruence.
    }
    {
      simpl in Hval.
      destruct (zeq idle
               (highest_not_finished_map idle (Z.of_nat nrange)
                  (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
                  (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf))).
      {
        inversion Hval.
        xomega.
      }
      {
        destruct (ZMap.get
             (highest_not_finished_map idle (Z.of_nat nrange)
                (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
                (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf)) conf) as [T C].
        destruct (if zlt te (Z.of_nat nrange / T * T + T)
             then true
             else
              if zeq te (Z.of_nat nrange / T * T + T)
              then
               if zlt id
                    (highest_not_finished_map idle 
                       (Z.of_nat nrange) (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf)
                       conf (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf))
               then true
               else false
              else false).
        {
          inversion Hval.
          xomega.
        }
        {
          specialize (IHnrange Hval).
          xomega.
        }
      }
    }
  Qed.

  Lemma start_of_burst_result_nonneg:
    forall N conf ts id deadline,
      0 <= start_of_burst N conf ts id deadline.
  Proof.
    intros.
    unfold start_of_burst.
    destruct (last_idle_or_long_overlap_aux N N conf id deadline (Z.to_nat ts)) eqn:Hlast.
    {
      pose proof (last_idle_or_long_overlap_aux_result_range _ _ _ _ _ _ _ Hlast).
      omega.
    }
    {
      omega.
    }
  Qed.
  
  Lemma start_of_burst_result_range:
    forall N conf ts id deadline,
      0 <= ts ->
      0 <= start_of_burst N conf ts id deadline <= ts.
  Proof.
    intros.
    unfold start_of_burst.
    destruct (last_idle_or_long_overlap_aux N N conf id deadline (Z.to_nat ts)) eqn:Hlast.
    {
      pose proof (last_idle_or_long_overlap_aux_result_range _ _ _ _ _ _ _ Hlast).
      rewrite Z2Nat.id in H0; xomega.
    }
    {
      omega.
    }
  Qed.
  
  Lemma last_idle_or_long_overlap_aux_res_idle_or_long:
    forall N idle conf id te nrange v i T C,
      last_idle_or_long_overlap_aux N idle conf id te nrange = Some v ->
      highest_not_finished_map idle v (TAC_pool_at_t v (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) v conf) = i ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      (i = idle \/ (priEntry_lt (PriEntry te id) (PriEntry (v / T * T + T) i) = true)).
  Proof.
    intros N idle conf id te nrange.    
    induction nrange.
    {
      intros.
      simpl in *.
      congruence.
    }
    {
      intros * Hlast Hval Hconfi.
      simpl in Hlast.
      destruct (zeq idle
                (highest_not_finished_map idle (Z.of_nat nrange)
                   (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
                   (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf))) as [Heq|Hneq].
      {
        inversion Hlast.
        rewrite H0 in *.
        left.
        omega.
      }
      {
        destruct (ZMap.get
              (highest_not_finished_map idle (Z.of_nat nrange)
                 (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
                 (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf)) conf) as [Tn Cn] eqn:Hconfn.
        destruct (if zlt te (Z.of_nat nrange / Tn * Tn + Tn)
              then true
              else
               if zeq te (Z.of_nat nrange / Tn * Tn + Tn)
               then
                if zlt id
                     (highest_not_finished_map idle 
                        (Z.of_nat nrange)
                        (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
                        (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf))
                then true
                else false
               else false) eqn:Hpri.
        {
          inversion Hlast.
          rewrite H0 in *.
          rewrite Hval in *.
          rewrite Hconfn in Hconfi.
          inversion Hconfi.
          right.
          rewrite H1 in Hpri.
          auto.
        }
        {
          apply IHnrange with (C:=C); auto.
        }
      }
    }
  Qed.        
  
  Lemma idle_slot_is_found_if_exists:
    forall N idle conf id nrange te t,
      highest_not_finished_map idle t (TAC_pool_at_t t (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) t conf) = idle ->
      0 <= t < Z.of_nat nrange ->
      (exists v, t <= v /\ last_idle_or_long_overlap_aux N idle conf id te nrange = Some v).
  Proof.
    intros * Hidle Ht.
    induction nrange.
    {
      simpl in Ht.
      omega.
    }
    {
      destruct (zeq t (Z.of_nat nrange)) as [Heq | Hneq].
      {
        exists t.
        split; try omega.
        simpl.
        rewrite <- Heq in *.
        rewrite Hidle.
        rewrite zeq_true.
        auto.
      }
      {
        assert (H1: 0 <= t < Z.of_nat nrange). xomega.
        specialize (IHnrange H1).
        simpl.
        destruct (zeq idle
           (highest_not_finished_map idle (Z.of_nat nrange)
              (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
              (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf))) as [Hcur|Hbusy].
        {
          exists (Z.of_nat nrange).
          split; auto; try omega.
        }
        {
          destruct (ZMap.get
         (highest_not_finished_map idle (Z.of_nat nrange)
            (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf)) conf) as [T C].
          destruct (if zlt te (Z.of_nat nrange / T * T + T)
         then true
         else
          if zeq te (Z.of_nat nrange / T * T + T)
          then
           if zlt id
                (highest_not_finished_map idle (Z.of_nat nrange)
                   (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
                   (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf))
           then true
           else false
          else false).
          {
            exists (Z.of_nat nrange).
            split; auto; try omega.
          }
          {
            auto.
          }
        }
      }
    }
  Qed.
          
  Lemma no_idle_within_burst:
    forall N conf id v ts te t i,
      highest_not_finished_map N t (TAC_pool_at_t t (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) t conf) = i ->
      0 < N ->
      v <= t < ts ->
      0 <= v ->
      start_of_burst N conf ts id te = v ->
      0 <= i < N.
  Proof.
    intros * Hval HN Ht Hv Hstart.
    assert (H1: 0 <= i <= N).
    {
      rewrite <- Hval.
      apply highest_not_finished_map_res_bound; auto.
      - remember (Z.to_nat N) as Nnat.
        replace N with (Z.of_nat Nnat).
        apply real_priqueue_sound.
        + rewrite HeqNnat.
          rewrite Z2Nat.id; omega.
    }

    destruct (zeq i N) as [Heq | ?]; auto; try omega.
    unfold start_of_burst in Hstart.
    rewrite Heq in Hval.
    assert (H2: 0 <= t < Z.of_nat (Z.to_nat ts)).
    { rewrite Z2Nat.id; omega. }
    pose proof (idle_slot_is_found_if_exists _ _ _ id _ te _ Hval H2) as H.
    destruct H as [v0 [Hv0 Hlast]].
    rewrite Hlast in Hstart.
    omega.
  Qed.

  Lemma lower_priority_slot_is_found_if_exists:
    forall N idle conf id nrange te t i T C,
      highest_not_finished_map idle t (TAC_pool_at_t t (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) t conf) = i ->
      0 <= t < Z.of_nat nrange ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      priEntry_lt (PriEntry te id) (PriEntry (t / T * T + T) i) = true ->
      (exists v, t <= v /\ last_idle_or_long_overlap_aux N idle conf id te nrange = Some v).
  Proof.
    intros * Hval Ht Hconfi Hlower.
    induction nrange.
    {
      simpl in Ht.
      omega.
    }
    {
      simpl.
      destruct (zeq idle
           (highest_not_finished_map idle (Z.of_nat nrange)
              (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
              (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf))) as [Hidle|Hslot].
      {
        exists (Z.of_nat nrange).
        split; auto; try xomega.
      }
      {
        destruct (zeq t (Z.of_nat nrange)) as [Heq | Hneq].
        {
          rewrite <- Heq in *.
          rewrite Hval.
          rewrite Hconfi.
          exists t.
          split; try omega.
          unfold priEntry_lt in Hlower.
          rewrite Hlower.
          auto.
        }
        {
          destruct (ZMap.get
         (highest_not_finished_map idle (Z.of_nat nrange)
            (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf)) conf) as [Tn Cn].
          destruct (if zlt te (Z.of_nat nrange / Tn * Tn + Tn)
         then true
         else
          if zeq te (Z.of_nat nrange / Tn * Tn + Tn)
          then
           if zlt id
                (highest_not_finished_map idle (Z.of_nat nrange)
                   (TAC_pool_at_t (Z.of_nat nrange) (Z.to_nat N) conf) conf
                   (real_priqueue (Z.to_nat N) (Z.of_nat nrange) conf))
           then true
           else false
          else false).
          {
            exists (Z.of_nat nrange).
            split; auto; try xomega.
          }
          {
            apply IHnrange.
            xomega.
          }
        }
      }
    }
  Qed.
  
  Lemma lower_priority_not_exists_within_burst:
    forall N conf id v ts te t i T C,
      highest_not_finished_map N t (TAC_pool_at_t t (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) t conf) = i ->
      0 < N ->
      v <= t < ts ->
      0 <= v ->
      start_of_burst N conf ts id te = v ->      
      ZMap.get i conf = mkPrioConfigValid T C ->
      priEntry_lt (PriEntry te id) (PriEntry (t / T * T + T) i) = false.
  Proof.
    intros * Hval HN Ht Hv Hstart Hconfi.
    assert (Hi: 0 <= i < N).
    { eapply no_idle_within_burst; eauto. }

    destruct (priEntry_lt (PriEntry te id) (PriEntry (t / T * T + T) i)) eqn:Hpri; auto.
    unfold start_of_burst in Hstart.
    assert (H1: 0 <= t < Z.of_nat (Z.to_nat ts)).
    { rewrite Z2Nat.id; omega. }
    
    pose proof (lower_priority_slot_is_found_if_exists _ _ _ _ _ _ _ _ _ _
               Hval H1 Hconfi Hpri).
    destruct H as [v0 [Hrange Hlast]].
    rewrite Hlast in Hstart.
    omega.
  Qed.

  Lemma ready_task_must_be_within_burst:
    forall N conf id v ts te t i T C,
      1 < T ->
      0 <= v <= t ->
      start_of_burst N conf ts id te = v ->      
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 <= i < N ->
      vt_map_not_finished t T C (ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf)) = true ->
      priEntry_lt (PriEntry te id) (PriEntry (t / T * T + T) i) = false ->
      v <= t / T * T.
  Proof.
    intros * HT Hv Hstart Hconfi Hi Hfinish Hpri.
    unfold start_of_burst in Hstart.
    destruct (last_idle_or_long_overlap_aux N N conf id te (Z.to_nat ts)) as [v0|] eqn:Hlast.
    {
      rewrite <- Hstart in *.
      destruct (zle (t / T * T) v0); try omega.

      assert (H1: 0 <= t / T * T).
      { apply div_mul_nonneg; omega. }
      assert (H2: t / T * T <= t).
      { apply divide_mul_le; omega. }
      assert (H3: (t / T * T) mod T = 0).
      { apply Z_mod_mult. }
      assert (H4: t < t / T * T + T).
      { apply x_lt_floor_plus; omega. }
      assert (H5: v0 / T * T = t / T * T).
      { apply divide_mul_plus_equal2; try omega. }
      
      assert (Hfinish0: vt_map_not_finished v0 T C (ZMap.get i (TAC_pool_at_t v0 (Z.to_nat N) conf)) = true).
      {
        unfold vt_map_not_finished in *.
        rewrite H5.
        
        assert (H10: ZMap.get i (TAC_pool_at_t (t/ T*T) (Z.to_nat N) conf) (t / T * T) = ZMap.get i (TAC_pool_at_t v0 (Z.to_nat N) conf) (t / T * T)).
        {
          symmetry.
          rewrite vertical_map_canonical; omega.
        }
        rewrite <- H10.

        assert (H11: ZMap.get i (TAC_pool_at_t v0 (Z.to_nat N) conf) v0 <= ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t).
        { apply TAC_pool_at_t_monotonic_two_points; auto; try omega. }

        assert (H12: ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) (t / T * T) = ZMap.get i (TAC_pool_at_t (t/T*T) (Z.to_nat N) conf) (t / T * T)).
        {
          rewrite vertical_map_canonical; omega.
        }        
        rewrite H12 in Hfinish.
        
        destruct (zlt
              (ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t -
               ZMap.get i (TAC_pool_at_t (t/T*T) (Z.to_nat N) conf) (t / T * T)) C); try discriminate.

        destruct (zlt
     (ZMap.get i (TAC_pool_at_t v0 (Z.to_nat N) conf) v0 -
      ZMap.get i (TAC_pool_at_t (t / T * T) (Z.to_nat N) conf) (t / T * T)) C); try omega.
        auto.
      }

      remember (highest_not_finished_map N v0 (TAC_pool_at_t v0 (Z.to_nat N) conf) conf
                                         (real_priqueue (Z.to_nat N) v0 conf)) as j.
      destruct (ZMap.get j conf) as [Tj Cj] eqn:Hconfj.
      assert (Hlast': j = N \/ (priEntry_lt (PriEntry te id) (PriEntry (v0 / Tj * Tj + Tj) j) = true)).
      {
        eapply last_idle_or_long_overlap_aux_res_idle_or_long with (N:=N) (nrange:=Z.to_nat ts); eauto.
      }

      assert (Hj: 0 <= j < N).
      {
        rewrite Heqj.
        eapply exists_ready_task_highest_not_finished_map_schedule_something; eauto.
      }
      destruct Hlast'; try omega.

      destruct (zeq i j) as [Heq | Hneq].
      {
        subst.
        congruence.
      }
      {
        symmetry in Heqj.
        pose proof (highest_not_finished_map_has_highest_priority_among_ready_ones _ _ _ _ _ _ _ Hj Hconfj Heqj) as Hjj.
        specialize (Hjj _ _ _ Hi Hneq Hconfi).
        destruct Hjj; try congruence.
        rewrite H5 in *.
        
        assert (is_true (priEntry_lt (PriEntry te id) (PriEntry (t / T * T + T) i))).
        {
          exploit priEntry_lt_transitive.
          - apply H.
          - apply H0.
          - auto.
        }
        congruence.
      }
    }
    {
      rewrite <- Hstart.
      apply div_mul_nonneg; omega.
    }
  Qed.
  
  Lemma sched_etype_is_busy_within_burst:
    forall N conf id v ts te t i T C T0 C0,
      highest_not_finished_map N t (TAC_pool_at_t t (Z.to_nat N) conf) conf
                               (real_priqueue (Z.to_nat N) t conf) = i ->
      v <= t < ts ->
      ts < te ->
      0 <= v ->
      0 < N ->
      start_of_burst N conf ts id te = v ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      ZMap.get id conf = mkPrioConfigValid T0 C0 ->
      ts mod T0 = 0 ->
      config_pool_valid N conf ->
      (0 <= i < N /\ get_esched_type i T id t v ts te <> ESched_other).
  Proof.
    intros * Hvalue Ht Hte Hv HN Hstart Hconfi Hconfid Htsmod Hconfvalid.
    assert (H1: 0 <= i < N).
    { eapply no_idle_within_burst; eauto. }
    split; auto.

    unfold not.
    intros H.
    unfold get_esched_type in H.

    assert (H2: 1 < T).
    { specialize (Hconfvalid _ _ _ H1 Hconfi). omega. }
    
    assert (H3: t / T * T <= t).
    { apply divide_mul_le; omega. }

    assert (Hlelt: v > t / T * T \/ t / T * T >= ts -> False).
    {
      intros o.
      destruct o; try omega.
      symmetry in Hvalue.

      assert (Hfinish: vt_map_not_finished t T C (ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf)) = true).
      {
        remember (Z.to_nat N) as Nnat.
        assert (HNN: N = Z.of_nat Nnat).
        { rewrite HeqNnat. rewrite Z2Nat.id; omega. }
        rewrite HNN in *.
        eapply highest_not_finished_map_condition_met; eauto.
      }

      assert (Hpri: priEntry_lt (PriEntry te id) (PriEntry (t / T * T + T) i) = false).
      { eapply lower_priority_not_exists_within_burst; eauto. }
      
      assert (v <= t / T * T).
      {
        eapply ready_task_must_be_within_burst; eauto.
        omega.
      }
      omega.
    }

    destruct (zle v (t / T * T)); auto.
    destruct (zlt (t / T * T) ts); auto.
    {
      destruct (zle (t / T * T + T) ts); try congruence.
      destruct (zeq i id) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite Hconfi in Hconfid.
        inv Hconfid.
        assert (ts mod T0 <> 0).
        {
          apply within_boundary_mod_not_0 with (c := t / T0 * T0 + T0); try omega.
          rewrite mod_0_preserve; try omega.
          apply Z_mod_mult.
        }
        omega.
      }
      {
        destruct (priEntry_lt (PriEntry (t / T * T + T) i) (PriEntry te id)) eqn:Hpri; try congruence.
        assert (Hpri': priEntry_lt  (PriEntry te id) (PriEntry (t / T * T + T) i) = false).
        {
          eapply lower_priority_not_exists_within_burst; eauto.
        }
        unfold priEntry_lt in *.
        destruct (zlt (t / T * T + T) te); try congruence.
        destruct (zlt te (t / T * T + T)); try congruence.
        assert (H5: t /T * T + T = te). omega.
        rewrite H5 in *.
        rewrite zeq_true in *.
        destruct (zlt i id); try congruence.
        rewrite zlt_true in Hpri'; try omega.
        congruence.
      }
    }
  Qed.

  Lemma sched_etype_is_complete_when_schedule_is_busy:
    forall N conf id v ts te t1 t2 Tid Cid,
      start_of_burst N conf ts id te = v ->
      v <= t1 <= t2 ->
      t2 <= ts < te ->
      0 < N ->
      ZMap.get id conf = mkPrioConfigValid Tid Cid ->
      ts mod Tid = 0 ->
      config_pool_valid N conf ->
      count_etype_occurrence N N t1 t2 conf id v ts te ESched_whole +
      count_etype_occurrence N N t1 t2 conf id v ts te ESched_overlap = t2 - t1.
  Proof.
    intros * Hstart Ht1 Ht2 HN Hconfid Htsmod Hconfvalid.
    unfold count_etype_occurrence.
    remember (Z.to_nat (t2 - t1)) as nrange.

    assert (H1: t2 = t1 + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite H1 in *.
    clear Heqnrange H1.
    replace (t1 + Z.of_nat nrange - t1) with (Z.of_nat nrange) by omega.
    
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H3: v <= t1 <= t1 + Z.of_nat nrange). xomega.
      assert (H4: t1 + Z.of_nat nrange <= ts < te). xomega.
      specialize (IHnrange H3 H4).
      
      remember ((highest_not_finished_map N (t1 + Z.of_nat nrange)
            (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (t1 + Z.of_nat nrange) conf))) as i.
      destruct (ZMap.get i conf) as [T C] eqn:Hconf.
      symmetry in Heqi.
      assert (H2: v <= t1 + Z.of_nat nrange < ts). xomega.
      assert (H10: ts < te). omega.
      assert (H11: 0 <= v).
      {
        rewrite <- Hstart.
        apply start_of_burst_result_nonneg.
      }
       
      pose proof (sched_etype_is_busy_within_burst _ _ _ _ _ _ _ _ _ _ _ _ Heqi H2 H10 H11 HN Hstart Hconf Hconfid Htsmod Hconfvalid) as H.
      destruct H as [Hbound Htype].
      repeat rewrite zeq_false; try omega.      

      destruct (get_esched_type i T id (t1 + Z.of_nat nrange) v ts te); try congruence.
      - rewrite esched_type_eq_true.
        rewrite esched_type_eq_false; try congruence.
        xomega.
      - rewrite esched_type_eq_true.
        rewrite esched_type_eq_false; try congruence.
        xomega.
    }
  Qed.

End BURST.


Section ESCHED_OVERLAP.

  Lemma is_overlapping_etype_is_overlap_generalized:
    forall fromid Tf toid v ts te t,
      0 <= ts < te ->
      1 < Tf ->
      ts / Tf * Tf <= t < ts ->
      is_overlapping fromid Tf toid v ts te = true ->
      fromid <> toid ->
      get_esched_type fromid Tf toid t v ts te = ESched_overlap. 
  Proof.
    intros * Hts HTf Ht Hoverlap Hid.
    unfold is_overlapping in Hoverlap.
    unfold get_esched_type.
    rewrite zeq_false; auto.

    assert (H1: (ts / Tf * Tf) mod Tf = 0). apply Z_mod_mult.
    assert (H2: ts < ts / Tf * Tf + Tf).
    {
      apply x_lt_floor_plus; omega.
    }
    assert (H3: 0 <= ts / Tf * Tf).
    { apply div_mul_nonneg; omega. }
      
    assert (H4: t / Tf * Tf = ts / Tf * Tf).
    {
      apply mod_lower_bound; try omega.
    }

    rewrite H4 in *.

    destruct (zle v (ts / Tf * Tf)); try congruence.
    destruct (zlt (ts / Tf * Tf) ts); try congruence.
    destruct (priEntry_lt (PriEntry (ts / Tf * Tf + Tf) fromid)
                          (PriEntry te toid)) eqn:Hpri; try congruence.
    rewrite zle_false; try omega.
    auto.
  Qed.

  Lemma is_overlapping_yet_not_e_overlap_when_t_is_small:
    forall fromid Tf toid v ts te t,
      0 <= v ->      
      v <= t < ts / Tf * Tf ->
      1 < Tf ->
      is_overlapping fromid Tf toid v ts te = true ->      
      get_esched_type fromid Tf toid t v ts te <> ESched_overlap.
  Proof.
    intros * Hv Ht HTf Hoverlap.
    unfold get_esched_type, not.
    intros H.
    destruct (zle v (t / Tf * Tf)); try congruence.
    destruct (zlt (t / Tf * Tf) ts); try congruence.
    destruct (zle (t / Tf * Tf + Tf) ts); try congruence.    

    assert (H1: t / Tf * Tf + Tf <= ts / Tf * Tf + Tf - Tf).
    {
      apply mod_0_bound_increase; try omega.
      - rewrite mod_0_preserve; try omega.
        apply Z_mod_mult.
    }
    assert (H2: ts / Tf * Tf <= ts).
    {
      apply divide_mul_le; omega.
    }
    omega.
  Qed.
  
  Lemma is_overlapping_then_single_task_etype_count_is_0_with_small_t:
    forall N fromid toid v ts te Tf Cf conf nrange,
      0 <= v ->
      v + Z.of_nat nrange <= ts / Tf * Tf ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf ->
      is_overlapping fromid Tf toid v ts te = true ->
      count_single_task_etype_aux N N fromid v nrange conf toid v ts te ESched_overlap = 0.      
  Proof.
    intros * Hv Hrange Hconf HTf Hoverlap.    
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H1: Z.of_nat nrange < Z.of_nat (S nrange)). xomega.
      rewrite IHnrange; try omega.
      destruct (zeq
         (highest_not_finished_map N (v + Z.of_nat nrange)
            (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (v + Z.of_nat nrange) conf)) fromid) as [Heq | Hneq]; auto.
      rewrite Heq.
      rewrite Hconf.
      rewrite esched_type_eq_false; auto.
      apply is_overlapping_yet_not_e_overlap_when_t_is_small; auto; try omega.
    }
  Qed.
  
  Lemma is_overlapping_then_single_task_etype_count_is_valid:
    forall N fromid toid v ts te Tf Cf conf nrange,
      0 < N ->
      0 <= v ->
      ts < te ->
      ts / Tf * Tf < v + Z.of_nat nrange <= ts ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf  ->
      0 < Cf ->
      is_overlapping fromid Tf toid v ts te = true ->
      fromid <> toid ->
      0 <= fromid < N ->
      count_single_task_etype_occurrence N N fromid v (v + Z.of_nat nrange) conf toid v ts te ESched_overlap =
      Z.min Cf (ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (v + Z.of_nat nrange) - ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) ((v + Z.of_nat nrange) / Tf * Tf)).
  Proof.
    intros * HN Hv Hte Hrange Hconf HTf HCf Hoverlap Hid Hfrom.
    unfold count_single_task_etype_occurrence.
    replace (v + Z.of_nat nrange - v) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl in Hrange.
      unfold is_overlapping in Hoverlap.
      destruct (zle v (ts / Tf * Tf)); try congruence.
      destruct (zlt (ts / Tf * Tf) ts); try congruence.
      omega.
    }
    {
      assert (HNN: Z.of_nat (Z.to_nat N) = N).
      { rewrite Z2Nat.id; omega. }      
      assert (H1: Z.of_nat nrange < Z.of_nat (S nrange)). xomega.
      assert (H2: v + Z.of_nat (S nrange) = (v + Z.of_nat nrange) + 1). xomega.
      rewrite H2.
      assert (H3: (ts / Tf * Tf) mod Tf = 0).
      { apply Z_mod_mult. }      
      
      assert (H10: 0 <= ts / Tf * Tf).
      {
        apply Zmult_gt_0_le_0_compat; try omega.
        apply Z_div_pos; omega.
      }
      assert (H11: 0 <= fromid < Z.of_nat (Z.to_nat N)). omega.
      
      assert (H12: (v + Z.of_nat nrange + 1) mod Tf <> 0).
      {            
        apply a_between_floor_and_self_mod_not_0 with (c := ts); omega.
      }
      assert (H15: 0 <= (v + Z.of_nat nrange) / Tf * Tf).
      { 
        apply Zmult_gt_0_le_0_compat; try omega.
        apply Z_div_pos; omega.
      }
      assert (H20: (v + Z.of_nat nrange) / Tf * Tf <= v + Z.of_nat nrange).
      {
        apply divide_mul_le; omega.
      }      
      
      
      simpl.
      destruct (zeq
         (highest_not_finished_map N (v + Z.of_nat nrange)
            (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (v + Z.of_nat nrange) conf)) fromid) as [Hsched|Hnot].
      {
        rewrite Hsched in *.
        rewrite Hconf.

        assert (Htype: get_esched_type fromid Tf toid (v + Z.of_nat nrange) v ts te = ESched_overlap).
        {
          apply is_overlapping_etype_is_overlap_generalized; auto; xomega.
        }
        rewrite Htype.
        rewrite esched_type_eq_true.

        
        destruct (zeq (v + Z.of_nat nrange) (ts / Tf * Tf)) as [Heq|Hneq].
        {
          rewrite Heq in *.

          assert (H0: count_single_task_etype_aux N N fromid v nrange conf toid v ts te ESched_overlap = 0).
          {
            apply is_overlapping_then_single_task_etype_count_is_0_with_small_t with (Tf:=Tf) (Cf:=Cf); auto; xomega.            
          }
          rewrite H0.

          assert (H4: (ts / Tf * Tf + 1) / Tf * Tf = ts / Tf * Tf).
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; omega.
          }
          rewrite H4.

          assert (H5: ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (ts / Tf * Tf + 1) =
                      ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (ts / Tf * Tf) + 1).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_increase_for_the_scheduled_task; auto; try omega.
            rewrite Z2Nat.id; omega.            
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          rewrite IHnrange; try omega.
          rewrite mod_equal; try omega.

          assert (H14: ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (v + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (v + Z.of_nat nrange) + 1).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.

            apply vt_increase_for_the_scheduled_task; auto; try xomega.
            rewrite Z2Nat.id; omega.            
          }
          rewrite H14.

          assert (H16: ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (v + Z.of_nat nrange) -
                       ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) ((v + Z.of_nat nrange) / Tf * Tf) < Cf).
          {
            symmetry in Hsched.
            replace N with (Z.of_nat (Z.to_nat N)) in Hsched.
            
            rewrite Nat2Z.id in Hsched.
            pose proof (highest_not_finished_map_condition_met _ _ _ _ _ _ _ Hsched H11 Hconf).
            unfold vt_map_not_finished in H.
            rewrite vertical_map_canonical; try omega.
            assert ((v + Z.of_nat nrange) / Tf * Tf <= v + Z.of_nat nrange).
            { apply divide_mul_le; omega. }
            assert (ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf)
                             ((v + Z.of_nat nrange) / Tf * Tf) = ZMap.get fromid (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf)
                                                                           ((v + Z.of_nat nrange) / Tf * Tf)).
            {
              rewrite vertical_map_canonical; try omega.
              symmetry.
              rewrite vertical_map_canonical; try omega.                
            }
            rewrite H4.
            destruct (zlt
        (ZMap.get fromid (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf)
           (v + Z.of_nat nrange) -
         ZMap.get fromid (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf)
           ((v + Z.of_nat nrange) / Tf * Tf)) Cf); try omega.
            discriminate H.
          }

          repeat rewrite Z.min_r; omega.          
        }
      }
      {
        (* fromid not scheduled *)
        destruct (zeq (v + Z.of_nat nrange) (ts / Tf * Tf)) as [Heq | Hneq].
        {          
          rewrite Heq in *.

          assert (H0: count_single_task_etype_aux N N fromid v nrange conf toid v ts te ESched_overlap = 0).
          {
            apply is_overlapping_then_single_task_etype_count_is_0_with_small_t with (Tf:=Tf) (Cf:=Cf); auto; xomega.            
          }
          rewrite H0.

          assert (H4: (ts / Tf * Tf + 1) / Tf * Tf = ts / Tf * Tf).
          {
            rewrite divide_plus_equal; try omega.
            rewrite mod_0_divide_mul; omega.
          }
          rewrite H4.

          assert (H6: vt_map_not_finished (ts / Tf * Tf) Tf Cf
                                          (ZMap.get fromid (TAC_pool_at_t (ts / Tf * Tf) (Z.to_nat N) conf)) = true).
          {
            apply vt_map_not_finished_true_at_beginning; auto; omega.
          }
          
          assert (H5: ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (ts / Tf * Tf + 1) =
                      ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (ts / Tf * Tf)).
          {
            rewrite vertical_map_canonical; try omega.
            symmetry.
            rewrite vertical_map_canonical; try omega.
            symmetry.
            apply vt_stagnates_if_task_not_scheduled with (T:=Tf) (C:=Cf); auto; try omega.
            - rewrite Z2Nat.id; try omega.
              apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=fromid) (T:=Tf) (C:=Cf); auto.
            - rewrite Z2Nat.id; omega.
          }
          rewrite H5.
          rewrite Z.min_r; omega.
        }
        {
          rewrite IHnrange; try omega.
          rewrite mod_equal; try omega.

          destruct (vt_map_not_finished (v + Z.of_nat nrange) Tf Cf
     (ZMap.get fromid (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf))) eqn:Hfinish.
          {
            assert (H5: ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf)
                                 (v + Z.of_nat nrange + 1) =
                   ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf)
                            (v + Z.of_nat nrange)).
            {
              rewrite vertical_map_canonical; try omega.
              symmetry.
              rewrite vertical_map_canonical; try omega.
              symmetry.
              apply vt_stagnates_if_task_not_scheduled with (T:=Tf) (C:=Cf); auto; try omega.
              - rewrite Z2Nat.id; try omega.
                apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=fromid) (T:=Tf) (C:=Cf); auto.
              - rewrite Z2Nat.id; omega.
            }
            rewrite H5.
            auto.            
          }
          {
            unfold vt_map_not_finished in Hfinish.            

            assert (H16: ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (v + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (v + Z.of_nat nrange) \/
                         ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (v + Z.of_nat nrange + 1) = ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf) (v + Z.of_nat nrange) + 1).
            {
              assert (H17: monotonic_continuous (ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf))).
              {
                apply TAC_pool_all_monotonic_continuous; auto; try omega.
              }
              apply H17.
              omega.
            }

            destruct H16 as [Heq | Hinc]; auto.
            {
              rewrite Heq; auto.
            }
            {
              rewrite Hinc.
              
              rewrite vertical_map_canonical; try omega.
              assert (H18: ZMap.get fromid (TAC_pool_at_t te (Z.to_nat N) conf)
        ((v + Z.of_nat nrange) / Tf * Tf) = ZMap.get fromid (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf)
                                                      ((v + Z.of_nat nrange) / Tf * Tf)).
              {              
                rewrite vertical_map_canonical; try omega.
                symmetry.
                rewrite vertical_map_canonical; try omega.                
              }
              rewrite H18.

              destruct (zlt
              (ZMap.get fromid (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf)
                 (v + Z.of_nat nrange) -
               ZMap.get fromid (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf)
                 ((v + Z.of_nat nrange) / Tf * Tf)) Cf); try discriminate.
              rewrite Z.min_l; try omega.
              rewrite Z.min_l; try omega.              
            }
          }
        }
      }
    }    
  Qed.

  Lemma count_single_task_self_esched_overlap_is_0:
    forall N1 N2 t1 t2 conf i v ts te,      
      count_single_task_etype_occurrence N1 N2 i t1 t2 conf i v ts te ESched_overlap = 0.
  Proof.
    intros *.
    unfold count_single_task_etype_occurrence.
    remember (Z.to_nat (t2 - t1)) as range.
    clear Heqrange.
    induction range.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHrange.
      destruct (zeq
         (highest_not_finished_map N2 (t1 + Z.of_nat range)
            (TAC_pool_at_t (t1 + Z.of_nat range) (Z.to_nat N1) conf) conf
            (real_priqueue (Z.to_nat N1) (t1 + Z.of_nat range) conf)) i) as [Heq|Hneq]; auto.
      {
        rewrite Heq.        
        destruct (ZMap.get i conf) as [Ti Ci] eqn:Hconf.
        rewrite esched_type_eq_false; auto.
        unfold not.
        intros H.
        unfold get_esched_type in H.
        rewrite zeq_true in H.
        destruct (zle v ((t1 + Z.of_nat range) / Ti * Ti)); try congruence.
        destruct (zlt ((t1 + Z.of_nat range) / Ti * Ti) ts); try congruence.
        destruct (zle ((t1 + Z.of_nat range) / Ti * Ti + Ti) ts); try congruence.
      }
    }
  Qed.

  Lemma none_overlapping_then_type_not_overlap:
    forall fromid Tf toid v ts te t,
      0 <= v ->
      v <= t < ts ->
      0 < Tf ->
      is_overlapping fromid Tf toid v ts te = false ->
      get_esched_type fromid Tf toid t v ts te <> ESched_overlap.
  Proof.
    intros * Hv Ht HTf Hnone.
    unfold is_overlapping in Hnone.
    unfold get_esched_type.

    destruct (zle v (t / Tf * Tf)); try congruence.
    destruct (zlt (t / Tf * Tf) ts); try congruence.
    destruct (zle (t / Tf * Tf + Tf) ts); try congruence.
    destruct (zeq fromid toid); try congruence.

    assert (H1: (t / Tf * Tf) mod Tf = 0). apply Z_mod_mult.
    assert (H2: ts < ts / Tf * Tf + Tf).
    {
      apply x_lt_floor_plus; omega.
    }
    assert (H3: 0 <= ts / Tf * Tf).
    { apply div_mul_nonneg; omega. }
    
    assert (H4: ts / Tf * Tf = t / Tf * Tf).
    {
      apply mod_lower_bound; try omega.
    }
    rewrite <- H4.

    destruct (priEntry_lt (PriEntry (ts / Tf * Tf + Tf) fromid) (PriEntry te toid)) eqn:Hpri; try congruence.
    destruct (zle v (ts / Tf * Tf)); try omega.
    destruct (zlt (ts / Tf * Tf) ts); try omega.
    congruence.
  Qed.
  
  Lemma non_overlapping_not_counted_by_single_task_etype_count:
    forall N fromid toid v ts te Tf Cf conf nrange,
      0 <= v ->
      v + Z.of_nat nrange <= ts ->
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      1 < Tf ->
      is_overlapping fromid Tf toid v ts te = false ->
      count_single_task_etype_occurrence N N fromid v (v + Z.of_nat nrange) conf toid v ts te ESched_overlap = 0.
  Proof.
    intros * Hv Hrange Hconf HTf Hoverlap.
    unfold count_single_task_etype_occurrence.
    replace (v + Z.of_nat nrange - v) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      assert (H1: Z.of_nat nrange < Z.of_nat (S nrange)). xomega.
      simpl.
      rewrite IHnrange; try omega.
      destruct (zeq
         (highest_not_finished_map N (v + Z.of_nat nrange)
            (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (v + Z.of_nat nrange) conf)) fromid) as [Heq | Hneq]; auto.

      rewrite Heq.
      rewrite Hconf.
      assert (Htype: get_esched_type fromid Tf toid (v + Z.of_nat nrange) v ts te <> ESched_overlap).
      {
        apply none_overlapping_then_type_not_overlap; auto; try omega.
      }
      rewrite esched_type_eq_false; auto.
    }
  Qed.
  
  Lemma count_e_occurrence_overlapping_bounded_aux:
    forall Nnat N0 id v ts te conf,
      Z.of_nat Nnat <= N0 ->
      0 < N0 ->
      0 <= v ->
      v <= ts < te ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      count_etype_occurrence2_aux Nnat N0 N0
                           v ts conf id v ts te ESched_overlap =
      workload_overlapping_remain_total Nnat id v ts te ts conf (TAC_pool_at_t te (Z.to_nat N0) conf).
  Proof.
    intros * HNnat HN0 Hv Hts Hconfvalid.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (Hconfvalid': config_pool_valid (Z.of_nat Nnat) conf).
      { apply config_pool_valid_shrink with (N2 := Z.of_nat (S Nnat)); auto; omega. }
      assert (H10: Z.of_nat Nnat <= N0). xomega.
      specialize (IHNnat H10 Hconfvalid').
      simpl.
      rewrite <- IHNnat.

      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq]; auto.
      {
        (* id = Z.of_nat Nnat *)
        rewrite <- Heq.
        rewrite count_single_task_self_esched_overlap_is_0; omega.
      }
      {
        (* id <> Z.of_nat Nnat *)
        destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tf Cf] eqn:Hconff.
        destruct (is_overlapping (Z.of_nat Nnat) Tf id v ts te) eqn:Hoverlap.
        {
          assert (H: forall a b c, b = c -> a + b = a + c). intros. omega.
          apply H.

          remember (Z.to_nat (ts - v)) as nrange.
          assert (HTT: ts = v + Z.of_nat nrange).
          { rewrite Heqnrange. rewrite Z2Nat.id; omega. }          
          rewrite HTT in *. clear HTT Heqnrange.

          assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.          
          assert (H3: 1 < Tf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          assert (H4: 0 < Cf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }
          
          apply is_overlapping_then_single_task_etype_count_is_valid; auto; try omega.
          - pose proof (is_overlapping_true_end_not_aligned _ _ _ _ _ _ Hoverlap).
            assert ((v + Z.of_nat nrange) / Tf * Tf <= v + Z.of_nat nrange).
            { apply divide_mul_le; omega. }
            omega.
        }
        {
          assert (H: forall a b, b = 0 -> a + b = a). intros. omega.
          apply H.

          remember (Z.to_nat (ts - v)) as nrange.
          assert (HTT: ts = v + Z.of_nat nrange).
          { rewrite Heqnrange. rewrite Z2Nat.id; omega. }          
          rewrite HTT in *. clear HTT Heqnrange.
          
          assert (H2: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)). omega.
          assert (H3: 1 < Tf).
          {
            specialize (Hconfvalid _ _ _ H2 Hconff).
            omega.
          }

          apply non_overlapping_not_counted_by_single_task_etype_count with (Tf := Tf) (Cf := Cf); auto; try omega.
        }
      }
    }
  Qed.

  Lemma count_e_occurrence_overlapping_bounded:
    forall Nnat id v ts te conf,
      0 < Z.of_nat Nnat ->
      0 <= v ->
      v <= ts < te ->
      config_pool_valid (Z.of_nat Nnat) conf ->
      count_etype_occurrence (Z.of_nat Nnat) (Z.of_nat Nnat)
                           v ts conf id v ts te ESched_overlap =
      workload_overlapping_remain_total Nnat id v ts te ts conf (TAC_pool_at_t te Nnat conf).
  Proof.
    intros * HN Hv Hts Hconfvalid.
    
    rewrite count_etype_occurrence2_eq; try omega.
    assert (H1: (Z.to_nat (Z.of_nat Nnat)) = Nnat).
    { rewrite Nat2Z.id. auto. }
    rewrite H1.

    rewrite count_e_occurrence_overlapping_bounded_aux; auto; try omega.
    - rewrite H1. auto.
  Qed.    

  Lemma non_overlapping_no_remain_generalized:
    forall fromid toid Tf Cf v ts te N conf,
      0 <= v <= ts ->
      0 <= fromid < N ->
      start_of_burst N conf ts toid te = v ->   
      ZMap.get fromid conf = mkPrioConfigValid Tf Cf ->
      config_pool_valid N conf ->
      is_overlapping fromid Tf toid v ts te = false ->
      workload_remain toid fromid Tf Cf ts te (ZMap.get fromid (TAC_pool_at_t ts (Z.to_nat N) conf)) = 0.      
  Proof.
    intros * Hv Hid Hburst Hconf Hconfvalid Hoverlap.
    unfold is_overlapping in Hoverlap.
    unfold workload_remain.
    destruct (zeq (ts mod Tf) 0); auto.

    assert (Hlelt: v > (ts / Tf * Tf) \/ (ts / Tf * Tf) >= ts ->
           (if priEntry_lt (PriEntry (ts / Tf * Tf + Tf) fromid) (PriEntry te toid)
    then
     Z.max 0
       (Cf - ZMap.get fromid (TAC_pool_at_t ts (Z.to_nat N) conf) ts +
        ZMap.get fromid (TAC_pool_at_t ts (Z.to_nat N) conf) (ts / Tf * Tf))
            else 0) = 0).
    {
      intros o.
      assert (H1: 1 < Tf).
      {
        specialize (Hconfvalid _ _ _ Hid Hconf).
        omega.
      }
      assert (H2: ts - ts / Tf * Tf >= 1).
      { apply z_gt_mod; try omega. }      
      destruct o; try omega.

      destruct (priEntry_lt (PriEntry (ts / Tf * Tf + Tf) fromid) (PriEntry te toid)) eqn:Hpri; auto.
      assert (Hfinish: vt_map_not_finished ts Tf Cf (ZMap.get fromid (TAC_pool_at_t ts (Z.to_nat N) conf)) = false).
      {
        destruct (vt_map_not_finished ts Tf Cf (ZMap.get fromid (TAC_pool_at_t ts (Z.to_nat N) conf))) eqn:Hfinish; auto.
        exfalso.        

        assert (v <= ts / Tf * Tf).
        {          
          apply ready_task_must_be_within_burst with (id:=toid) (i:=fromid) (N:=N) (conf:=conf) (ts:=ts) (te:=te) (C:=Cf); auto.
          - apply priEntry_lt_neg_tf; auto.
        }
        omega.
      }
      unfold vt_map_not_finished in Hfinish.
      destruct (zlt
              (ZMap.get fromid (TAC_pool_at_t ts (Z.to_nat N) conf) ts -
               ZMap.get fromid (TAC_pool_at_t ts (Z.to_nat N) conf) (ts / Tf * Tf)) Cf); try discriminate.
      rewrite Z.max_l; omega.
    }

    destruct (zle v (ts / Tf * Tf)); auto.
    destruct (zlt (ts / Tf * Tf) ts); auto.
    {
      destruct (priEntry_lt (PriEntry (ts / Tf * Tf + Tf) fromid) (PriEntry te toid));
        try congruence.
    }
  Qed.
  
  Lemma workload_overlapping_remain_total_eq_burst:
    forall Nnat i v ts te conf N,
      Z.of_nat Nnat <= N ->
      start_of_burst N conf ts i te = v -> 
      0 <= ts < te ->
      config_pool_valid N conf ->
      workload_overlapping_total Nnat i v ts te conf -
      workload_remain_total Nnat i ts te conf (TAC_pool_at_t ts (Z.to_nat N) conf) = 
      workload_overlapping_remain_total Nnat i v ts te ts conf (TAC_pool_at_t ts (Z.to_nat N) conf).
  Proof.
    intros * HN Hburst Hts Hconfvalid.
    induction Nnat.
    {
      simpl.
      auto.
    }
    {
      simpl.
      assert (H2: Z.of_nat Nnat <= N). xomega.
      assert (H3: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      specialize (IHNnat H2).      

      destruct (zeq i (Z.of_nat Nnat)) as [Heq | Hneq]; auto.      
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tn Cn] eqn:Hconfn.

      assert (H6: 0 <= Z.of_nat Nnat < N). omega.
      assert (H7: 1 < Tn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }
      assert (H9: 0 < Cn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }
      
      
      destruct (is_overlapping (Z.of_nat Nnat) Tn i v ts te) eqn:Hoverlap.
      {
        rewrite is_overlapping_implies_remain with (d1:=v); auto; try omega. 
        replace (Cn + workload_overlapping_total Nnat i v ts te conf -
   (Cn - Z.min Cn (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf) ts - ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf) (ts / Tn * Tn)) +
    workload_remain_total Nnat i ts te conf (TAC_pool_at_t ts (Z.to_nat N) conf))) with
            (workload_overlapping_total Nnat i v ts te conf -
             workload_remain_total Nnat i ts te conf (TAC_pool_at_t ts (Z.to_nat N) conf) + 
   Z.min Cn (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf) ts - ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf) (ts / Tn * Tn))) by omega.
        rewrite IHNnat.
        auto.
      }
      {
        assert (Hremain: workload_remain i (Z.of_nat Nnat) Tn Cn ts te (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf)) = 0).
        {
          apply non_overlapping_no_remain_generalized with (v:=v); auto; try omega.
          - rewrite <- Hburst.
            apply start_of_burst_result_range; auto; omega.
        }
        omega.
      }
    }
  Qed.

  Lemma workload_overlapping_remain_total_le:
    forall Nnat i v ts te conf N,
      Z.of_nat Nnat <= N ->
      0 <= v ->
      v <= ts < te ->
      config_pool_valid N conf ->
      workload_overlapping_total Nnat i v ts te conf -
      workload_remain_total Nnat i ts te conf (TAC_pool_at_t ts (Z.to_nat N) conf) <=
      workload_overlapping_remain_total Nnat i v ts te ts conf (TAC_pool_at_t ts (Z.to_nat N) conf).
  Proof.
    intros * HN Hv Hts Hconfvalid.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      simpl.
      assert (H2: Z.of_nat Nnat <= N). xomega.
      assert (H3: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      specialize (IHNnat H2).      

      destruct (zeq i (Z.of_nat Nnat)) as [Heq | Hneq]; auto.      
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tn Cn] eqn:Hconfn.

      assert (H6: 0 <= Z.of_nat Nnat < N). omega.
      assert (H7: 1 < Tn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }
      assert (H9: 0 < Cn).
      { specialize (Hconfvalid _ _ _ H6 Hconfn). omega. }
      
      
      destruct (is_overlapping (Z.of_nat Nnat) Tn i v ts te) eqn:Hoverlap.
      {
        rewrite is_overlapping_implies_remain with (d1:=v); auto; try omega. 
        replace (Cn + workload_overlapping_total Nnat i v ts te conf -
   (Cn - Z.min Cn (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf) ts - ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf) (ts / Tn * Tn)) +
    workload_remain_total Nnat i ts te conf (TAC_pool_at_t ts (Z.to_nat N) conf))) with
            (workload_overlapping_total Nnat i v ts te conf -
             workload_remain_total Nnat i ts te conf (TAC_pool_at_t ts (Z.to_nat N) conf) + 
             Z.min Cn (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf) ts - ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf) (ts / Tn * Tn))) by omega.
        assert (H: forall a b c, a <= b -> a + c <= b + c). intros. omega.
        apply H.
        auto.
      }
      {
        assert (Hremain: 0 <= workload_remain i (Z.of_nat Nnat) Tn Cn ts te (ZMap.get (Z.of_nat Nnat) (TAC_pool_at_t ts (Z.to_nat N) conf))).
        {          
          apply workload_remain_nonneg.
        }
        omega.
      }
    }
  Qed.

  
End ESCHED_OVERLAP.


Section ESCHED_WHOLE.

  Lemma e_count_single_task_interference_value_in_one_period_aux:
    forall N t1 conf toid i T C nrange v ts te,
      0 <= v <= t1 ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      t1 mod T = 0 ->
      Z.of_nat nrange <= T ->
      t1 + T <= ts ->
      count_single_task_etype_occurrence N N i t1 (t1 + Z.of_nat nrange) conf toid v ts te ESched_whole =
      Z.min C (ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) (t1 + Z.of_nat nrange) -
               ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1).
  Proof.
    intros * Ht1 Hconf HT HC Hi Hmod Hrange Hts.
    unfold count_single_task_etype_occurrence.
    replace (t1 + Z.of_nat nrange - t1) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.    
    induction nrange.
    {
      simpl.
      replace (t1 + 0) with t1 by omega.
      rewrite Z.min_r; omega.
    }
    {
      assert (H0: Z.of_nat nrange < (Z.of_nat (S nrange))). xomega.
      assert (H1: t1 + Z.of_nat (S nrange) = (t1 + Z.of_nat nrange) + 1). xomega.
      rewrite H1. clear H1.
      simpl.

      assert (HNN: N = Z.of_nat (Z.to_nat N)). rewrite Z2Nat.id; omega.      
      assert (H2: (t1 + Z.of_nat nrange) / T * T = t1).
      {
        rewrite divide_plus_equal; try omega.
        rewrite mod_0_divide_mul; try omega.
      }      
      
      rewrite IHnrange; try xomega.
      destruct (zeq
         (highest_not_finished_map N (t1 + Z.of_nat nrange)
            (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (t1 + Z.of_nat nrange) conf)) i) as [Hsched|Hother].
      {
        rewrite Hsched.
        rewrite Hconf.

        assert (Htype: get_esched_type i T toid (t1 + Z.of_nat nrange) v ts te = ESched_whole).
        {
          unfold get_esched_type.
          rewrite H2.
          destruct (zle v t1); try omega.
          destruct (zlt t1 ts); try omega.
          rewrite zle_true; auto.
        }
        rewrite Htype.
        rewrite esched_type_eq_true.

        assert (H10: ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange + 1) (Z.to_nat N) conf) (t1 + Z.of_nat nrange + 1) = ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) (t1 + Z.of_nat nrange) + 1).
        {
          apply vt_increase_for_the_scheduled_task; auto; try omega.
          rewrite <- HNN.
          auto.
        }
        rewrite H10.
        
        assert (H11: ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) (t1 + Z.of_nat nrange) - ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1 < C).
        {
          assert (H12: vt_map_not_finished (t1 + Z.of_nat nrange) T C (ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf)) = true).
          {
            apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N) (conf:= conf); auto; try omega.
            rewrite <- HNN.
            auto.
          }
          unfold vt_map_not_finished in H12.
          rewrite divide_plus_equal in H12; try omega.
          rewrite mod_0_divide_mul in H12; try omega.

          assert (H14: ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) t1 = ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1).
          { rewrite vertical_map_canonical; omega. }
          rewrite H14 in H12.
          
          destruct (zlt
                      (ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) (t1 + Z.of_nat nrange) -
                       ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1) C); try omega; try discriminate. 
        }

        rewrite Z.min_r; try omega.
        rewrite Z.min_r; try omega.
      }
      {
        (* i not scheduled *)
        destruct (vt_map_not_finished (t1 + Z.of_nat nrange) T C (ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf))) eqn:Hfinish.
        {
          (* i is ready, vt stagnates *)
          assert (Hvt: ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange + 1) (Z.to_nat N) conf) (t1 + Z.of_nat nrange + 1) = ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) (t1 + Z.of_nat nrange)).
          {            
            apply vt_stagnates_if_task_not_scheduled with (T:=T) (C:=C); auto; try omega.
            - rewrite <- HNN.
              apply exists_ready_task_highest_not_finished_map_schedule_something with (id:=i) (T:=T) (C:=C); auto.
            - rewrite <- HNN.
              auto.
          }
          rewrite Hvt.
          auto.
        }
        {
          (* i finished *)
          assert (H10: 0 <= t1 + Z.of_nat nrange). omega.
          pose proof (monotonic_vt_step_bound _ _ _ conf H10 Hi) as H11.
          
          unfold vt_map_not_finished in Hfinish.
          rewrite divide_plus_equal in Hfinish; try omega.
          rewrite mod_0_divide_mul in Hfinish; try omega.
          assert (H12: ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) t1 = ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1).
          {
            rewrite vertical_map_canonical; auto; omega.
          }
          rewrite H12 in Hfinish.

          destruct (zlt
              (ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) (t1 + Z.of_nat nrange) -
               ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1) C); try discriminate.
          repeat rewrite Z.min_l; try omega.
        }
      }
    }
  Qed.

  Lemma e_count_single_task_interference_bounded_by_budget_in_one_period:
    forall N t1 conf toid i T C v ts te,
      0 <= v <= t1 ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      t1 mod T = 0 ->
      t1 + T <= ts ->
      count_single_task_etype_occurrence N N i t1 (t1 + T) conf toid v ts te ESched_whole <= C.
  Proof.
    intros.
    remember (Z.to_nat T) as nT.
    assert (HT: T = Z.of_nat nT).
    { rewrite HeqnT. rewrite Z2Nat.id; omega. }
    rewrite HT.
    
    rewrite e_count_single_task_interference_value_in_one_period_aux with (T:=T) (C:=C); auto; try omega.
    apply Z.le_min_l.
  Qed.

  Lemma e_count_single_task_interference_bounded_by_budget_total_aux1:
    forall N t1 n conf toid i T C v ts te,
      0 <= v <= t1 ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      t1 mod T = 0 ->
      t1 + T* Z.of_nat n <= ts ->
      count_single_task_etype_occurrence N N i t1 (t1 + T * Z.of_nat n) conf toid v ts te ESched_whole <= C * Z.of_nat n.
  Proof.
    intros * Ht1 Hconf HT HC Hi Hmod Hn.
    induction n.
    {
      simpl.
      replace (t1 + T * 0) with t1 by omega.
      rewrite count_single_task_etype_occurance_0_when_empty_duration.
      omega.
    }
    {
      assert (H1: Z.of_nat (S n) = Z.of_nat n + 1). xomega.
      assert (H2: t1 + T * Z.of_nat (S n) = t1 + T * Z.of_nat n + T). lia.
      rewrite H2 in *. clear H2.

      destruct (zeq (Z.of_nat n) 0) as [Heq | Hneq].
      {
        rewrite H1 in *.
        rewrite Heq in *.
        replace (t1 + T * 0) with t1 in * by omega.
        replace (C * (0 + 1)) with C by nia.
        apply e_count_single_task_interference_bounded_by_budget_in_one_period; auto.
      }
      {
        assert (H3: t1 <= t1 + T * Z.of_nat n <= t1 + T * Z.of_nat n + T). nia.
        pose proof (count_single_task_etype_occurrence_composition N N i _ _ _ conf toid v ts te ESched_whole H3) as H4.
        rewrite <- H4. clear H4.        
        replace (C * Z.of_nat (S n)) with (C * Z.of_nat n + C) by nia.

        assert (H5: t1 + T * Z.of_nat n <= ts). omega.
        specialize (IHn H5).

        assert (H: forall a b c d, a <= c -> b <= d -> a + b <= c + d). intros. omega.
        apply H; auto.

        apply e_count_single_task_interference_bounded_by_budget_in_one_period; auto; try omega.
        - rewrite Z.mul_comm.
          rewrite Z_mod_plus_full.
          auto.
      }
    }
  Qed.

  Lemma e_count_single_task_interference_bounded_by_budget_total_aux1_z:
    forall N t1 p conf toid i T C v ts te,
      0 <= p ->
      0 <= v <= t1 ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      t1 mod T = 0 ->
      t1 + T* p <= ts ->
      count_single_task_etype_occurrence N N i t1 (t1 + T * p) conf toid v ts te ESched_whole <= C * p.
  Proof.
    intros.
    remember (Z.to_nat p) as np.
    assert (Hp: p = Z.of_nat np).
    {
      rewrite Heqnp.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hp in *.
    eapply e_count_single_task_interference_bounded_by_budget_total_aux1; eauto.
  Qed.

  Lemma e_count_single_task_interference_0_aux2_aux:
    forall N t1 conf toid i T C v ts te nrange,
      0 <= t1 ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      t1 mod T = 0 ->
      Z.of_nat nrange <= T ->
      ts < t1 + T ->
      count_single_task_etype_occurrence N N i t1 (t1 + Z.of_nat nrange) conf toid v ts te ESched_whole = 0.
  Proof.
    intros * Ht1 Hconf HT Hmod Hrange Hts. 
    unfold count_single_task_etype_occurrence.
    replace (t1 + Z.of_nat nrange - t1) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHnrange; try xomega.
      destruct (zeq
         (highest_not_finished_map N (t1 + Z.of_nat nrange)
            (TAC_pool_at_t (t1 + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (t1 + Z.of_nat nrange) conf)) i); auto.
      rewrite e.
      
      rewrite Hconf.
      rewrite esched_type_eq_false; auto.
      
      unfold not.
      intros.
      unfold get_esched_type in H.
      rewrite divide_plus_equal in H; try xomega.
      rewrite mod_0_divide_mul in H; try omega.

      destruct (zle v t1); try congruence.
      destruct (zlt t1 ts); try congruence.
      rewrite zle_false in H; try omega.
      destruct (zeq i toid); try congruence.
      destruct (priEntry_lt (PriEntry (t1 + T) i) (PriEntry te toid)); try congruence.
    }
  Qed.
  
  Lemma e_count_single_task_interference_0_aux2:
    forall N t1 conf toid i T C v ts te,
      0 <= t1 <= ts ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      t1 mod T = 0 ->
      ts < t1 + T ->
      count_single_task_etype_occurrence N N i t1 ts conf toid v ts te ESched_whole = 0.
  Proof.
    intros.
    remember (Z.to_nat (ts - t1)) as nrange.
    assert (Hte: ts = t1 + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; try xomega.
    }
    rewrite Hte.

    apply e_count_single_task_interference_0_aux2_aux with (T:=T) (C:=C); auto; try xomega.
  Qed.

  Lemma e_count_single_task_interference_0_aux3_aux:
    forall N conf toid i T C v ts te nrange,
      0 <= v ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      v mod T <> 0 ->
      v + Z.of_nat nrange <= v / T * T + T ->
      count_single_task_etype_occurrence N N i v (v + Z.of_nat nrange) conf toid v ts te ESched_whole = 0.
  Proof.
    intros * Hv Hconf HT Hmod Hrange.
    assert (H1: v / T * T < v).
    {
      assert (v - v / T * T >= 1).
      {
        apply z_gt_mod; omega.
      }
      omega.
    }
    assert (H2: v < v / T * T + T).
    {
      apply x_lt_floor_plus; omega.
    }
      
    assert (H3: 0 <= v / T * T).
    {
      apply Z.mul_nonneg_nonneg; try omega.
      apply Z_div_pos; omega.
    }
      
    assert (H4: (v / T * T) mod T = 0).
    {
      apply Z_mod_mult.
    }
    
    unfold count_single_task_etype_occurrence.
    replace (v + Z.of_nat nrange - v) with (Z.of_nat nrange) by omega.
    rewrite Nat2Z.id.
    induction nrange.
    {
      simpl.
      auto.
    }
    {
      simpl.
      rewrite IHnrange; try xomega.
      destruct (zeq
         (highest_not_finished_map N (v + Z.of_nat nrange)
            (TAC_pool_at_t (v + Z.of_nat nrange) (Z.to_nat N) conf) conf
            (real_priqueue (Z.to_nat N) (v + Z.of_nat nrange) conf)) i) as [Hsched|Hother]; auto.
      rewrite Hsched.
      rewrite Hconf.

      unfold get_esched_type.
      assert (H5: (v + Z.of_nat nrange) / T * T = v / T * T).
      {
        apply divide_mul_plus_equal2; try xomega.
      }
      rewrite H5.
      destruct (zle v (v / T * T)); try omega.
      rewrite esched_type_eq_false; auto.
      congruence.
    }
  Qed.

  Lemma e_count_single_task_interference_0_aux3:
    forall N conf toid i T C v ts te,
      0 <= v ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      v mod T <> 0 ->
      count_single_task_etype_occurrence N N i v (v / T * T + T) conf toid v ts te ESched_whole = 0.
  Proof.
    intros * Hv Hconf HT Hmod.
    assert (H1: v / T * T < v).
    {
      assert (v - v / T * T >= 1).
      {
        apply z_gt_mod; omega.
      }
      omega.
    }
    assert (H2: v < v / T * T + T).
    {
      apply x_lt_floor_plus; omega.
    }
      
    assert (H3: 0 <= v / T * T).
    {
      apply Z.mul_nonneg_nonneg; try omega.
      apply Z_div_pos; omega.
    }
      
    assert (H4: (v / T * T) mod T = 0).
    {
      apply Z_mod_mult.
    }

    remember (Z.to_nat (v / T * T + T - v)) as nrange.
    assert (H5: v / T * T + T = v + Z.of_nat nrange).
    {
      rewrite Heqnrange.
      rewrite Z2Nat.id; omega.
    }
    rewrite H5.
    eapply e_count_single_task_interference_0_aux3_aux; eauto.
    xomega.
  Qed.
  
  Lemma e_count_single_task_interference_bounded_by_whole:
    forall N conf i T C toid v ts te,
      0 <= v ->
      v <= ts < te ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      count_single_task_etype_occurrence N N i v ts conf toid v ts te ESched_whole <= workload_extended_whole T C v ts.
  Proof.
    intros * Hv Hte Hconf HT HC Hi.
    unfold workload_extended_whole.
    
    destruct (zeq (v mod T) 0) as [Hvmod | Hvmod].
    {
      destruct (zlt (ts - v) T).
      {
        rewrite e_count_single_task_interference_0_aux2 with (T:=T) (C:=C); auto; try omega.
      }
      {
        assert (H1: v <= ts / T * T).
        {
          assert (ts < ts / T * T + T).
          { apply x_lt_floor_plus; omega. }
          omega.
        }
        assert (H2: ts / T * T <= ts).
        {
          apply divide_mul_le; omega.
        }
        assert (H3: (ts / T * T) mod T = 0).
        { apply Z_mod_mult. }
        assert (H: count_single_task_etype_occurrence N N i v (ts/T*T) conf toid v ts te ESched_whole + count_single_task_etype_occurrence N N i (ts/T*T) ts conf toid v ts te ESched_whole = count_single_task_etype_occurrence N N i v ts conf toid v ts te ESched_whole).
        {
          apply count_single_task_etype_occurrence_composition; try omega.          
        }
        rewrite <- H.

        assert (H10: ts < ts / T * T + T).
        {
          apply x_lt_floor_plus; omega.
        }
        rewrite e_count_single_task_interference_0_aux2 with (T:=T) (C:=C); auto; try omega.

        rewrite Z.add_0_r.

        remember ((ts / T * T - v) / T) as p.
        replace (p*C) with (C*p) by lia.
        assert (Hp: 0 <= p).
        {
          rewrite Heqp.
          apply Z_div_pos; omega.
        }
        assert (H12: (ts / T * T - v) mod T = 0).
        {
          rewrite Zminus_mod.
          rewrite H3.
          rewrite Hvmod.
          auto.
        }
        assert (H11: ts / T * T = v + T * p).
        {
          rewrite Heqp.
          symmetry.
          rewrite Z.mul_comm.
          rewrite mod_0_divide_mul; try omega.
          
        }
        rewrite H11.        
        
        apply e_count_single_task_interference_bounded_by_budget_total_aux1_z; auto; try omega.
      }
    }
    {
      assert (H1: v - v / T * T >= 1).
      {
        apply z_gt_mod; omega.
      }
      
      destruct (zlt ts (v / T * T + T)).
      {
        rewrite zlt_true; try omega.
        remember (Z.to_nat (ts - v)) as nrange.
        assert (H10: ts = v + Z.of_nat nrange).
        {
          rewrite Heqnrange.
          rewrite Z2Nat.id; omega.
        }
        rewrite H10 in *.

        rewrite e_count_single_task_interference_0_aux3_aux with (T:=T) (C:=C); auto; try omega.
      }
      {
        assert (H2: v < v / T * T + T).
        {
          apply x_lt_floor_plus; omega.
        }
        assert (Hcom1: count_single_task_etype_occurrence N N i v (v/T*T+T) conf toid v ts te ESched_whole + count_single_task_etype_occurrence N N i (v/T*T+T) ts conf toid v ts te ESched_whole = count_single_task_etype_occurrence N N i v ts conf toid v ts te ESched_whole).
        {
          apply count_single_task_etype_occurrence_composition; try omega.          
        }

        rewrite <- Hcom1.
        rewrite e_count_single_task_interference_0_aux3 with (C:=C); auto.
        simpl.

        assert (H3: (v / T * T + T) mod T = 0).
        {
          replace (v / T * T + T) with ((v / T + 1) * T) by lia.
          apply Z_mod_mult.
        }

        destruct (zlt ts (v/T*T+T+T)).
        {
          rewrite e_count_single_task_interference_0_aux2 with (T:=T) (C:=C); auto; try omega.
          destruct (zlt (ts - v) T); try omega.
          apply div_mul_nonneg; try omega.

          assert (H10: (v / T * T + T) / T * T <= ts / T * T).
          {
            apply div_mul_le; omega.
          }
          rewrite mod_0_divide_mul in H10; try omega.
        }
        {
          rewrite zlt_false; try omega.
          assert (H4: ts / T * T <= ts).
          {
            apply divide_mul_le; omega.
          }
          assert (H6: (v/T*T+T+T) mod T = 0).
          {
            apply mod_0_preserve; omega.
          }
          assert (H5: (v / T * T + T + T) / T * T <= ts / T * T).
          {
            apply div_mul_le; omega.
          }
          rewrite mod_0_divide_mul in H5; try omega.
          
          assert (Hcom2: count_single_task_etype_occurrence N N i (v/T*T+T) (ts/T*T) conf toid v ts te ESched_whole + count_single_task_etype_occurrence N N i (ts/T*T) ts conf toid v ts te ESched_whole = count_single_task_etype_occurrence N N i (v/T*T+T) ts conf toid v ts te ESched_whole).
          {
            apply count_single_task_etype_occurrence_composition; try omega.          
          }
          rewrite <- Hcom2.

          assert (H7: (ts / T * T) mod T = 0). apply Z_mod_mult.
          assert (H8: ts < ts / T * T + T).
          { apply x_lt_floor_plus; omega. }            
            
          rewrite e_count_single_task_interference_0_aux2 with (T:=T) (C:=C); auto; try omega.
          rewrite Z.add_0_r.
          
          remember ((ts / T * T - (v / T * T + T)) / T) as p.
          assert (Hp: 0 <= p).
          {
            rewrite Heqp.
            apply Z_div_pos; omega.
          }

          assert (H9: ts / T * T = v / T * T + T + T * p).
          {
            rewrite Heqp.
            replace (T * ((ts / T * T - (v / T * T + T)) / T)) with (ts / T * T - (v / T * T + T)).
            omega.
            - symmetry.
              rewrite Z.mul_comm.
              rewrite mod_0_divide_mul; try omega.
              replace (ts / T * T - (v / T * T + T)) with ((ts / T - v / T - 1) * T) by nia.
              apply Z_mod_mult.
          }
          rewrite H9.

          replace (p*C) with (C*p) by lia.
          apply e_count_single_task_interference_bounded_by_budget_total_aux1_z with (C:=C); auto; try omega.
        }
      }
    }
  Qed.

  Lemma e_count_occurrence_whole_bounded_aux:
    forall Nnat N0 id v ts te conf,
      0 <= v ->
      v <= ts < te ->
      Z.of_nat Nnat <= N0 ->
      config_pool_valid N0 conf ->
      count_etype_occurrence2_aux Nnat N0 N0 v ts conf id v ts te ESched_whole <=
      workload_extended_whole_total Nnat conf v ts.
  Proof.
    intros * Hv Hts HNbound Hconfvalid.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (H2: Z.of_nat Nnat <= N0). omega.
      specialize (IHNnat H2).

      simpl.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [T C] eqn:Hconf.
      
      assert (H: forall a b c d, a <= c -> b <= d -> a + b <= c + d). intros. omega.
      apply H; auto.

      assert (H3: 0 <= Z.of_nat Nnat < N0). omega.
      specialize (Hconfvalid _ _ _ H3 Hconf).
      apply e_count_single_task_interference_bounded_by_whole; auto; omega.
    }
  Qed.
  
End ESCHED_WHOLE.



Section ESCHED_WHOLE_SCHEDULABLE.

  Lemma e_count_single_task_interference_is_budget_in_one_period_if_schedulable:
    forall N t1 conf toid i T C v ts te,
      0 <= v <= t1 ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      t1 mod T = 0 ->
      t1 + T <= ts ->
      system_schedulable N conf ->
      count_single_task_etype_occurrence N N i t1 (t1 + T) conf toid v ts te ESched_whole = C.
  Proof.
    intros.
    remember (Z.to_nat T) as nT.
    assert (HT: T = Z.of_nat nT).
    { rewrite HeqnT. rewrite Z2Nat.id; omega. }
    rewrite HT.
    
    rewrite e_count_single_task_interference_value_in_one_period_aux with (T:=T) (C:=C); auto; try omega.
    assert (Hsched: C <= ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nT) (Z.to_nat N) conf) (t1 + Z.of_nat nT) -
                    ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1).
    {
      unfold system_schedulable in H6.
      unfold schedulable_upto in H6.
      unfold schedulable_per_level_upto in H6.

      assert (H10: ZMap.get i (TAC_pool_at_t (t1 + Z.of_nat nT) (Z.to_nat N) conf) t1 = ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1).
      {
        rewrite vertical_map_canonical; omega.
      }
      rewrite <- H10.

      remember (t1 / Z.of_nat nT) as k.
      assert (H11: t1 = k * Z.of_nat nT).
      {
        rewrite Heqk.
        rewrite HT in H4.
        rewrite mod_0_divide_mul; try omega.
      }
      rewrite H11.

      assert (H12: 0 <= k * Z.of_nat nT + Z.of_nat nT). xomega.
      assert (H14: 0 <= k).
      {
        rewrite Heqk.
        apply Z_div_pos; omega.
      }
      rewrite HT in *.
      specialize (H6 _ H12 _ H3).
      rewrite H0 in H6.
      apply H6; try omega.
    }

    rewrite Z.min_l; omega.
  Qed.
  
  Lemma e_count_single_task_interference_is_budget_total_if_schedulable_aux1:
    forall N t1 n conf toid i T C v ts te,
      0 <= v <= t1 ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      t1 mod T = 0 ->
      t1 + T* Z.of_nat n <= ts ->
      system_schedulable N conf ->
      count_single_task_etype_occurrence N N i t1 (t1 + T * Z.of_nat n) conf toid v ts te ESched_whole = C * Z.of_nat n.
  Proof.
    intros * Ht1 Hconf HT HC Hi Hmod Hn Hschedulable.
    induction n.
    {
      simpl.
      replace (t1 + T * 0) with t1 by omega.
      rewrite count_single_task_etype_occurance_0_when_empty_duration.
      omega.
    }
    {
      assert (H1: Z.of_nat (S n) = Z.of_nat n + 1). xomega.
      assert (H2: t1 + T * Z.of_nat (S n) = t1 + T * Z.of_nat n + T). lia.
      rewrite H2 in *. clear H2.

      destruct (zeq (Z.of_nat n) 0) as [Heq | Hneq].
      {
        rewrite H1 in *.
        rewrite Heq in *.
        replace (t1 + T * 0) with t1 in * by omega.
        replace (C * (0 + 1)) with C by nia.
        apply e_count_single_task_interference_is_budget_in_one_period_if_schedulable; auto.
      }
      {
        assert (H3: t1 <= t1 + T * Z.of_nat n <= t1 + T * Z.of_nat n + T). nia.
        pose proof (count_single_task_etype_occurrence_composition N N i _ _ _ conf toid v ts te ESched_whole H3) as H4.
        rewrite <- H4. clear H4.        
        replace (C * Z.of_nat (S n)) with (C * Z.of_nat n + C) by nia.

        assert (H5: t1 + T * Z.of_nat n <= ts). omega.
        specialize (IHn H5).

        assert (H: forall a b c d, a = c -> b = d -> a + b = c + d). intros. omega.
        apply H; auto.

        apply e_count_single_task_interference_is_budget_in_one_period_if_schedulable; auto; try omega.
        - rewrite Z.mul_comm.
          rewrite Z_mod_plus_full.
          auto.
      }
    }
  Qed.

  Lemma e_count_single_task_interference_is_budget_total_if_schedulable_aux1_z:
    forall N t1 p conf toid i T C v ts te,
      0 <= p ->
      0 <= v <= t1 ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      t1 mod T = 0 ->
      t1 + T* p <= ts ->
      system_schedulable N conf ->
      count_single_task_etype_occurrence N N i t1 (t1 + T * p) conf toid v ts te ESched_whole = C * p.
  Proof.
    intros.
    remember (Z.to_nat p) as np.
    assert (Hp: p = Z.of_nat np).
    {
      rewrite Heqnp.
      rewrite Z2Nat.id; omega.
    }
    rewrite Hp in *.
    eapply e_count_single_task_interference_is_budget_total_if_schedulable_aux1; eauto.
  Qed.

  Lemma e_count_single_task_interference_is_whole_if_schedulable:
    forall N conf i T C toid v ts te,
      0 <= v ->
      v <= ts < te ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      0 < T ->
      0 < C ->
      0 <= i < N ->
      system_schedulable N conf ->
      count_single_task_etype_occurrence N N i v ts conf toid v ts te ESched_whole = workload_extended_whole T C v ts.
  Proof.
    intros * Hv Hte Hconf HT HC Hi Hschedulable.
    unfold workload_extended_whole.
    
    destruct (zeq (v mod T) 0) as [Hvmod | Hvmod].
    {
      destruct (zlt (ts - v) T).
      {
        rewrite e_count_single_task_interference_0_aux2 with (T:=T) (C:=C); auto; try omega.
      }
      {
        assert (H1: v <= ts / T * T).
        {
          assert (ts < ts / T * T + T).
          { apply x_lt_floor_plus; omega. }
          omega.
        }
        assert (H2: ts / T * T <= ts).
        {
          apply divide_mul_le; omega.
        }
        assert (H3: (ts / T * T) mod T = 0).
        { apply Z_mod_mult. }
        assert (H: count_single_task_etype_occurrence N N i v (ts/T*T) conf toid v ts te ESched_whole + count_single_task_etype_occurrence N N i (ts/T*T) ts conf toid v ts te ESched_whole = count_single_task_etype_occurrence N N i v ts conf toid v ts te ESched_whole).
        {
          apply count_single_task_etype_occurrence_composition; try omega.          
        }
        rewrite <- H.

        assert (H10: ts < ts / T * T + T).
        {
          apply x_lt_floor_plus; omega.
        }
        rewrite e_count_single_task_interference_0_aux2 with (T:=T) (C:=C); auto; try omega.

        rewrite Z.add_0_r.

        remember ((ts / T * T - v) / T) as p.
        replace (p*C) with (C*p) by lia.
        assert (Hp: 0 <= p).
        {
          rewrite Heqp.
          apply Z_div_pos; omega.
        }
        assert (H12: (ts / T * T - v) mod T = 0).
        {
          rewrite Zminus_mod.
          rewrite H3.
          rewrite Hvmod.
          auto.
        }
        assert (H11: ts / T * T = v + T * p).
        {
          rewrite Heqp.
          symmetry.
          rewrite Z.mul_comm.
          rewrite mod_0_divide_mul; try omega.
          
        }
        rewrite H11.        
        
        apply e_count_single_task_interference_is_budget_total_if_schedulable_aux1_z; auto; try omega.
      }
    }
    {
      assert (H1: v - v / T * T >= 1).
      {
        apply z_gt_mod; omega.
      }
      
      destruct (zlt ts (v / T * T + T)).
      {
        rewrite zlt_true; try omega.
        remember (Z.to_nat (ts - v)) as nrange.
        assert (H10: ts = v + Z.of_nat nrange).
        {
          rewrite Heqnrange.
          rewrite Z2Nat.id; omega.
        }
        rewrite H10 in *.

        rewrite e_count_single_task_interference_0_aux3_aux with (T:=T) (C:=C); auto; try omega.
      }
      {
        assert (H2: v < v / T * T + T).
        {
          apply x_lt_floor_plus; omega.
        }
        assert (Hcom1: count_single_task_etype_occurrence N N i v (v/T*T+T) conf toid v ts te ESched_whole + count_single_task_etype_occurrence N N i (v/T*T+T) ts conf toid v ts te ESched_whole = count_single_task_etype_occurrence N N i v ts conf toid v ts te ESched_whole).
        {
          apply count_single_task_etype_occurrence_composition; try omega.          
        }

        rewrite <- Hcom1.
        rewrite e_count_single_task_interference_0_aux3 with (C:=C); auto.
        simpl.

        assert (H3: (v / T * T + T) mod T = 0).
        {
          replace (v / T * T + T) with ((v / T + 1) * T) by lia.
          apply Z_mod_mult.
        }

        destruct (zlt ts (v/T*T+T+T)).
        {
          rewrite e_count_single_task_interference_0_aux2 with (T:=T) (C:=C); auto; try omega.
          destruct (zlt (ts - v) T); try omega.

          assert (H10: ts / T * T = v / T * T + T).
          {
            apply divide_mul_plus_equal2; try omega.
          }
          rewrite H10.
          replace (v / T * T + T - (v / T * T + T)) with 0 by omega.
          auto.
        }
        {
          rewrite zlt_false; try omega.
          assert (H4: ts / T * T <= ts).
          {
            apply divide_mul_le; omega.
          }
          assert (H6: (v/T*T+T+T) mod T = 0).
          {
            apply mod_0_preserve; omega.
          }
          assert (H5: (v / T * T + T + T) / T * T <= ts / T * T).
          {
            apply div_mul_le; omega.
          }
          rewrite mod_0_divide_mul in H5; try omega.
          
          assert (Hcom2: count_single_task_etype_occurrence N N i (v/T*T+T) (ts/T*T) conf toid v ts te ESched_whole + count_single_task_etype_occurrence N N i (ts/T*T) ts conf toid v ts te ESched_whole = count_single_task_etype_occurrence N N i (v/T*T+T) ts conf toid v ts te ESched_whole).
          {
            apply count_single_task_etype_occurrence_composition; try omega.          
          }
          rewrite <- Hcom2.

          assert (H7: (ts / T * T) mod T = 0). apply Z_mod_mult.
          assert (H8: ts < ts / T * T + T).
          { apply x_lt_floor_plus; omega. }            
            
          rewrite e_count_single_task_interference_0_aux2 with (T:=T) (C:=C); auto; try omega.
          rewrite Z.add_0_r.
          
          remember ((ts / T * T - (v / T * T + T)) / T) as p.
          assert (Hp: 0 <= p).
          {
            rewrite Heqp.
            apply Z_div_pos; omega.
          }

          assert (H9: ts / T * T = v / T * T + T + T * p).
          {
            rewrite Heqp.
            replace (T * ((ts / T * T - (v / T * T + T)) / T)) with (ts / T * T - (v / T * T + T)).
            omega.
            - symmetry.
              rewrite Z.mul_comm.
              rewrite mod_0_divide_mul; try omega.
              replace (ts / T * T - (v / T * T + T)) with ((ts / T - v / T - 1) * T) by nia.
              apply Z_mod_mult.
          }
          rewrite H9.

          replace (p*C) with (C*p) by lia.
          apply e_count_single_task_interference_is_budget_total_if_schedulable_aux1_z with (C:=C); auto; try omega.
        }
      }
    }
  Qed.
  
  Lemma e_count_occurrence_is_whole_if_schedulable_aux:
    forall Nnat N0 id v ts te conf,
      0 <= v ->
      v <= ts < te ->
      Z.of_nat Nnat <= N0 ->
      config_pool_valid N0 conf ->
      system_schedulable N0 conf ->
      count_etype_occurrence2_aux Nnat N0 N0 v ts conf id v ts te ESched_whole =
      workload_extended_whole_total Nnat conf v ts.
  Proof.
    intros * Hv Hts HNbound Hconfvalid Hschedulable.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      assert (H1: Z.of_nat Nnat < Z.of_nat (S Nnat)). xomega.
      assert (H2: Z.of_nat Nnat <= N0). omega.
      specialize (IHNnat H2).

      simpl.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [T C] eqn:Hconf.
      
      assert (H: forall a b c d, a = c -> b = d -> a + b = c + d). intros. omega.
      apply H; auto.

      assert (H3: 0 <= Z.of_nat Nnat < N0). omega.
      specialize (Hconfvalid _ _ _ H3 Hconf).
      apply e_count_single_task_interference_is_whole_if_schedulable; auto; omega.
    }
  Qed.
  
End ESCHED_WHOLE_SCHEDULABLE.

  
Section BUSY_SCHEDULE.

  Theorem workloads_in_a_busy_schedule_one_burst:
    forall i T C N conf v ts,
      0 <= i < N ->
      0 <= ts ->
      ts mod T = 0 ->
      config_pool_valid N conf ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      start_of_burst N conf ts i (ts + T) = v ->
      workload_overlapping_total (Z.to_nat N) i v ts (ts+T) conf -
      workload_remain_total (Z.to_nat N) i ts (ts+T) conf (TAC_pool_at_t (ts+T) (Z.to_nat N) conf) >= ts - v - workload_extended_whole_total (Z.to_nat N) conf v ts.
  Proof.
    intros * Hi Hts Htsmod Hconfvalid Hconfi Hstart.

    assert (HT: 1 < T).
    { specialize (Hconfvalid _ _ _ Hi Hconfi). omega. }
    assert (Hv: 0 <= v <= ts).
    {
      rewrite <- Hstart.
      apply start_of_burst_result_range; omega.
    }
    
    assert (Hcount: count_etype_occurrence N N v ts conf i v ts (ts+T) ESched_whole +
      count_etype_occurrence N N v ts conf i v ts (ts+T) ESched_overlap = ts - v).
    {
      eapply sched_etype_is_complete_when_schedule_is_busy; eauto; try omega.
    }
    
    assert (Hoverlap: workload_overlapping_total (Z.to_nat N) i v ts (ts + T) conf -
                      workload_remain_total (Z.to_nat N) i ts (ts + T) conf (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
                      = count_etype_occurrence N N v ts conf i v ts (ts + T) ESched_overlap).
    {
      remember (Z.to_nat N) as Nnat.
      assert (HNN: N = Z.of_nat Nnat). rewrite HeqNnat. rewrite Z2Nat.id; omega.
      rewrite HNN in *.
      
      rewrite count_e_occurrence_overlapping_bounded; auto; try omega.
      rewrite HeqNnat.
      rewrite <- HNN.

      assert (Hremain: workload_remain_total (Z.to_nat N) i ts (ts + T) conf
                                             (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) =
             workload_remain_total (Z.to_nat N) i ts (ts + T) conf
                                   (TAC_pool_at_t ts (Z.to_nat N) conf)).
      {
        rewrite workload_remain_total_canonical; auto; try omega.
        rewrite HNN.
        rewrite Nat2Z.id.
        auto.
      }
      rewrite Hremain.

      assert (Hoverlap: workload_overlapping_remain_total (Z.to_nat N) i v ts 
                                                          (ts + T) ts conf (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) =
             workload_overlapping_remain_total (Z.to_nat N) i v ts 
                                               (ts + T) ts conf (TAC_pool_at_t ts (Z.to_nat N) conf)).
      {
        apply workload_overlapping_remain_total_canonical; auto; try omega.
        rewrite HNN.
        rewrite Nat2Z.id.
        auto.
      }
      rewrite Hoverlap.
      
      apply workload_overlapping_remain_total_eq_burst; auto; try omega.
      - rewrite Z2Nat.id; omega.
      - rewrite HNN. auto.
      - rewrite HNN. auto.
    }

    assert (Hwhole: count_etype_occurrence N N v ts conf i v ts (ts + T) ESched_whole <= workload_extended_whole_total (Z.to_nat N) conf v ts).
    {
      rewrite count_etype_occurrence2_eq; try omega.
      apply e_count_occurrence_whole_bounded_aux; auto; try omega.
      - rewrite Z2Nat.id; omega.
    }
    omega.      
  Qed.

End BUSY_SCHEDULE.


Section SCHEDULABLE_WORKLOAD.

  Theorem all_components_fit_in_one_burst:
    forall i T C N conf v ts,
      0 < N ->
      0 <= v <= ts ->
      0 <= i < N ->
      ts mod T = 0 ->
      config_pool_valid N conf ->
      system_schedulable N conf ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      workload_extended_whole_total (Z.to_nat N) conf v ts + 
      workload_overlapping_total (Z.to_nat N) i v ts (ts+T) conf -       
      workload_remain_total (Z.to_nat N) i ts (ts+T) conf (TAC_pool_at_t (ts+T) (Z.to_nat N) conf) <= ts - v.
  Proof.
    intros * HN Hts Hi Htsmod Hconfvalid Hschedulable Hconfi.
    assert (Hcount: count_etype_occurrence N N v ts conf i v ts (ts+T) ESched_whole +
                    count_etype_occurrence N N v ts conf i v ts (ts+T) ESched_overlap <= ts - v).
    {
      assert (H1: count_etype_occurrence N N v ts conf i v ts (ts+T) ESched_whole +
                  count_etype_occurrence N N v ts conf i v ts (ts+T) ESched_overlap +
                  count_etype_occurrence N N v ts conf i v ts (ts+T) ESched_other <=
                  Z.of_nat (Z.to_nat (ts - v))).
      { apply sched_etype_sum_aux_bounded_by_duration; auto. }

      rewrite Z2Nat.id in H1; try omega.

      assert (H2: 0 <= count_etype_occurrence N N v ts conf i v ts (ts + T) ESched_other).
      { apply count_etype_occurrence_aux_nonneg; auto. }
      omega.
    }

    assert (HT: 1 < T).
    { specialize (Hconfvalid _ _ _ Hi Hconfi). omega. }
    
    assert (Hoverlap: workload_overlapping_total (Z.to_nat N) i v ts (ts + T) conf -
                      workload_remain_total (Z.to_nat N) i ts (ts + T) conf (TAC_pool_at_t (ts + T) (Z.to_nat N) conf)
                      <= count_etype_occurrence N N v ts conf i v ts (ts + T) ESched_overlap).
    {
      remember (Z.to_nat N) as Nnat.
      assert (HNN: N = Z.of_nat Nnat). rewrite HeqNnat. rewrite Z2Nat.id; omega.
      rewrite HNN in *.
      
      rewrite count_e_occurrence_overlapping_bounded; auto; try omega.
      rewrite HeqNnat.
      rewrite <- HNN.
      
      assert (Hremain: workload_remain_total (Z.to_nat N) i ts (ts + T) conf
                                             (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) =
             workload_remain_total (Z.to_nat N) i ts (ts + T) conf
                                   (TAC_pool_at_t ts (Z.to_nat N) conf)).
      {
        rewrite workload_remain_total_canonical; auto; try omega.
        rewrite HNN.
        rewrite Nat2Z.id.
        auto.
      }
      rewrite Hremain.

      assert (Hoverlap: workload_overlapping_remain_total (Z.to_nat N) i v ts 
                                                          (ts + T) ts conf (TAC_pool_at_t (ts + T) (Z.to_nat N) conf) =
             workload_overlapping_remain_total (Z.to_nat N) i v ts 
                                               (ts + T) ts conf (TAC_pool_at_t ts (Z.to_nat N) conf)).
      {
        apply workload_overlapping_remain_total_canonical; auto; try omega.
        rewrite HNN.
        rewrite Nat2Z.id.
        auto.
      }
      rewrite Hoverlap.
      apply workload_overlapping_remain_total_le; auto; try omega.
      - rewrite Z2Nat.id; omega.
      - rewrite HNN. auto.
    }

    assert (Hwhole: count_etype_occurrence N N v ts conf i v ts (ts + T) ESched_whole = workload_extended_whole_total (Z.to_nat N) conf v ts).
    {
      rewrite count_etype_occurrence2_eq; try omega.
      apply e_count_occurrence_is_whole_if_schedulable_aux; auto; try omega.      
      - rewrite Z2Nat.id; omega.
    }
    omega.    
  Qed.

End SCHEDULABLE_WORKLOAD.    


Section A_TASK_IS_ENLARGED.

  Lemma workload_whole_divide:
    forall i id T C d1 d2 d3,
      1 < T ->
      0 <= d1 ->
      d1 <= d2 < d3 ->
      d2 mod T = 0 ->
      workload_extended_whole T C d1 d2 + workload_whole i id T C d2 d3 =
      workload_whole i id T C d1 d3.
  Proof.
    intros * HT Hd1 Hd Hd2mod.

    destruct (zeq d1 d2) as [Heq | Hneq].
    {
      rewrite Heq in *.
      rewrite workload_extended_whole_empty_slot; try omega.
    }
    {      
      unfold workload_whole, workload_extended_whole.
      rewrite mod_0_divide_mul; try omega.
      rewrite Hd2mod.
      assert (H0: (if zeq 0 0 then d2 else d2 + T) = d2).
      { rewrite zeq_true; auto. }
      rewrite H0.
      
      destruct (zlt (d2 - d1) T).
      {
        assert (H1: d1 mod T <> 0).
        {
          apply within_boundary_mod_not_0 with (c:=d2); omega.
        }
        assert (H2: d1 / T * T + T = d2).
        {
          apply div_mul_round_up; omega.
        }
        rewrite H2.
        assert (H3: (if zeq (d1 mod T) 0 then d1 else d2) = d2).
        { rewrite zeq_false; auto. }
        rewrite H3.
          
        destruct (zlt (d3 - d2) T).
        {
          assert (H11: d3 mod T <> 0).
          {
            replace d3 with (d2 + (d3 - d2)) by omega.
            apply mod_0_not_0; omega.
          }
          destruct (zlt (d3 - d1) T); try omega.
          rewrite zeq_false; auto.

          assert (H12: d3 / T * T = d2).
          { apply divide_mul_plus_equal2; omega. }
          rewrite H12.          
          simpl.

          replace (d2 - d2) with 0 by omega.
          simpl.
          auto.
        }
        {
          destruct (zlt (d3 - d1) T); try omega.        
        }
      }
      {
        assert (H1: 0 <= (d2 - (if zeq (d1 mod T) 0 then d1 else d1 / T * T + T))).
        {
          destruct (zeq (d1 mod T) 0) as [Hd1mod | Hd1mod]; try omega.
          - assert (d1 + T <> d2).
            {
              unfold not.
              intros H.              
              rewrite <- H in *.
              assert ((d1 + T) mod T <> 0). apply mod_not_0_preserve; omega.
              omega.
            }
            assert (d1 / T * T + T <= d2 - T).
            { apply mod_0_bound_increase; try omega. }
            omega.
        }
        
        destruct (zlt (d3 - d2) T).
        {
          destruct (zlt (d3 - d1) T); try omega.

          assert (H11: d3 mod T <> 0).
          {
            replace d3 with (d2 + (d3 - d2)) by omega.
            apply mod_0_not_0; omega.
          }          

          assert (H12: (if zeq (d3 mod T) 0 then if zlt id i then d3 / T * T else d3 - T else d3 / T * T) = d2).
          {
            rewrite zeq_false; auto.
            assert (H: d3 / T * T = d2).
            { apply divide_mul_plus_equal2; omega. }
            rewrite H.          
            auto.
          }
          rewrite H12.
          omega.
        }
        {
          destruct (zlt (d3 - d1) T); try omega.
          rewrite mod_0_divide_mul_assoc; try omega.
          - replace (d2 - (if zeq (d1 mod T) 0 then d1 else d1 / T * T + T) +
    ((if zeq (d3 mod T) 0 then if zlt id i then d3 / T * T else d3 - T else d3 / T * T) - d2)) with ((if zeq (d3 mod T) 0 then if zlt id i then d3 / T * T else d3 - T else d3 / T * T) -
                                                                                                     (if zeq (d1 mod T) 0 then d1 else d1 / T * T + T)) by omega.
            auto.
          - rewrite Zminus_mod.
            rewrite Hd2mod.
            destruct (zeq (d1 mod T) 0) as [Hd1mod | Hd1mod].
            + rewrite Hd1mod.
              simpl.
              auto.
            + replace (d1 / T * T + T) with ((d1 / T + 1) * T) by nia.
              rewrite Z_mod_mult.
              simpl.
              auto.
          - rewrite Zminus_mod.
            rewrite Hd2mod.
            destruct (zeq (d3 mod T) 0) as [Hd3mod | Hd3mod].
            + destruct (zlt id i).
              * rewrite Z_mod_mult.
                simpl.
                auto.
              * replace ((d3 - T) mod T - 0) with ((d3 - T) mod T) by omega.
                rewrite Zminus_mod.
                rewrite Hd3mod.
                rewrite Z_mod_same_full.
                simpl.
                auto.
            + replace ((d3 / T * T) mod T - 0) with ((d3 / T * T) mod T) by omega.
              rewrite Z_mod_mult.
              simpl.
              auto.
        }
      }
    }
  Qed.
  
  Lemma workload_whole_decompose:
    forall id T C i v ts te,
      0 <= v ->
      v <= ts < te ->
      1 < T ->
      (if is_overlapping id T i v ts te then C else 0) + workload_extended_whole T C v ts + workload_whole i id T C ts te = workload_whole i id T C v te.
  Proof.
    intros * Hv Ht HT.
    unfold is_overlapping.
    destruct (zeq (ts mod T) 0) as [Hmodts | Hmodts].
    {
      rewrite mod_0_divide_mul; try omega.
      destruct (zle v ts); try omega.      
      destruct (zlt ts ts); try omega.
      simpl.
      rewrite workload_whole_divide; try omega.
    }
    {
      assert (H1: ts - ts / T * T >= 1).
      { apply z_gt_mod; omega. }
      assert (H2: 0 <= ts / T * T).
      { apply div_mul_nonneg; omega. }        
      assert (H5: (ts / T * T) mod T = 0).
      { apply Z_mod_mult. }
      assert (H3: (ts /T * T + T) mod T = 0).
      { apply mod_0_preserve; try omega. }
      assert (H4: ts < ts / T * T + T).
      { apply x_lt_floor_plus; omega. }
      assert (H6: (ts / T * T + T) / T * T = ts / T * T + T).
      { rewrite mod_0_divide_mul; try omega. }
      
      destruct (zlt (ts / T * T + T) te).
      {
        rewrite workload_whole_left_align; try omega.
        
        assert (H10: workload_whole i id T C v te = workload_extended_whole T C v (ts / T * T + T) + workload_whole i id T C (ts / T * T + T) te).
        {
          rewrite <- workload_whole_divide with (d2 := ts / T * T + T); auto; try omega.
        }          
        rewrite H10. clear H10.

        rewrite workload_extended_whole_right_align; try omega.
        destruct (zle v (ts / T * T)).
        {          
          destruct (zlt (ts / T * T) ts); try omega.
          unfold priEntry_lt.
          rewrite zlt_true; try omega.
          rewrite workload_extended_whole_right_step; try omega.
        }
        {
          unfold workload_extended_whole.
          rewrite zlt_true; try omega.
          rewrite zlt_true; try omega.          
        }
      }
      {
        assert (H10: workload_whole i id T C ts te = 0).
        {
          unfold workload_whole.
          rewrite zlt_true; try omega.
        }
        rewrite H10.

        destruct (zeq te (ts / T * T + T)) as [Heq | Hneq].
        {
          rewrite Heq in *.
          rewrite workload_extended_whole_right_align; try omega.          

          destruct (zle v (ts / T * T)).
          {
            destruct (zlt (ts / T * T) ts); try omega.
            rewrite <- workload_whole_divide with (d2:=ts/T*T); try omega.
            unfold workload_whole.
            rewrite zlt_false; try omega.
            
            rewrite H3.
            rewrite zeq_true.
            unfold priEntry_lt.
            rewrite zlt_false; try omega.
            rewrite zeq_true.
            rewrite H5.
            rewrite zeq_true.
            rewrite H6.
            
            destruct (zlt id i).
            {
              replace (ts / T * T + T - ts / T * T) with T by omega.
              rewrite Z_div_same_full; try omega.
            }
            {
              replace (ts / T * T + T - T - ts / T * T) with 0 by omega.
              rewrite Zdiv_0_l.
              omega.
            }
          }
          {
            unfold workload_whole.
            rewrite zlt_true; try omega.
            unfold workload_extended_whole.
            rewrite zlt_true; try omega.            
          }
        }
        {
          assert (Hoverlap: (if if zle v (ts / T * T)
       then
        if zlt (ts / T * T) ts
        then
         if priEntry_lt (PriEntry (ts / T * T + T) id) (PriEntry te i)
         then true
         else false
        else false
       else false
    then C
    else 0) = 0).
          {
            destruct (zle v (ts / T * T)); try omega.
            destruct (zlt (ts / T * T) ts); try omega.
            unfold priEntry_lt.
            rewrite zlt_false; try omega.
            rewrite zeq_false; try omega.
          }
          rewrite Hoverlap.

          assert (H11: te mod T <> 0).
          { apply within_boundary_mod_not_0 with (c:=ts/T* T+T); try omega. }
          
          destruct (zle v (ts / T *T)).
          {
            rewrite <- workload_whole_divide with (d2:=ts/T*T); try omega.
            rewrite workload_extended_whole_right_align; try omega.
            unfold workload_whole.
            rewrite zlt_true; try omega.
          }
          {
            unfold workload_whole.
            rewrite zlt_true; try omega.
            unfold workload_extended_whole.
            rewrite zlt_true; try omega.
          }
        }
      }
    }
  Qed.

  Lemma workload_extended_whole_enlarge_nonincr:
    forall k T C t1 t2,
      0 < k ->
      0 < T ->
      0 < C ->
      0 <= t1 <= t2 ->
      workload_extended_whole (k * T) (k * C) t1 t2 <= workload_extended_whole T C t1 t2.
  Proof.
    intros * Hk HT HC Ht.
    destruct (zeq k 1) as [Heq|Hneq].
    {
      rewrite Heq in *.
      replace (1*T) with T by omega.
      replace (1*C) with C by omega.
      omega.
    }
    {
      destruct (zlt (t2 - t1) (k*T)).
      {
        assert (H1: workload_extended_whole (k * T) (k * C) t1 t2 = 0).
        {
          unfold workload_extended_whole.
          rewrite zlt_true; auto.
        }
        rewrite H1.
        apply workload_extended_whole_nonneg; omega.
      }
      {
        unfold workload_extended_whole.
        rewrite zlt_false; auto.
        rewrite zlt_false; try nia.
        apply period_enlarge_le_le; try omega.

        assert (H2: t2 / (k*T) * (k*T) <= t2 / T * T).
        {
          apply period_enlarge_le; omega.
        }

        destruct (zeq (t1 mod (k*T)) 0) as [Hmod1 | Hmod1].
        {
          assert (H3: t1 mod T = 0).
          {
            rewrite <- Fcore_Zaux.Zmod_mod_mult with (a := k); try omega.
            rewrite Hmod1.
            auto.
          }
          rewrite H3.
          rewrite zeq_true.
          split; try omega.

          assert (H4: t1 < t2 / (k*T) * (k*T)).
          {
            apply off_by_one_period_le; try nia.
          }
          omega.
        }
        {
          split.
          {
            assert (H3: (t1 + k*T) / (k*T) * (k*T) <= t2 / (k * T) * (k * T)).
            {
              apply div_mul_le; try nia.
            }
            rewrite div_T_inc_1 in H3; try nia.
          }
          {            
            assert (H3: (if zeq (t1 mod T) 0 then t1 else t1 / T * T + T) <=
                        t1 / (k * T) * (k * T) + k * T).
            {
              destruct (zeq (t1 mod T) 0).
              {
                assert (t1 < t1 / (k * T) * (k * T) + k * T).
                {
                  apply x_lt_floor_plus; try nia.
                }
                omega.
              }
              {
                apply period_enlarge_roundup_le; try omega.
              }
            }
            omega.
          }
        }
      }
    }
  Qed.

  Lemma workload_whole_enlarge_nonincr:
    forall i id k T C t1 t2,
      0 < k ->
      1 < T ->
      0 < C ->
      0 <= t1 <= t2 ->
      workload_whole i id (k * T) (k * C) t1 t2 <= workload_whole i id T C t1 t2.
  Proof.
    intros * Hk HT HC Ht.
    destruct (zeq k 1) as [Heq|Hneq].
    {
      rewrite Heq in *.
      replace (1*T) with T by omega.
      replace (1*C) with C by omega.
      omega.
    }
    {
      destruct (zlt (t2 - t1) (k*T)).
      {
        assert (H1: workload_whole i id (k * T) (k * C) t1 t2 = 0).
        {
          unfold workload_whole.
          rewrite zlt_true; auto.
        }
        rewrite H1.
        apply workload_whole_nonneg; try omega.
      }
      {
        unfold workload_whole.
        rewrite zlt_false; auto.
        
        assert (HH: (if zlt (t2 - t1) T
    then 0
    else
     ((if zeq (t2 mod T) 0 then if zlt id i then t2 / T * T else t2 - T else t2 / T * T) -
      (if zeq (t1 mod T) 0 then t1 else t1 / T * T + T)) / T * C) =
               (
     ((if zeq (t2 mod T) 0 then if zlt id i then t2 / T * T else t2 - T else t2 / T * T) -
      (if zeq (t1 mod T) 0 then t1 else t1 / T * T + T)) / T * C)).
        {
          rewrite zlt_false; try nia.
        }
        rewrite HH. clear HH.
        apply period_enlarge_le_le; try omega.

        assert (H2: t2 / (k*T) * (k*T) <= t2 / T * T).
        {
          apply period_enlarge_le; omega.
        }

        assert (Hadd: (if zeq (t2 mod (k * T)) 0
    then if zlt id i then t2 / (k * T) * (k * T) else t2 - k * T
                       else t2 / (k * T) * (k * T)) <=
                      (if zeq (t2 mod T) 0 then if zlt id i then t2 / T * T else t2 - T else t2 / T * T)).
        {
          destruct (zeq (t2 mod (k*T)) 0) as [Hmod1 | Hmod1].
          {
            assert (H3: t2 mod T = 0).
            {
              rewrite <- Fcore_Zaux.Zmod_mod_mult with (a := k); try omega.
              rewrite Hmod1.
              auto.
            }
            rewrite H3.
            rewrite zeq_true.
            destruct (zlt id i); try nia.
          }
          {
            destruct (zeq (t2 mod T) 0) as [HmodT | HmodT].
            {
              assert (H4: t2 / T * T = t2).
              {
                apply mod_0_divide_mul; omega.
              }
              rewrite H4 in *.                
              destruct (zlt id i); try omega.
              apply mod_not_0_large; omega.
            }
            {
              omega.
            }
          }
        }
        
        assert (Hsub: (if zeq (t1 mod T) 0 then t1 else t1 / T * T + T) <=
                  (if zeq (t1 mod (k * T)) 0 then t1 else t1 / (k * T) * (k * T) + k * T)).
        {
          destruct (zeq (t1 mod (k*T)) 0) as [Hmod1 | Hmod1].
          {
            assert (H3: t1 mod T = 0).
            {
              rewrite <- Fcore_Zaux.Zmod_mod_mult with (a := k); try omega.
              rewrite Hmod1.
              auto.
            }
            rewrite H3.
            rewrite zeq_true.
            destruct (zlt id i); try nia.
          }
          {
            transitivity (t1 / T * T + T).
            - destruct (zeq (t1 mod T) 0).
              + rewrite mod_0_divide_mul; omega.
              + omega.
            - apply period_enlarge_roundup_le; omega.
          }
        }

        assert (Hnonneg: (if zeq (t1 mod (k * T)) 0 then t1 else t1 / (k * T) * (k * T) + k * T) <=
               (if zeq (t2 mod (k * T)) 0
    then if zlt id i then t2 / (k * T) * (k * T) else t2 - k * T
    else t2 / (k * T) * (k * T))).
        {
          assert (H4: t1 < t2 / (k*T) * (k*T)).
          {
            apply off_by_one_period_le; try nia.
          }

          transitivity (if zeq (t2 mod (k * T)) 0
    then t2 - k * T
    else t2 / (k * T) * (k * T)).
          - apply roundup_le_rounddown_off_by_period; try omega.
            nia.
          - destruct (zeq (t2 mod (k * T)) 0); try omega.
            + rewrite mod_0_divide_mul; try nia.
              destruct (zlt id i); try nia.
        }

        split; omega.
      }
    }
  Qed.
  
  Lemma enlarged_task_whole_workload_noninc:
    forall id k T C i v ts te,
      0 <= v ->
      v <= ts <= te ->
      0 < k ->
      1 < T ->
      0 < C ->
      (if is_overlapping id (k * T) i v ts te then k * C else 0) +
      workload_extended_whole (k * T) (k * C) v ts +
      workload_whole i id (k * T) (k * C) ts te <=
      (if is_overlapping id T i v ts te then C else 0) + workload_extended_whole T C v ts + workload_whole i id T C ts te.
  Proof.
    intros * Hv Ht Hk HT HC.
    assert (H1: 1 < k * T). nia.      

    destruct (zeq ts te) as [Hright | Hright].
    {
      rewrite Hright.
      repeat rewrite workload_whole_empty_window_0; auto; try omega.
      repeat rewrite is_overlapping_false_empty_slot2; try omega.
      simpl.

      repeat rewrite Z.add_0_r.
      apply workload_extended_whole_enlarge_nonincr; try omega.
    }
    {
      rewrite workload_whole_decompose; try omega.
      rewrite workload_whole_decompose; try omega.
      apply workload_whole_enlarge_nonincr; try omega.
    }
  Qed.
  
  Theorem compare_workloads_when_a_task_is_enlarged:
    forall i T C N conf conf2 v ts id k,
      0 <= v <= ts ->
      1 < k ->
      0 <= i < N ->
      0 <= id < N ->
      i <> id ->
      id_is_enlarged_by_k id k conf conf2 ->
      enlarged_task_has_largest_period id k N conf ->      
      config_pool_valid N conf ->
      config_pool_valid N conf2 ->      
      ZMap.get i conf = mkPrioConfigValid T C ->
      workload_overlapping_total (Z.to_nat N) i v ts (ts+T) conf2 + workload_extended_whole_total (Z.to_nat N) conf2 v ts + workload_whole_total (Z.to_nat N) i ts (ts+T) conf2 <=
      workload_overlapping_total (Z.to_nat N) i v ts (ts+T) conf + workload_extended_whole_total (Z.to_nat N) conf v ts + workload_whole_total (Z.to_nat N) i ts (ts+T) conf.
  Proof.
    intros * Hv Hk Hi Hid Hid' Henlarge Hperiod Hconfvalid Hconf2valid Hconfi.
    assert (HT: 1 < T).
    {
      specialize (Hconfvalid _ _ _ Hi Hconfi).
      omega.
    }
    assert (HC: 0 < C).
    {
      specialize (Hconfvalid _ _ _ Hi Hconfi).
      omega.
    }
    
    remember (Z.to_nat N) as Nnat. clear HeqNnat.
    induction Nnat.
    {
      simpl.
      omega.
    }
    {
      simpl.
      destruct (zeq i (Z.of_nat Nnat)) as [Heqi | Hneqi].
      {
        rewrite <- Heqi in *.
        assert (Hconf2i: ZMap.get i conf2 = mkPrioConfigValid T C).
        {
          destruct (ZMap.get id conf) as [Tid Cid] eqn:Hconfid.          
          specialize (Henlarge i _ _ Hconfid).
          rewrite Henlarge.
          rewrite ZMap.gso; auto.
        }
        rewrite Hconf2i.
        rewrite Hconfi.
        omega.
      }
      {
        destruct (zeq id (Z.of_nat Nnat)) as [Heqid | Hneqid].
        {
          rewrite <- Heqid in *.
          destruct (ZMap.get id conf) as [Tid Cid] eqn:Hconfid.
          assert (Hconf2id: ZMap.get id conf2 = mkPrioConfigValid (k*Tid) (k*Cid)).
          {
            rewrite (Henlarge _ _ _ Hconfid).
            rewrite ZMap.gss.
            auto.
          }
          rewrite Hconf2id.

          assert (Hsum: forall (cond: bool) a b c d e f, (if cond then a + b else b) + (c + d) + (e + f) =
                 (if cond then a else 0) + d + e + (b + c + f)).
          {
            intros.
            destruct cond; omega.
          }
          repeat rewrite Hsum.

          assert (Hsingle: (if is_overlapping id (k * Tid) i v ts (ts + T) then k * Cid else 0) +
   workload_extended_whole (k * Tid) (k * Cid) v ts +
   workload_whole i id (k * Tid) (k * Cid) ts (ts + T) <=
                 (if is_overlapping id Tid i v ts (ts + T) then Cid else 0) + workload_extended_whole Tid Cid v ts +
                 workload_whole i id Tid Cid ts (ts + T)).
          {
            apply enlarged_task_whole_workload_noninc; try omega.
            - specialize (Hconfvalid _ _ _ Hid Hconfid).
              omega.
            - specialize (Hconfvalid _ _ _ Hid Hconfid).
              omega.              
          }
          omega.
        }
        {
          destruct (ZMap.get (Z.of_nat Nnat) conf) as [Tn Cn] eqn:Hconfn.
          assert (Hconf2n: ZMap.get (Z.of_nat Nnat) conf2 = mkPrioConfigValid Tn Cn).
          {
            destruct (ZMap.get id conf) as [Tid Cid] eqn:Hconfid.          
            rewrite (Henlarge _ _ _ Hconfid).
            rewrite ZMap.gso; auto.
          }
          rewrite Hconf2n.
          destruct (is_overlapping (Z.of_nat Nnat) Tn i v ts (ts + T)); omega.
        }
      }
    }        
  Qed.

End A_TASK_IS_ENLARGED.


