Require Import Coqlib.
Require Import Maps.
Require Import Values.
Require Import Integers.
Require Import ZArith.
Require Import VTDefs.
Require Import LemmaMaths.
Require Import Psatz.
Require Import Coq.Sorting.Mergesort.
Require Import Coq.Structures.Orders.
Require Import Coq.Sorting.Permutation.


Open Scope Z_scope.


Section VTMAPAVAIL.

  Function vt_map_available (pt: Z) (T: Z) (wcet: Z) (map: vtmap) : bool :=
    if zlt 0 (map (pt + 1) - (map pt)) then
      let pt0 := pt / T * T in
        zlt ((map pt) - (map pt0)) wcet 
    else false.

  (* returns true if the task is ready at pt .
    This is equivalent to specifying that  quantum > 0 *)
  Function vt_map_not_finished (pt: Z) (T: Z) (wcet: Z) (map: vtmap) : bool :=
    zlt (map pt - map (pt / T * T)) wcet.

  Lemma vt_map_not_finished_true_at_beginning:
    forall t T C vt,
      0 <= t ->
      0 < T ->
      0 < C ->
      t mod T = 0 ->      
      vt_map_not_finished t T C vt = true.
  Proof.
    intros.
    unfold vt_map_not_finished.
    rewrite mod_0_divide_mul; try omega.
    destruct (zlt (vt t -vt t) C); auto; omega.
  Qed.  
  
End VTMAPAVAIL.

Section CALQUANTUM.

  Function cal_quantum (t: Z) (c: vtmap) (T: Z) (C: Z) :=
    C - Z.min C (c (t + 1) - c (t / T * T)).

  Definition cal_quantum_valid_at_i (t: Z) (i: Z) (quantum: Z) (confs: PrioConfigPool)
             (mappings: TimeContainerPool): Prop :=
    match (ZMap.get i confs) with
      | mkPrioConfigValid T C =>
        quantum = cal_quantum t (ZMap.get i mappings) T C
    end.

  Lemma cal_quantum_nonneg:
    forall t c T C,
      t >= 0 ->
      cal_quantum t c T C >= 0.
  Proof.
    intros.
    unfold cal_quantum.
    pose proof (Z.le_min_l C (c (t + 1) - c (t / T * T))).
    omega.
  Qed.

  Lemma vt_available_new_period_or_cal_quantum_pos:
    forall t T C c,
      0 <= t ->
      1 < T ->
      vt_map_available (t + 1) T C c = true ->
      (t + 1) mod T = 0 \/ (((t + 1) mod T <> 0 ) /\ cal_quantum t c T C > 0).
  Proof.
    intros * Ht HT Havail.
    unfold vt_map_available in Havail.
    unfold cal_quantum.
    destruct (zeq ((t + 1) mod T) 0); try solve [left; auto].
    right.
    assert (c (t + 1) - c (t / T * T) < C).
    {
      destruct (zlt 0 (c (t + 1 + 1) - c (t + 1))); try discriminate Havail.
      destruct (zlt (c (t + 1) - c ((t + 1) / T * T)) C); try discriminate Havail. 
      clear Havail.
      assert ((t + 1) / T * T = t / T * T).
      {
        eapply mod_equal; omega.
      }
      rewrite H in l0.
      omega.
    }
    rewrite Z.min_r; omega.
  Qed.

  Lemma cal_quantum_zero_implies_vt_not_available:
    forall t T C c,
      0 <= t ->
      1 < T ->
      (t + 1) mod T <> 0 ->
      cal_quantum t c T C = 0 ->
      vt_map_available (t + 1) T C c = false.
  Proof.
    intros * Ht HT Hmod Hq.
    unfold vt_map_available.
    destruct (zlt 0 (c (t + 1 + 1) - c (t + 1))) as [Hdes1 | Hdes1]; try reflexivity.
    destruct (zlt (c (t + 1) - c ((t + 1) / T * T)) C) as [Hdes2 | Hdes2]; try reflexivity.
    simpl.
    unfold cal_quantum in Hq.
    destruct (zle C (c (t + 1) - c (t / T * T))) as [Hle | Hgt].
    {
      assert (Hm1: t / T = (t + 1) / T).
      {
        apply z_divide_same; omega.
      }
      rewrite <- Hm1 in *.
      omega.
    }
    {
      rewrite Z.min_r in Hq; omega.
    }
  Qed.

End CALQUANTUM.

Section PRIQUEUE.

  Inductive priEntry: Type :=
  | PriEntry: Z (* deadline *) -> Z (* task ID *) -> priEntry.
  
  Function priEntry_lt e1 e2 :=
    match e1, e2 with
    | PriEntry d1 id1, PriEntry d2 id2 =>
      if (zlt d1 d2) then true
      else (if zeq d1 d2 then
              (if zlt id1 id2 then true else false) else false)
    end.

  Function priEntry_le e1 e2 :=
    match e1, e2 with
    | PriEntry d1 id1, PriEntry d2 id2 =>
      if (zlt d1 d2) then true
      else (if zeq d1 d2 then
              (if zle id1 id2 then true else false) else false)
    end.
  
  (* A list of entries, from high to low priority  *)
  Definition priQueue := list priEntry.

  Fixpoint extract_id_queue (l: list priEntry) :=
    match l with
    | nil => nil
    | PriEntry d id :: tl => id :: extract_id_queue tl
    end.  

  Lemma id_in_extract_queue_if_in_queue:
    forall d id queue,
      In (PriEntry d id) queue ->
      In id (extract_id_queue queue).
  Proof.
    intros * H.
    induction queue.
    {
      simpl in H.
      omega.
    }
    {
      simpl in H.
      destruct H.
      - simpl.
        rewrite H.
        simpl.
        auto.
      - simpl.
        destruct a.
        simpl.
        right.
        apply IHqueue; auto.
    }
  Qed.
        
  Definition priqueue_sound (N t: Z) conf (queue: priQueue): Prop :=
    (Sorted.StronglySorted
         (fun x x0 : priEntry =>
            is_true ((fun x1 y : priEntry => priEntry_le x1 y) x x0)) queue) /\
    (forall d id T C, In (PriEntry d id) queue ->
                 ZMap.get id conf = mkPrioConfigValid T C ->
                 (0 <= id < N /\ d = t / T * T + T)).

  Lemma priqueue_sound_equal_id_equal_entry:
    forall N t conf queue d1 d2 id,
      priqueue_sound N t conf queue ->
      In (PriEntry d1 id) queue ->
      In (PriEntry d2 id) queue ->
      d1 = d2.
  Proof.
    intros * Hsound Hin1 Hin2.
    destruct Hsound as [_ Hval].
    destruct (ZMap.get id conf) as [T C] eqn:Hconf.
    pose proof (Hval _ _ _ _ Hin1 Hconf).
    pose proof (Hval _ _ _ _ Hin2 Hconf).    
    omega.
  Qed.
  
  Lemma priqueue_sound_shrink:
    forall N t conf a queue,
      priqueue_sound N t conf (a :: queue) -> priqueue_sound N t conf queue.
  Proof.
    intros.
    unfold priqueue_sound in *.
    destruct H as [Hsort Hval].
    split.
    {
      inv Hsort.
      auto.
    }
    {
      intros.
      apply Hval with (C := C); auto.
      simpl.
      right.
      auto.
    }
  Qed.

  Lemma id_in_extract_queue_then_entry_in_queue:
    forall i T C t conf N queue,
      priqueue_sound N t conf queue ->
      In i (extract_id_queue queue) ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      In (PriEntry (t / T * T + T) i) queue.
  Proof.
    intros *.    
    induction queue.
    {
      intros.
      simpl in *.
      omega.
    }
    {
      intros Hsound Hin Hconf.
      destruct a as [d id].
      simpl in Hin.      
      destruct Hin as [Hhead | Htail].
      {
        rewrite Hhead in *.
        destruct Hsound as [_ Hval].
        assert (H1: In (PriEntry d i) (PriEntry d i :: queue)).
        { simpl. auto. }

        specialize (Hval _ _ _ _ H1 Hconf).
        inv Hval.
        simpl.
        auto.
      }
      {
        simpl.
        right.
        apply IHqueue; auto.
        eapply priqueue_sound_shrink; eauto.
      }
    }
  Qed.
  
  Lemma priEntry_id_diff_lt:
    forall d1 i1 d2 i2,
      i1 <> i2 ->
      priEntry_le (PriEntry d1 i1) (PriEntry d2 i2) =
      priEntry_lt (PriEntry d1 i1) (PriEntry d2 i2).      
  Proof.
    intros.
    unfold priEntry_le, priEntry_lt.
    destruct (zlt d1 d2); auto.
    destruct (zeq d1 d2); auto.
    destruct (zlt i1 i2).
    - rewrite zle_true; auto; omega.
    - rewrite zle_false; auto; omega.
  Qed.

  Lemma priEntry_le_transitive:
    Transitive (fun x x0 : priEntry => is_true (priEntry_le x x0)).
  Proof.
    unfold Transitive.
    intros * H1 H2.
    unfold is_true in *.
    unfold priEntry_le in *.
    destruct x as [d1 id1].
    destruct y as [d2 id2].
    destruct z as [d3 id3].
    assert (Hd12: d1 <= d2).
    {
      destruct (zlt d1 d2); try omega.
      destruct (zeq d1 d2); try omega.
      congruence.
    }
    assert (Hd23: d2 <= d3).
    {
      destruct (zlt d2 d3); try omega.
      destruct (zeq d2 d3); try omega.
      congruence.
    }    
    destruct (zlt d1 d3); auto.
    destruct (zeq d1 d3) as [Heq | Hfalse].
    {
      assert (d1 = d2). omega.
      repeat subst.
      rewrite zlt_false in H2; try omega.
      rewrite zeq_true in H2.
      destruct (zle id2 id3); try congruence.
      rewrite zlt_false in H1; try omega.
      rewrite zeq_true in H1.
      destruct (zle id1 id2); try congruence.
      destruct (zle id1 id3); auto; try omega.
    }
    {
      omega.
    }
  Qed.

  Lemma priEntry_lt_transitive:
    Transitive (fun x x0 : priEntry => is_true (priEntry_lt x x0)).
  Proof.
    unfold Transitive.
    intros * H1 H2.
    unfold is_true in *.
    unfold priEntry_lt in *.
    destruct x as [d1 id1].
    destruct y as [d2 id2].
    destruct z as [d3 id3].
    assert (Hd12: d1 <= d2).
    {
      destruct (zlt d1 d2); try omega.
      destruct (zeq d1 d2); try omega.
      congruence.
    }
    assert (Hd23: d2 <= d3).
    {
      destruct (zlt d2 d3); try omega.
      destruct (zeq d2 d3); try omega.
      congruence.
    }    
    destruct (zlt d1 d3); auto.
    destruct (zeq d1 d3) as [Heq | Hfalse].
    {
      assert (d1 = d2). omega.
      repeat subst.
      rewrite zlt_false in H2; try omega.
      rewrite zeq_true in H2.
      destruct (zlt id2 id3); try congruence.
      rewrite zlt_false in H1; try omega.
      rewrite zeq_true in H1.
      destruct (zlt id1 id2); try congruence.
      destruct (zlt id1 id3); auto; try omega.
    }
    {
      omega.
    }
  Qed.

  Lemma priEntry_lt_self_false:
    forall a,
      priEntry_lt a a = false.
  Proof.
    intros.
    destruct a.
    unfold priEntry_lt.
    rewrite zlt_false; try omega.
    rewrite zeq_true.
    rewrite zlt_false; try omega.
    auto.
  Qed.
  
  Lemma priEntry_lt_deadline_le:
    forall d1 id1 d2 id2,
      priEntry_lt (PriEntry d1 id1) (PriEntry d2 id2) = true ->
      d1 <= d2.
  Proof.
    intros.
    unfold priEntry_lt in H.
    destruct (zlt d1 d2); try omega.
    destruct (zeq d1 d2); try omega.
    congruence.
  Qed.

  Lemma priEntry_lt_neg_tf:
    forall d1 id1 d2 id2,
      priEntry_lt (PriEntry d1 id1) (PriEntry d2 id2) = true ->
      priEntry_lt (PriEntry d2 id2) (PriEntry d1 id1) = false.
  Proof.
    intros * Ht.
    unfold priEntry_lt in *.
    destruct (zlt d1 d2).
    {
      rewrite zlt_false; try omega.
      rewrite zeq_false; try omega.
      auto.
    }
    {
      destruct (zeq d1 d2); try discriminate.
      rewrite e in *.
      rewrite zlt_false; try omega.
      rewrite zeq_true.
      destruct (zlt id1 id2); try discriminate.
      rewrite zlt_false; auto; omega.
    }
  Qed.

  Lemma priEntry_lt_neg_ft:
    forall d1 id1 d2 id2,
      id1 <> id2 ->
      priEntry_lt (PriEntry d1 id1) (PriEntry d2 id2) = false ->
      priEntry_lt (PriEntry d2 id2) (PriEntry d1 id1) = true.
  Proof.
    intros * Hid Ht.
    unfold priEntry_lt in *.
    destruct (zlt d1 d2); try congruence.
    destruct (zeq d1 d2).
    {
      rewrite e in *.
      rewrite zlt_false; try omega.
      rewrite zeq_true.
      destruct (zlt id1 id2); try discriminate.
      rewrite zlt_true; auto; omega.
    }
    {
      rewrite zlt_true; try omega.
      auto.
    }
  Qed.
  
  Lemma permutation_extract_id_queue:
    forall l l',
      Permutation l l' ->
      Permutation (extract_id_queue l) (extract_id_queue l').
  Proof.
    intros * H.
    induction H.
    - simpl.
      constructor.
    - destruct x.
      simpl.
      constructor; auto.
    - destruct x, y.
      simpl.
      constructor; auto.
    - eapply Permutation_trans; eauto.
  Qed.   
      
End PRIQUEUE.

Module PriOrder <: TotalLeBool.

  Definition t := priEntry.
  
  Function leb (x y: priEntry) :=
    priEntry_le x y.
  
  Infix "<=p" := leb (at level 35).
  
  Theorem leb_total : forall a1 a2, a1 <=p a2 = true  \/ a2 <=p a1 = true.
  Proof.
    intros.
    destruct a1 as [d1 id1], a2 as [d2 id2].
    unfold leb, priEntry_le.
    destruct (zlt d1 d2) as [Hdlt | Hdge]; auto.
    destruct (zeq d1 d2) as [Hdeq | Hdneq].
    - rewrite Hdeq. clear Hdge Hdeq.
      rewrite zeq_true.
      rewrite zlt_false; try omega.
      destruct (zle id1 id2) as [H | H]; auto.
      + right.
        rewrite zle_true; auto.
        omega.
    - rewrite zlt_true; auto.
      omega.
  Qed.      
    
End PriOrder.

Module Import prisort := Sort PriOrder.

Section PRIQUEUEUPDATE.

(*
Eval compute in prisort.sort ((PriEntry 10 0) :: (PriEntry 4 3) :: (PriEntry 6 2)
                                      :: (PriEntry 4 1) :: nil).
*)

  Fixpoint priqueue_task_arrive_aux (deadline id: Z) (queue: priQueue) :=
    match queue with
    | nil => nil
    | PriEntry dy idy :: l =>
      if (zeq id idy) then
        PriEntry deadline id :: l
      else
        PriEntry dy idy :: priqueue_task_arrive_aux deadline id l
    end.
  
  Function priqueue_task_arrive (deadline id: Z) (queue: priQueue) :=
    prisort.sort (priqueue_task_arrive_aux deadline id queue).

  Fixpoint priqueue_tick (N: nat) (t: Z) (conf: PrioConfigPool) (queue: priQueue) :=
    match N with
    | O => queue
    | S n =>
      match (ZMap.get (Z.of_nat n) conf) with
      | mkPrioConfigValid T C =>
        if (zeq (t mod T) 0) then
          priqueue_task_arrive (t + T) (Z.of_nat n)
                               (priqueue_tick n t conf queue)
        else
          priqueue_tick n t conf queue
      end
    end.

  Fixpoint real_priqueue_aux (N: nat) (t: Z) (conf: PrioConfigPool):=
    match N with
    | O => nil
    | S n =>
      match (ZMap.get (Z.of_nat n) conf) with
      | mkPrioConfigValid T C =>
        PriEntry (t / T * T + T) (Z.of_nat n) :: (real_priqueue_aux n t conf)
      end
    end.

  Definition real_priqueue (N: nat) (t: Z) (conf: PrioConfigPool) :=
    prisort.sort (real_priqueue_aux N t conf).


  (* properties *)
  Lemma entry_same_in_aux_and_real:
    forall a l,
      In a l <->
      In a (sort l).
  Proof.
    intros.
    pose proof (Permuted_sort l).
    split.
    - apply Permutation_in; auto.
    - intros.
      apply Permutation_sym in H.
      apply Permutation_in with (l := sort l); auto.
  Qed.
    
  Lemma real_priqueue_aux_contents_valid:
    forall Nnat t d id T C conf,
      In (PriEntry d id) (real_priqueue_aux Nnat t conf) ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      0 <= id < Z.of_nat Nnat /\ d = t / T * T + T.
  Proof.
    intro Nnat.
    induction Nnat.
    {
      intros.
      simpl in H.
      contradiction.
    }
    {
      intros * Hin Hconf.
      assert (H1: 0 <= Z.of_nat Nnat < Z.of_nat (S Nnat)).
      { rewrite Nat2Z.inj_succ; omega. }
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq].
      {
        rewrite Heq in *. 
        split; auto.
        simpl in Hin.
        rewrite Hconf in Hin.
        inv Hin.
        - inv H; auto.
        - specialize (IHNnat t d (Z.of_nat Nnat) T C conf H Hconf).
          destruct IHNnat.
          auto.
      }
      {
        simpl in Hin.
        destruct (ZMap.get (Z.of_nat Nnat) conf).
        inv Hin.
        - inv H; omega.
        - specialize (IHNnat _ _ _ _ _ _ H Hconf).
          destruct IHNnat.
          split; auto.
          rewrite Nat2Z.inj_succ.
          omega.
      }
    }
  Qed.

  Lemma real_priqueue_aux_complete:
    forall Nnat t id T C conf,
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      In (PriEntry (t / T * T + T) id) (real_priqueue_aux Nnat t conf).      
  Proof.
    intros Nnat.
    induction Nnat.
    {
      simpl.
      intros.
      omega.
    }
    {
      intros * Hid Hconf.
      destruct (zeq id (Z.of_nat Nnat)) as [Heq | Hneq].
      {
        subst.
        simpl.
        rewrite Hconf.
        simpl.
        auto.
      }
      {
        simpl.
        destruct (ZMap.get (Z.of_nat Nnat) conf).
        simpl.
        right.
        apply IHNnat with (C := C); auto.
        rewrite Nat2Z.inj_succ in Hid.
        omega.
      }
    }
  Qed.

  Lemma real_priqueue_aux_extract_id_bound:
    forall id Nnat t conf,
      In id (extract_id_queue (real_priqueue_aux Nnat t conf)) ->
      0 <= id < Z.of_nat Nnat.
  Proof.
    intros *.
    induction Nnat.
    {
      simpl.      
      omega.
    }
    {
      intros Hin.
      simpl in Hin.
      destruct (ZMap.get (Z.of_nat Nnat) conf).
      simpl in Hin.
      rewrite Nat2Z.inj_succ.
      destruct Hin as [H1 | H2]; try omega.
      specialize (IHNnat H2).
      omega.
    }
  Qed.  
  
  Lemma real_priqueue_aux_ids_nodup:
    forall Nnat t conf,
      NoDup (extract_id_queue (real_priqueue_aux Nnat t conf)).
  Proof.
    intros.
    induction Nnat.
    {
      simpl.
      constructor.
    }
    {
      simpl.
      destruct (ZMap.get (Z.of_nat Nnat) conf) as [T C].
      simpl.
      constructor; auto.
      unfold not.
      intros Hin.
      pose proof (real_priqueue_aux_extract_id_bound _ _ _ _ Hin).
      omega.
    }
  Qed.

  Lemma permutation_preserves_extract_id_nodup (A: Type):
    forall (l l': list A),
      Permutation l l' ->
      NoDup l ->
      NoDup l'.
  Proof.
    intros * Hperm Hdup.
    induction Hperm.
    - constructor.
    - inv Hdup.
      specialize (IHHperm H2).
      constructor; auto.
      unfold not.
      intros.
      assert (In x l).
      {
        apply Permutation_in with (l := l'); auto.
        apply Permutation_sym; auto.
      }
      congruence.
    - inv Hdup.
      inv H2.
      assert (Hneq: x <> y).
      {
        unfold not.
        intros.
        apply H1.
        constructor.
        auto.
      }
      constructor.
      + unfold not.
        intros.        
        inv H; try congruence.
      + constructor; auto.
        unfold not.
        intros.
        apply H1.
        simpl.
        auto.
    - auto.
  Qed.  
  
  Lemma real_priqueue_ids_nodup:
    forall Nnat t conf,
      NoDup (extract_id_queue (real_priqueue Nnat t conf)).
  Proof.
    intros.
    unfold real_priqueue.
    pose proof (real_priqueue_aux_ids_nodup Nnat t conf).
    apply permutation_preserves_extract_id_nodup with (l := (extract_id_queue (real_priqueue_aux Nnat t conf))); auto.
    apply permutation_extract_id_queue.
    pose proof (Permuted_sort (real_priqueue_aux Nnat t conf)) as Hperm.
    auto.
  Qed.
  
  Lemma real_priqueue_sound:
    forall Nnat t conf,
      priqueue_sound (Z.of_nat Nnat) t conf (real_priqueue Nnat t conf).
  Proof.
    intros.
    unfold priqueue_sound.
    split.
    {
      (* sorted *)
      apply StronglySorted_sort.
      apply priEntry_le_transitive.
    }
    {
      intros * Hin Hconf.
      unfold real_priqueue in Hin.
      apply real_priqueue_aux_contents_valid with (conf := conf) (C := C); auto.
      apply entry_same_in_aux_and_real; auto.
    }
  Qed.

  Lemma real_priqueue_complete:
    forall Nnat t id T C conf,
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      In (PriEntry (t / T * T + T) id) (real_priqueue Nnat t conf).      
  Proof.
    intros.
    unfold real_priqueue.
    pose proof (entry_same_in_aux_and_real (PriEntry (t / T * T + T) id) (real_priqueue_aux Nnat t conf)) as Hsort.
    apply Hsort.
    apply real_priqueue_aux_complete with (C := C); auto.
  Qed.

  Lemma real_priqueue_extract_id_complete:
    forall Nnat t id conf,
      0 <= id < Z.of_nat Nnat ->
      In id (extract_id_queue (real_priqueue Nnat t conf)).
  Proof.
    intros.
    destruct (ZMap.get id conf) as [T C] eqn:Hconf.
    pose proof (real_priqueue_complete _ t _ _ _ _ H Hconf) as H1.
    pose proof (id_in_extract_queue_if_in_queue _ _ _ H1).
    auto.
  Qed.
    
End PRIQUEUEUPDATE.

Section HIGHESTAVAIL.

  Fixpoint highest_available_map (idle: Z) (t: Z) (TAC: TimeContainerPool) (conf: PrioConfigPool) (queue: priQueue) :=
    match queue with
    | nil => idle
    | (PriEntry deadline tid) :: l =>
      match (ZMap.get tid conf) with
      | mkPrioConfigValid T C =>
        if (vt_map_available t T C (ZMap.get tid TAC)) then
          tid
        else highest_available_map idle t TAC conf l
      end
    end.     

  Fixpoint highest_not_finished_map (idle: Z) (t: Z) (TAC: TimeContainerPool) (conf: PrioConfigPool) (queue: priQueue) :=
    match queue with
    | nil => idle
    | (PriEntry deadline tid) :: l =>
      match (ZMap.get tid conf) with
      | mkPrioConfigValid T C =>
        if (vt_map_not_finished t T C (ZMap.get tid TAC)) then
          tid
        else highest_not_finished_map idle t TAC conf l
      end
    end.

  Lemma highest_not_finished_map_res_bound:
    forall N t TAC conf queue,
      0 < N ->
      priqueue_sound N t conf queue ->
      0 <= highest_not_finished_map N t TAC conf queue <= N.
  Proof.
    intros * HN Hq.
    induction queue.
    {
      simpl.
      omega.
    }
    {
      destruct a as [d i].
      assert (Hq': priqueue_sound N t conf queue).
      {
        apply priqueue_sound_shrink with (a := PriEntry d i); auto.
      }
      specialize (IHqueue Hq').
      destruct Hq as [Hsort Hvalid].
      simpl.
      destruct (ZMap.get i conf) as [T C] eqn:Hconf.
      assert (Hin: In (PriEntry d i) (PriEntry d i :: queue)).
      { simpl. auto. }
      specialize (Hvalid _ _ _ _ Hin Hconf).
      destruct (vt_map_not_finished t T C (ZMap.get i TAC)); omega.
    }
  Qed.
  
End HIGHESTAVAIL.

Section TIMEMAPTICK.
    
  (* update the timemap pool at time t*)
  (* queue is already updated to time t *)
  (* returns time map which is valid up to (t + 1) *)
  Fixpoint time_map_tick (t: Z) (TAC: TimeContainerPool) (conf: PrioConfigPool)
           (queue: priQueue) :=
    match queue with
    | nil => TAC
    | (PriEntry deadline tid) :: l =>
      match (ZMap.get tid conf) with
      | mkPrioConfigValid T C =>
        let map_id := ZMap.get tid TAC in
        let map_id' := fun x => if (zle (t + 1) x) then (map_id t + 1) else (map_id x) in        
        if (vt_map_not_finished t T C (ZMap.get tid TAC)) then
          ZMap.set tid map_id' TAC else
          ZMap.set tid map_id' (time_map_tick t TAC conf l)        
      end
    end.

  Definition vt_constant_starting_t (t: Z) (vt: vtmap) :Prop :=
    forall x, t <= x -> vt x = vt t.

  Definition TAC_constant_starting_t (t: Z) (TAC: TimeContainerPool) :=
    forall i x,
      t <= x ->
      ZMap.get i TAC x = ZMap.get i TAC t.

  Lemma TAC_constant_starting_t_extend:
    forall t1 t2 TAC,
      t1 <= t2 ->
      TAC_constant_starting_t t1 TAC ->
      TAC_constant_starting_t t2 TAC.
  Proof.
    unfold TAC_constant_starting_t.
    intros.
    rewrite H0; try omega.
    symmetry.
    rewrite H0; try omega.
  Qed.
  
  Lemma time_map_tick_incremental:
    forall t TAC conf queue id x,
      x <= t ->
      (ZMap.get id (time_map_tick t TAC conf queue)) x =      
      (ZMap.get id TAC) x.
  Proof.
    intros * Hx.
    induction queue; auto.
    simpl.
    destruct a as [_ aid].
    destruct (zeq id aid) as [Heq | Hneq].
    {
      subst.
      destruct (ZMap.get aid conf).
      destruct (vt_map_not_finished t z z0 (ZMap.get aid TAC)); rewrite ZMap.gss;
        rewrite zle_false; omega.
    }
    {
      destruct (ZMap.get aid conf).
      destruct (vt_map_not_finished t z z0 (ZMap.get aid TAC)); rewrite ZMap.gso; auto.
    }
  Qed.
  
  Lemma time_map_tick_preserves_monotonic_continuous:
    forall t (TAC: TimeContainerPool) conf queue Nnat,
      (forall i, 0 <= i < Z.of_nat Nnat -> monotonic_continuous (ZMap.get i TAC)) ->
      (forall i, 0 <= i < Z.of_nat Nnat -> monotonic_continuous (ZMap.get i
                                                                   (time_map_tick t TAC conf queue))).
  Proof.
    intros *.
    induction queue.
    {
      intros.
      simpl.
      auto.
    }
    {
      intros H i Hi.
      simpl.
      destruct a as [d id].
      destruct (ZMap.get id conf) as [T C].

      destruct (zeq i id) as [Heq | Hneq].
      - rewrite Heq in *. clear Heq.
        assert (H1: monotonic_continuous (fun x : Z => if zle (t + 1) x then ZMap.get id TAC t + 1 else ZMap.get id TAC x)).
        {
          specialize (H _ Hi).
          unfold monotonic_continuous.
          intros x Hx.
          destruct (zle (t + 1) x) as [Hbefore | Hafter].
          - rewrite zle_true; try omega.
          - destruct (zeq t x) as [H1 | H1].
            + rewrite H1.
              rewrite zle_true; try omega.
            + rewrite zle_false; try omega.
              apply H; omega.
        }
        destruct (vt_map_not_finished t T C (ZMap.get id TAC)) as [Hready |Hskip]; rewrite ZMap.gss; auto.
      - assert (H1: monotonic_continuous (ZMap.get i (time_map_tick t TAC conf queue))).
        {
          apply IHqueue; auto.
        }
        destruct (vt_map_not_finished t T C (ZMap.get id TAC)) as [Hready |Hskip]; rewrite ZMap.gso; auto.
    }
  Qed.

  Lemma time_map_tick_extends_TAC_constant_starting_t_plus_1:
    forall t conf TAC queue,
      TAC_constant_starting_t t TAC ->
      TAC_constant_starting_t (t+1) (time_map_tick t TAC conf queue).
  Proof.
    intros * Hpre.
    induction queue.
    {
      simpl.
      apply TAC_constant_starting_t_extend with (t1 := t); auto; omega.
    }
    {
      destruct a as [d i].
      simpl.
      destruct (ZMap.get i conf) as [T C].
      unfold TAC_constant_starting_t.
      intros id x Hx.
      destruct (zeq id i) as [Heq | Hneq].
      {
        rewrite Heq.
        assert (H1: (if zle (t + 1) x then ZMap.get i TAC t + 1 else ZMap.get i TAC x) =
                    (if zle (t + 1) (t + 1) then ZMap.get i TAC t + 1 else ZMap.get i TAC (t + 1))).
        {
          rewrite zle_true; auto.
          rewrite zle_true; try omega.
        }
        destruct (vt_map_not_finished t T C (ZMap.get i TAC)); rewrite ZMap.gss; auto.
      }
      {
        destruct (vt_map_not_finished t T C (ZMap.get i TAC)); rewrite ZMap.gso; auto.
        rewrite Hpre; try omega.
        symmetry.
        rewrite Hpre; try omega.
      }
    }
  Qed.

End TIMEMAPTICK.      

Section UPDATEMAP.

  (* to get the time map up to time t, we simply tick from 0 to (t - 1) *)
  Fixpoint TAC_pool_at_t_aux (nt: nat) (N : nat) (conf: PrioConfigPool) : TimeContainerPool:=
    match nt with
    | O => real_time_container_pool
    | S n =>
      time_map_tick (Z.of_nat n)
                    (TAC_pool_at_t_aux n N conf) conf
                    (real_priqueue N (Z.of_nat n) conf)
    end.

  (* return valid map from time 0 up to t *)
  Function TAC_pool_at_t (t: Z) (N: nat) (conf: PrioConfigPool) : TimeContainerPool :=
    TAC_pool_at_t_aux (Z.to_nat t) N conf.

  (* formalization of schedulability *)
  
  Definition schedulable_per_level_upto (t: Z) (T: Z) (C: Z) (vt: vtmap) :=
    forall k,
      0 <= k ->
      k * T + T <= t ->
      C <= vt (k * T + T) - vt (k * T).      

  Definition schedulable_upto (t: Z) (N: Z) (conf: PrioConfigPool) (mappings: TimeContainerPool) :=
    forall p, 0 <= p < N ->
      let '(mkPrioConfigValid T C) := ZMap.get p conf in
      schedulable_per_level_upto t T C (ZMap.get p mappings).

  Definition system_schedulable (N: Z) (conf: PrioConfigPool): Prop:=
    forall t, 0 <= t ->
         schedulable_upto t N conf (TAC_pool_at_t t (Z.to_nat N) conf).  

  Lemma TAC_pool_aux_all_monotonic_continuous:
    forall nt N conf i,
      0 <= i < Z.of_nat N ->
      monotonic_continuous (ZMap.get i (TAC_pool_at_t_aux nt N conf)).
  Proof.
    intros nt N conf.
    induction nt.
    {
      intros * Hi.
      simpl.
      unfold real_time_container_pool.
      rewrite ZMap.gi.
      unfold monotonic_continuous.
      auto.
    }
    {
      intros i Hi.
      simpl.
      eapply time_map_tick_preserves_monotonic_continuous; eauto.
    }
  Qed.

  Lemma TAC_pool_all_monotonic_continuous:
    forall t N conf i,
      0 <= t ->
      0 <= i < Z.of_nat N ->
      monotonic_continuous (ZMap.get i (TAC_pool_at_t t N conf)).
  Proof.
    intros.
    unfold TAC_pool_at_t.
    apply TAC_pool_aux_all_monotonic_continuous; auto.
  Qed.
  
  Lemma vertical_map_aux_canonical_step:
    forall nt x id Nnat conf,
      x <= Z.of_nat nt ->
      ZMap.get id (TAC_pool_at_t_aux (S nt) Nnat conf) x =
      ZMap.get id (TAC_pool_at_t_aux nt Nnat conf) x.
  Proof.
    intros.
    simpl.
    eapply time_map_tick_incremental; eauto.    
  Qed.    

  Lemma vertical_map_aux_canonical:
    forall nt1 nt2 x id Nnat conf,
      x <= Z.of_nat nt1 ->
      Z.of_nat nt1 <= Z.of_nat nt2 ->
      ZMap.get id (TAC_pool_at_t_aux nt2 Nnat conf) x =
      ZMap.get id (TAC_pool_at_t_aux nt1 Nnat conf) x.
  Proof.
    intros nt1 nt2.
    revert nt1.
    induction nt2.
    {
      intros * H1 H2.
      simpl in H2.
      assert (H3: Z.of_nat nt1 = 0).
      {
        pose proof (Zle_0_nat nt1).
        omega.
      }
      assert (H4: Z.to_nat (Z.of_nat nt1) = Z.to_nat 0). rewrite H3. auto.
      rewrite Nat2Z.id in H4.
      simpl in H4.
      rewrite H4.
      auto.
    }
    {
      intros * Hx Hnt1.
      destruct (zeq (Z.of_nat nt1) (Z.of_nat (S nt2))) as [Heq | Hneq].
      {
        assert (Z.to_nat (Z.of_nat nt1) = Z.to_nat (Z.of_nat (S nt2))).
        {
          rewrite Heq.
          auto.
        }
        repeat rewrite Nat2Z.id in H.
        subst.
        auto.
      }
      {
        assert (H1: Z.of_nat nt1 <= Z.of_nat nt2).
        {
          rewrite Nat2Z.inj_succ in *.
          omega.
        }
        rewrite vertical_map_aux_canonical_step; try omega.
        apply IHnt2; auto.
      }
    }
  Qed.    
  
  Lemma vertical_map_canonical:
    forall x t id Nnat conf,
      0 <= t ->
      t <= x ->
      ZMap.get id (TAC_pool_at_t x Nnat conf) t =
      ZMap.get id (TAC_pool_at_t t Nnat conf) t.
  Proof.
    intros.
    remember (Z.to_nat x) as nx.
    assert (H1: x = Z.of_nat nx).
    {
      rewrite Heqnx.
      rewrite Z2Nat.id; omega.
    }
    clear Heqnx.
    subst.
    remember (Z.to_nat t) as nt.
    assert (H2: t = Z.of_nat nt).
    {
      rewrite Heqnt.
      rewrite Z2Nat.id; omega.
    }
    clear Heqnt.
    subst.
    apply vertical_map_aux_canonical; auto; try (repeat rewrite Z2Nat.id; omega).
  Qed.

  Lemma system_schedulable_implies_period_schedulable:
    forall N conf i T C t,
      0 <= t ->
      1 < T ->
      0 <= i < N ->
      system_schedulable N conf ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      t mod T = 0 ->
      C <= ZMap.get i (TAC_pool_at_t (t + T) (Z.to_nat N) conf) (t + T) -
          ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t.
  Proof.
    intros * Ht HT Hi Hsched Hconfi Hmod.
    unfold system_schedulable in Hsched.
    assert (H1: 0 <= t + T). omega.
    specialize (Hsched _ H1).
    unfold schedulable_upto in Hsched.
    specialize (Hsched _ Hi).
    rewrite Hconfi in Hsched.
    unfold schedulable_per_level_upto in Hsched.

    assert (Ht2: t / T * T = t).
    {
      rewrite mod_0_divide_mul; omega.
    }
    assert (Hvt1: ZMap.get i (TAC_pool_at_t (t + T) (Z.to_nat N) conf) (t + T) =
                  ZMap.get i (TAC_pool_at_t (t + T) (Z.to_nat N) conf) (t / T * T + T)).
    {      
      rewrite Ht2. auto.
    }
    rewrite Hvt1.
    assert (Hvt2: ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t =
                  ZMap.get i (TAC_pool_at_t (t + T) (Z.to_nat N) conf) (t / T * T)).
    {
      rewrite Ht2.
      symmetry.
      rewrite vertical_map_canonical; omega.
    }
    rewrite Hvt2.

    apply Hsched; try omega.
    apply Z_div_pos; omega.
  Qed.
  
  Lemma monotonic_vt_step_bound:
    forall i t N conf,
      0 <= t ->
      0 <= i < N ->
      ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t <=
      ZMap.get i (TAC_pool_at_t (t + 1) (Z.to_nat N) conf) (t + 1) <=
      ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t + 1.
  Proof.
    intros * Ht Hi.
    assert (Hvt: monotonic_continuous (ZMap.get i (TAC_pool_at_t (t + 1) (Z.to_nat N) conf))).
    {
      apply TAC_pool_all_monotonic_continuous; try omega.
      - rewrite Z2Nat.id; omega.
    }
    assert (H1: t >= 0). omega.
    specialize (Hvt _ H1).
    assert (H2: ZMap.get i (TAC_pool_at_t (t + 1) (Z.to_nat N) conf) t = ZMap.get i (TAC_pool_at_t t (Z.to_nat N) conf) t).
    {
      rewrite vertical_map_canonical; auto; omega.
    }
    rewrite H2 in *.
    
    destruct Hvt as [Hvt|Hvt]; rewrite Hvt; omega.
  Qed.

  Lemma TAC_pool_at_t_monotonic_two_points:
    forall i t1 t2 N conf,
      0 <= t1 <= t2 ->
      0 <= i < N ->
      ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1 <=
      ZMap.get i (TAC_pool_at_t t2 (Z.to_nat N) conf) t2.
  Proof.
    intros * Ht Hi.
    assert (Hvt: monotonic_continuous (ZMap.get i (TAC_pool_at_t t2 (Z.to_nat N) conf))).
    {
      apply TAC_pool_all_monotonic_continuous; try omega.
      - rewrite Z2Nat.id; omega.
    }

    replace (ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1) with (ZMap.get i (TAC_pool_at_t t2 (Z.to_nat N) conf) t1).
    - apply monotonic_continuous_monotonic; auto; try omega.
    - rewrite vertical_map_canonical; omega.
  Qed.
  
  Lemma TAC_constant_starting_t_real:
    TAC_constant_starting_t 0 real_time_container_pool.
  Proof.
    unfold TAC_constant_starting_t, real_time_container_pool.
    intros.
    rewrite ZMap.gi.
    auto.
  Qed.
  
  Lemma TAC_pool_at_t_aux_constant_starting_t:
    forall nt Nnat conf,
      TAC_constant_starting_t (Z.of_nat nt) (TAC_pool_at_t_aux nt Nnat conf).
  Proof.
    intros.
    induction nt.
    {
      simpl.
      apply TAC_constant_starting_t_real.
    }
    {
      simpl.
      replace (Z.pos (Pos.of_succ_nat nt)) with (Z.of_nat nt + 1).
      apply time_map_tick_extends_TAC_constant_starting_t_plus_1; auto.
      xomega.
    }
  Qed.

  Lemma TAC_pool_at_t_constant_starting_t:
    forall t Nnat conf,
      0 <= t ->
      TAC_constant_starting_t t (TAC_pool_at_t t Nnat conf).
  Proof.
    intros.
    unfold TAC_pool_at_t.
    remember (Z.to_nat t) as nt.
    replace t with (Z.of_nat nt).
    - apply TAC_pool_at_t_aux_constant_starting_t.
    - rewrite Heqnt.
      rewrite Z2Nat.id; omega.
  Qed.

  Lemma vertical_map_canonical2:
    forall i N conf t1 t2,
      0 <= i < N ->
      0 <= t1 <= t2 ->
      ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t2 =
      ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t1.
  Proof.
    intros * Hi Ht.
    assert (H1: 0 <= t1). omega.
    pose proof (TAC_pool_at_t_constant_starting_t _ (Z.to_nat N) conf H1).
    unfold TAC_constant_starting_t in H.
    rewrite H; omega.
  Qed.

  Lemma vt_map_not_finished_canonical:
    forall i t1 t2 T C N conf,
      0 <= t1 <= t2 ->
      0 < T ->
      vt_map_not_finished t1 T C (ZMap.get i (TAC_pool_at_t t2 (Z.to_nat N) conf)) =
      vt_map_not_finished t1 T C (ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf)).      
  Proof.
    intros * Ht HT.
    unfold vt_map_not_finished.
    rewrite vertical_map_canonical; try omega.
    assert (H1: t1 / T * T <= t1).
    {
      apply divide_mul_le; try omega.
    }
    assert (H2: 0 <= t1 / T * T).
    {
      apply div_mul_nonneg; omega.
    }
    assert (Hvt: ZMap.get i (TAC_pool_at_t t2 (Z.to_nat N) conf) (t1 / T * T) =
           ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) (t1 / T * T)).
    {
      rewrite vertical_map_canonical; try omega.
      symmetry.
      rewrite vertical_map_canonical; try omega.
    }
    rewrite Hvt.
    auto.
  Qed.
  
  Lemma TAC_pool_at_t_nonneg:
    forall i N conf t1 t2,
      0 <= i < N ->
      0 <= t1 ->
      0 <= t2 ->
      0 <= ZMap.get i (TAC_pool_at_t t1 (Z.to_nat N) conf) t2.
  Proof.
    intros * Hi Ht1 Ht2.
    assert (Hvt0: ZMap.get i (TAC_pool_at_t 0 (Z.to_nat N) conf) 0 = 0).
    {
      unfold TAC_pool_at_t.
      simpl.
      unfold real_time_container_pool.
      rewrite ZMap.gi.
      auto.
    }
    
    destruct (zle t1 t2).
    {
      rewrite vertical_map_canonical2; try omega.
      transitivity (ZMap.get i (TAC_pool_at_t 0 (Z.to_nat N) conf) 0).
      - omega.
      - apply TAC_pool_at_t_monotonic_two_points; try omega.
    }
    {
      rewrite vertical_map_canonical; try omega.
      transitivity (ZMap.get i (TAC_pool_at_t 0 (Z.to_nat N) conf) 0).
      - omega.
      - apply TAC_pool_at_t_monotonic_two_points; try omega.
    }
  Qed.
    
  Lemma vt_not_changed_if_id_not_in_queue:
    forall i t x TAC conf queue,
      ~ In i (extract_id_queue queue) ->
      (ZMap.get i (time_map_tick t TAC conf queue)) x =
      ZMap.get i TAC x.
  Proof.
    intros * Hnotin.
    induction queue.
    {
      simpl.
      auto.
    }
    {
      simpl.
      destruct a as [d id].
      assert (Hneq: i <> id).
      {
        destruct (zeq i id); auto.
        rewrite e in Hnotin.
        contradiction Hnotin.
        simpl.
        auto.
      }
      {
        destruct (ZMap.get id conf).
        destruct (vt_map_not_finished t z z0 (ZMap.get id TAC)) eqn:Hfinish.
        - rewrite ZMap.gso; auto.
        - rewrite ZMap.gso; auto.
          apply IHqueue; auto.
          unfold not.
          intros.
          apply Hnotin.
          simpl.
          auto.
      }
    }  
  Qed.

  Lemma highest_not_finished_map_res_in_queue:
    forall i N t TAC conf queue,      
      i = highest_not_finished_map N t TAC conf queue ->
      0 <= i < N ->
      In i (extract_id_queue queue).
  Proof.
    intros * Hhigh Hi.
    induction queue.
    {
      simpl in Hhigh.
      omega.
    }
    {
      simpl in Hhigh.
      destruct a as [d id].
      destruct (ZMap.get id conf) as [T C].
      destruct (vt_map_not_finished t T C (ZMap.get id TAC)).
      {
        simpl.
        auto.
      }
      {
        simpl.
        right.
        apply IHqueue; auto.
      }
    }
  Qed.

  Lemma highest_not_finished_map_condition_met_aux:
    forall N t TAC conf queue T C i,
      NoDup (extract_id_queue queue) ->
      i = highest_not_finished_map N t TAC conf queue ->
      0 <= i < N ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      vt_map_not_finished t T C (ZMap.get i TAC) = true.
  Proof.
    intros * Hdup Hi Hbound Hconf.
    induction queue.
    {
      simpl in Hi.
      omega.
    }
    {
      simpl in Hi.
      destruct a as [d id].
      destruct (zeq i id) as [Heq | Hneq].
      {
        rewrite <- Heq in *.
        rewrite Hconf in Hi.
        destruct (vt_map_not_finished t T C (ZMap.get i TAC)); auto.
        simpl in Hdup.
        inv Hdup.
        contradiction H1.
        rewrite Heq.
        eapply highest_not_finished_map_res_in_queue; eauto.
        omega.
      }
      {
        destruct (ZMap.get id conf).
        destruct (vt_map_not_finished t z z0 (ZMap.get id TAC)); try omega.
        inv Hdup.
        apply IHqueue; auto.
      }
    }
  Qed.

  Lemma highest_not_finished_map_condition_met:
    forall Nnat t TAC conf T C i,
      i = highest_not_finished_map (Z.of_nat Nnat) t TAC conf (real_priqueue Nnat t conf) ->
      0 <= i < Z.of_nat Nnat ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      vt_map_not_finished t T C (ZMap.get i TAC) = true.      
  Proof.
    intros.
    eapply highest_not_finished_map_condition_met_aux; eauto.
    apply real_priqueue_ids_nodup.
  Qed.

  Lemma exists_ready_task_highest_not_finished_map_schedule_something_aux:
    forall t id T C N conf TAC queue,
      0 <= id < N ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      In id (extract_id_queue queue) ->
      priqueue_sound N t conf queue ->      
      vt_map_not_finished t T C (ZMap.get id TAC) = true ->
      0 <= highest_not_finished_map N t TAC conf queue < N.
  Proof.
    intros * Hid Hconf Hin Hsound Hready.
    induction queue.
    {
      simpl in Hin.
      omega.
    }
    {
      destruct a as [d i].
      simpl in *.
      destruct (zeq i id) as [Heq | Hneq].
      {
        rewrite Heq in *.
        rewrite Hconf.
        rewrite Hready.
        auto.
      }
      {
        destruct Hin; try omega.
        destruct (ZMap.get i conf) as [Ti Ci] eqn:Hconfi.
        destruct (vt_map_not_finished t Ti Ci (ZMap.get i TAC)).
        {
          destruct Hsound as [_ Hvalid].
          assert (H1: In (PriEntry d i) (PriEntry d i :: queue)).
          { simpl. auto. }
          specialize (Hvalid _ _ _ _ H1 Hconfi).
          omega.
        }
        {
          apply IHqueue; auto.
          - apply priqueue_sound_shrink with (a := PriEntry d i); auto.
        }
      }
    }
  Qed.

  Lemma exists_ready_task_highest_not_finished_map_schedule_something:
    forall t id T C N conf,
      0 <= id < N ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      vt_map_not_finished t T C (ZMap.get id (TAC_pool_at_t t (Z.to_nat N) conf)) = true ->
      0 <= highest_not_finished_map N t
                                   (TAC_pool_at_t t (Z.to_nat N) conf)
                                   conf (real_priqueue (Z.to_nat N) t conf) < N.
  Proof.
    intros * Hid Hconf Hfinish.
    eapply exists_ready_task_highest_not_finished_map_schedule_something_aux; eauto.
    - eapply real_priqueue_extract_id_complete; eauto.
      rewrite Z2Nat.id; omega.
    - remember (Z.to_nat N) as Nnat.
      replace N with (Z.of_nat Nnat).
      + apply real_priqueue_sound.
      + rewrite HeqNnat.
        rewrite Z2Nat.id; omega.
  Qed.


  Section SCHEDULED_TASK.
    
    Lemma highest_not_finished_map_has_highest_priority_among_ready_ones_aux:
      forall t id T C N conf TAC queue,
        0 <= id < N ->
        ZMap.get id conf = mkPrioConfigValid T C ->
        priqueue_sound N t conf queue ->
        highest_not_finished_map N t TAC conf queue = id ->
        forall i te Ti Ci,
          i <> id ->
          In (PriEntry te i) queue ->
          ZMap.get i conf = mkPrioConfigValid Ti Ci ->
          ((vt_map_not_finished t Ti Ci (ZMap.get i TAC) = false)
           \/
           (priEntry_lt (PriEntry (t/T*T+T) id) (PriEntry te i)) = true).
    Proof.
      intros * Hid Hconfid.
      induction queue.
      {
        simpl.
        intros.
        omega.
      }
      {
        intros Hsound Hval * Hi Hin Hconfi.
        destruct a as [ate ai].
        simpl in Hval.
        destruct (ZMap.get ai conf) as [Tai Cai] eqn:Hconfai.
        destruct (vt_map_not_finished t Tai Cai (ZMap.get ai TAC)) eqn:Hfinish.
        {
          inv Hval.
          rewrite Hconfid in Hconfai.
          inv Hconfai.
          right.

          assert (Hin': In (PriEntry te i) queue).
          {            
            simpl in Hin.
            destruct Hin.
            - inv H. omega.
            - auto.
          }

          destruct Hsound as [Hsort Hvalue].
          assert (Hate: ate = t / Tai * Tai + Tai).
          {
            assert (H1: In (PriEntry ate id) (PriEntry ate id :: queue)).
            { simpl. left. auto. }
            specialize (Hvalue ate id Tai Cai H1 Hconfid).
            omega.
          }
          rewrite Hate in *. clear Hate.

          assert (H2: is_true (priEntry_le (PriEntry (t / Tai * Tai + Tai) id) (PriEntry te i))).
          {
            inv Hsort.
            rewrite Forall_forall in H2.
            apply H2; auto.
          }

          unfold is_true in H2.
          rewrite <- priEntry_id_diff_lt; auto.
        }
        {
          destruct (zeq i ai) as [Heq|Hneq].
          {
            rewrite Heq in *.
            rewrite Hconfai in Hconfi.
            inv Hconfi.
            left.
            auto.
          }
          {
            apply IHqueue; auto.
            - eapply priqueue_sound_shrink; eauto.
            - simpl in Hin.
              destruct Hin; try congruence.
          }
        }
      }
    Qed.

    Lemma highest_not_finished_map_has_highest_priority_among_ready_ones:
      forall t id T C N conf TAC,
        0 <= id < N ->
        ZMap.get id conf = mkPrioConfigValid T C ->
        highest_not_finished_map N t TAC conf (real_priqueue (Z.to_nat N) t conf) = id ->
        forall i Ti Ci,
          0 <= i < N ->
          i <> id ->
          ZMap.get i conf = mkPrioConfigValid Ti Ci ->
          ((vt_map_not_finished t Ti Ci (ZMap.get i TAC) = false)
           \/
           (priEntry_lt (PriEntry (t/T*T+T) id) (PriEntry (t/Ti*Ti+Ti) i)) = true).
    Proof.
      intros * Hid Hconfid Hval * Hi Hi' Hconfi.
      assert (H1: priqueue_sound N t conf (real_priqueue (Z.to_nat N) t conf)).
      {
        remember (Z.to_nat N) as Nnat.
        assert (HNN: N = Z.of_nat Nnat).
        {
          rewrite HeqNnat.
          rewrite Z2Nat.id; omega.
        }
        rewrite HNN.
        apply real_priqueue_sound.
      }
      
      eapply highest_not_finished_map_has_highest_priority_among_ready_ones_aux; eauto.
      - eapply real_priqueue_complete; eauto.
        rewrite Z2Nat.id; omega.
    Qed.

    Lemma highest_not_finished_map_idle_all_tasks_finished_aux:
      forall t id T C N conf TAC queue,
        0 <= id < N ->
        ZMap.get id conf = mkPrioConfigValid T C ->
        priqueue_sound N t conf queue ->
        highest_not_finished_map N t TAC conf queue = N ->
        In (PriEntry (t / T * T + T) id) queue ->
        vt_map_not_finished t T C (ZMap.get id TAC) = false.
    Proof.
      intros * Hid Hconfid Hsound Hres Hin.
      induction queue.
      {
        simpl in Hin.
        omega.
      }
      {
        destruct a as [d i].
        destruct (zeq i id) as [Heq | Hneq].
        {
          rewrite Heq in *. clear Heq.
          simpl in Hres.
          rewrite Hconfid in Hres.
          destruct (vt_map_not_finished t T C (ZMap.get id TAC)); auto; omega.
        }
        {
          simpl in Hin.
          destruct Hin; try congruence.
          simpl in Hres.
          destruct (ZMap.get i conf) eqn:Hconfi.
          assert (Hi: 0 <= i < N).
          {
            destruct Hsound.
            assert (Hini: In (PriEntry d i) (PriEntry d i :: queue)).
            { simpl. auto. }
            specialize (H1 _ _ _ _ Hini Hconfi).
            omega.
          }
          destruct (vt_map_not_finished t z z0 (ZMap.get i TAC)); try omega.

          assert (Hsound': priqueue_sound N t conf queue).
          {
            eapply priqueue_sound_shrink; eauto.
          }
          apply IHqueue; auto.
        }
      }
    Qed.

    Lemma highest_not_finished_map_idle_all_tasks_finished:
      forall t id T C N conf TAC,
        0 <= id < N ->
        ZMap.get id conf = mkPrioConfigValid T C ->
        highest_not_finished_map N t TAC conf (real_priqueue (Z.to_nat N) t conf) = N ->
        vt_map_not_finished t T C (ZMap.get id TAC) = false.
    Proof.
      intros * Hid Hconfid Hidle.

      assert (HNN: N = Z.of_nat (Z.to_nat N)).
      {
        rewrite Z2Nat.id; omega.
      }
      
      eapply highest_not_finished_map_idle_all_tasks_finished_aux; eauto.
      - rewrite HNN at 1.
        apply real_priqueue_sound.
      - eapply real_priqueue_complete; eauto.
        rewrite Z2Nat.id; omega.
    Qed.
      
  End SCHEDULED_TASK.

  
  Lemma vt_increase_for_the_scheduled_task_aux:
    forall nt id Nnat TAC conf queue,
      NoDup (extract_id_queue queue) ->
      id = highest_not_finished_map (Z.of_nat Nnat) (Z.of_nat nt) TAC conf queue ->
      0 <= id < Z.of_nat Nnat ->
      (ZMap.get id
               (time_map_tick (Z.of_nat nt) TAC conf queue)) (Z.pos (Pos.of_succ_nat nt)) =
      ZMap.get id TAC (Z.of_nat nt) + 1.
  Proof.
    intros *.
    induction queue.
    {
      simpl.
      omega.
    }
    {
      intros Hdup Hidval Hid.
      simpl.
      simpl in Hidval, Hdup.
      destruct a as [d i].
      destruct (ZMap.get i conf) as [T C] eqn:Hconfi.
      destruct (zeq i id) as [Heq | Hneq].
      {
        rewrite <- Heq in *. clear Heq.
        rewrite Zpos_P_of_succ_nat.        
        destruct (vt_map_not_finished (Z.of_nat nt) T C (ZMap.get i TAC)); rewrite ZMap.gss.
        - rewrite zle_true; try omega.
        - rewrite zle_true; try omega.
      }
      {
        destruct (vt_map_not_finished (Z.of_nat nt) T C (ZMap.get i TAC)) eqn:Hfinish; rewrite ZMap.gso; auto.
        - omega. 
        - apply IHqueue; auto.
          inv Hdup; auto.
      }
    }
  Qed.
    
  Lemma vt_increase_for_the_scheduled_task:
    forall id t Nnat conf,
      0 <= t ->
      id = highest_not_finished_map (Z.of_nat Nnat) t 
                                    (TAC_pool_at_t t Nnat conf) conf (real_priqueue Nnat t conf) ->
      0 <= id < Z.of_nat Nnat ->
      (ZMap.get id (TAC_pool_at_t (t + 1) Nnat conf)) (t + 1) =
      (ZMap.get id (TAC_pool_at_t t Nnat conf)) t + 1.
  Proof.
    intros * Htbound Hid Hbound.
    unfold TAC_pool_at_t.
    remember (Z.to_nat t) as nt.
    assert (Ht: t = Z.of_nat nt).
    {
      rewrite Heqnt.
      rewrite Z2Nat.id; omega.
    }
    assert (HSt: t + 1 = Z.of_nat (S nt)).
    {
      rewrite Nat2Z.inj_succ.
      omega.
    }
    rewrite HSt.
    rewrite Nat2Z.id.
    simpl.
    rewrite Ht in *.
    eapply vt_increase_for_the_scheduled_task_aux; eauto.
    - apply real_priqueue_ids_nodup.
    - unfold TAC_pool_at_t in Hid.
      rewrite <- Heqnt in Hid.
      auto.
  Qed.

  Lemma vt_stagnates_if_task_not_scheduled_aux:
    forall nt id T C Nnat TAC conf queue,
      TAC_constant_starting_t (Z.of_nat nt) TAC ->
      In id (extract_id_queue queue) ->
      0 <= highest_not_finished_map (Z.of_nat Nnat) 
             (Z.of_nat nt) TAC conf queue < Z.of_nat Nnat ->
      id <> highest_not_finished_map (Z.of_nat Nnat) (Z.of_nat nt) TAC conf queue ->
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      vt_map_not_finished (Z.of_nat nt) T C
                   (ZMap.get id TAC) = true ->
      ZMap.get id (time_map_tick (Z.of_nat nt) TAC conf queue)
               (Z.pos (Pos.of_succ_nat nt)) =
      ZMap.get id TAC (Z.of_nat nt).
  Proof.
    intros *.
    induction queue.
    {
      simpl.
      intros.
      omega.
    }
    {
      intros Hconst Hin Hsched Hidval Hid Hconf Hvtmap.
      simpl.
      simpl in Hidval, Hsched.
      destruct a as [d i].

      destruct (zeq i id) as [Heq | Hneq].
      {
        rewrite <- Heq in *. clear Heq.
        rewrite Hconf in *.
        rewrite Hvtmap in Hidval.
        congruence.
      }
      {
        destruct (ZMap.get i conf) as [Ti Ci] eqn:Hconfi.
        destruct (vt_map_not_finished (Z.of_nat nt) Ti Ci (ZMap.get i TAC)) eqn:Hisched; rewrite ZMap.gso; auto.
        {
          rewrite Hconst; auto.
          xomega.
        }
        {
          apply IHqueue; auto; try omega.
          simpl in Hin.
          destruct Hin; auto; try omega.
        }
      }
    }
  Qed.
  
  Lemma vt_stagnates_if_task_not_scheduled:
    forall id T C t Nnat conf,
      0 <= t ->
      0 <= highest_not_finished_map (Z.of_nat Nnat) t 
                                    (TAC_pool_at_t t Nnat conf) conf (real_priqueue Nnat t conf) < Z.of_nat Nnat ->      
      id <> highest_not_finished_map (Z.of_nat Nnat) t 
                                    (TAC_pool_at_t t Nnat conf) conf (real_priqueue Nnat t conf) ->
      0 <= id < Z.of_nat Nnat ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      vt_map_not_finished t T C (ZMap.get id (TAC_pool_at_t t Nnat conf)) = true ->                  
      (ZMap.get id (TAC_pool_at_t (t + 1) Nnat conf)) (t + 1) =
      (ZMap.get id (TAC_pool_at_t t Nnat conf)) t.
  Proof.
    intros * Htbound Hsched Hid Hidbound Hconf Hidnotfinish.
    unfold TAC_pool_at_t.
    remember (Z.to_nat t) as nt.
    assert (Ht: t = Z.of_nat nt).
    {
      rewrite Heqnt.
      rewrite Z2Nat.id; omega.
    }
    assert (HSt: t + 1 = Z.of_nat (S nt)).
    {
      rewrite Nat2Z.inj_succ.
      omega.
    }
    rewrite HSt.
    rewrite Nat2Z.id.
    simpl.
    rewrite Ht in *.
    assert (H1: TAC_pool_at_t (Z.of_nat nt) Nnat conf = TAC_pool_at_t_aux nt Nnat conf).
    {
      unfold TAC_pool_at_t.
      rewrite <-Heqnt.
      auto.
    }
    rewrite H1 in *.
    apply vt_stagnates_if_task_not_scheduled_aux with (T:=T) (C:=C) (Nnat:=Nnat); auto.
    - rewrite <- H1.
      apply TAC_pool_at_t_constant_starting_t.
      xomega.
    - apply real_priqueue_extract_id_complete; auto.
  Qed.
      
End UPDATEMAP.


Section VT_AND_PRIORITY.

  Definition id_not_blocked id t N conf TAC : Prop :=
    forall i T C Tid Cid,
      0 <= i < N ->
      i <> id ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      ZMap.get id conf = mkPrioConfigValid Tid Cid ->
      ((vt_map_not_finished t T C (ZMap.get i TAC) = false)
       \/
       (priEntry_lt (PriEntry (t/Tid*Tid+Tid) id) (PriEntry (t/T*T+T) i)) = true).

  Definition id_not_blocked_queue id did t conf TAC queue : Prop :=
    forall i T C d,
      i <> id ->
      ZMap.get i conf = mkPrioConfigValid T C ->
      In (PriEntry d i) queue ->
      ((vt_map_not_finished t T C (ZMap.get i TAC) = false)
       \/
       (priEntry_lt (PriEntry did id) (PriEntry d i)) = true).

  Definition id_is_blocked id t N conf TAC : Prop :=
    exists i, 0 <= i < N /\ i <> id /\
    (match (ZMap.get i conf), (ZMap.get id conf) with
       | mkPrioConfigValid T C, mkPrioConfigValid Tid Cid =>
      ((vt_map_not_finished t T C (ZMap.get i TAC) = true)
       /\
       (priEntry_lt (PriEntry (t/T*T+T) i) (PriEntry (t/Tid*Tid+ Tid) id)) = true)
     end).

  Definition id_is_blocked_queue id did t conf TAC queue : Prop :=
    exists i, i <> id /\
    (match (ZMap.get i conf) with
       | mkPrioConfigValid T C =>
        (In (PriEntry (t/T*T+T) i) queue /\
      (vt_map_not_finished t T C (ZMap.get i TAC) = true)
       /\
       (priEntry_lt (PriEntry (t/T*T+T) i) (PriEntry did id)) = true)
    end).

  Lemma id_is_blocked_canonical:
    forall i t1 t2 N conf,
      0 <= t1 <= t2 ->
      config_pool_valid N conf ->
      (id_is_blocked i t1 N conf (TAC_pool_at_t t2 (Z.to_nat N) conf) <->
      id_is_blocked i t1 N conf (TAC_pool_at_t t1 (Z.to_nat N) conf)).
  Proof.
    intros * Ht Hconf.    
    split; intros Hblock.
    {
      unfold id_is_blocked in *.
      destruct Hblock as [j [? [? Hval]]].
      exists j.
      split; auto.
      split; auto.
      destruct (ZMap.get j conf) eqn:Hconfj.
      destruct (ZMap.get i conf) eqn:Hconfi.
      assert (HT: 1 < z).
      {
        specialize (Hconf _ _ _ H Hconfj).
        omega.
      }
      rewrite vt_map_not_finished_canonical in Hval; try omega.
      auto.
    }
    {
      unfold id_is_blocked in *.
      destruct Hblock as [j [? [? Hval]]].
      exists j.
      split; auto.
      split; auto.
      destruct (ZMap.get j conf) eqn:Hconfj.
      destruct (ZMap.get i conf) eqn:Hconfi.
      assert (HT: 1 < z).
      {
        specialize (Hconf _ _ _ H Hconfj).
        omega.
      }
      rewrite vt_map_not_finished_canonical; try omega.
      auto.
    }
  Qed.

  Lemma id_not_blocked_canonical:
    forall i t1 t2 N conf,
      0 <= t1 <= t2 ->
      config_pool_valid N conf ->
      (id_not_blocked i t1 N conf (TAC_pool_at_t t2 (Z.to_nat N) conf) <->
      id_not_blocked i t1 N conf (TAC_pool_at_t t1 (Z.to_nat N) conf)).
  Proof.
    intros * Ht Hconf.
    split; intros Hblock.
    {
      unfold id_not_blocked in *.
      intros * Hj Hj' Hconfj Hconfi.
      specialize (Hblock _ _ _ _ _ Hj Hj' Hconfj Hconfi).
      assert (HT: 1 < T).
      {
        specialize (Hconf _ _ _ Hj Hconfj).
        omega.
      }
      rewrite vt_map_not_finished_canonical in Hblock; try omega.
      auto.
    }
    {
      unfold id_not_blocked in *.
      intros * Hj Hj' Hconfj Hconfi.
      specialize (Hblock _ _ _ _ _ Hj Hj' Hconfj Hconfi).
      assert (HT: 1 < T).
      {
        specialize (Hconf _ _ _ Hj Hconfj).
        omega.
      }
      rewrite vt_map_not_finished_canonical; try omega.
      auto.
    }
  Qed.
  
  Lemma id_blocked_dec:
    forall id t N conf,
      0 < N ->
      {id_is_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf)} + 
      {id_not_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf)}.
  Proof.
    intros * HN.
    remember (highest_not_finished_map N t (TAC_pool_at_t t (Z.to_nat N) conf) conf
             (real_priqueue (Z.to_nat N) t conf)) as k.
    destruct (zeq k N) as [Hidle|Hreal].
    {
      right.
      unfold id_not_blocked.
      intros * Hi Hid Hconfi Hconfid.
      rewrite Heqk in Hidle.

      left.
      apply highest_not_finished_map_idle_all_tasks_finished with (N:=N) (conf:=conf); auto.
    }
    {
      assert (HNN: N = Z.of_nat (Z.to_nat N)).
      {
        rewrite Z2Nat.id; omega.
      }
      assert (Hk: 0 <= k <= N).
      {
        rewrite Heqk.
        apply highest_not_finished_map_res_bound; auto.
        - rewrite HNN at 1.
          apply real_priqueue_sound.
      }

      destruct (ZMap.get id conf) as [T C] eqn:Hconfid.
      destruct (ZMap.get k conf) as [Tk Ck] eqn:Hconfk.
      destruct (priEntry_lt (PriEntry (t / Tk * Tk + Tk) k) (PriEntry (t / T * T + T) id)) eqn:Hpri.
      {
        left.
        unfold id_is_blocked.
        exists k.
        split; try omega.
        split.
        {
          unfold not.
          intros H.
          rewrite H in *. clear H.
          rewrite Hconfk in Hconfid.
          inversion Hconfid.
          rewrite H0 in Hpri.
          unfold priEntry_lt in Hpri.
          rewrite zlt_false in Hpri by omega.
          rewrite zeq_true in Hpri.
          rewrite zlt_false in Hpri by omega.
          congruence.
        }
        {
          rewrite Hconfk.
          rewrite Hconfid.
          split; auto.
          apply highest_not_finished_map_condition_met with (Nnat := Z.to_nat N) (conf := conf); auto.
          - rewrite <- HNN. auto.
          - omega.
        }
      }
      {
        right.
        unfold id_not_blocked.
        intros * Hi Hi' Hconfi Hconfid'.
        rewrite Hconfid' in Hconfid.
        inversion Hconfid.
        rewrite H0 in *.

        assert (Hk2: 0 <= k < N). omega.
        symmetry in Heqk.
        pose proof (highest_not_finished_map_has_highest_priority_among_ready_ones _ _ _ _ _ _ _ Hk2 Hconfk Heqk).
        destruct (zeq i k) as [Heq | Hneq].
        {
          rewrite Heq in *.
          rewrite Hconfk in Hconfi.
          inversion Hconfi.
          rewrite H3 in *.
          right.
          
          apply priEntry_lt_neg_ft; auto.
        }
        {
          specialize (H _ _ _ Hi Hneq Hconfi).
          destruct H as [Hfinish' | Hpri'].
          {
            left.
            auto.
          }
          {
            right.
            destruct (zeq k id) as [Hkk1 | Hkk2].
            {
              rewrite Hkk1 in *.
              rewrite Hconfk in Hconfid'.
              inversion Hconfid'.
              rewrite H2 in *.
              auto.
            }
            {
              rewrite priEntry_lt_transitive; auto.
              - instantiate (1 := (PriEntry (t / Tk * Tk + Tk) k)).
                unfold is_true.              
                apply priEntry_lt_neg_ft; auto.
              - auto.
            }
          }
        }
      }
    }
  Qed.

  Lemma id_blocked_absurd:
    forall id t N conf,
      0 < N ->
      id_is_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf) -> 
      id_not_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf) ->
      False.
  Proof.
    intros * HN Hb1 Hb2.
    destruct Hb1 as [x [H1 [H2 H3]]].
    unfold id_not_blocked in Hb2.
    destruct (ZMap.get x conf) eqn:Hconfx.
    destruct (ZMap.get id conf) eqn:Hconfid.
    assert (H4: mkPrioConfigValid z1 z2 = mkPrioConfigValid z1 z2). auto.
    specialize (Hb2 _ _ _ _ _ H1 H2 Hconfx H4).
    destruct H3.
    rewrite H in Hb2.
    rewrite priEntry_lt_neg_tf in Hb2; auto.
    destruct Hb2; try congruence.
  Qed.

  Lemma id_is_blocked_queue_self_false:
    forall N t conf dj j queue TAC,
      priqueue_sound N t conf (PriEntry dj j :: queue) ->
      id_is_blocked_queue j dj t conf TAC (PriEntry dj j :: queue) ->
      False.
  Proof.
    intros * Hsound Hblock.    
    unfold id_is_blocked_queue in Hblock.
    destruct Hblock as [i [Hi1 Hi2]].
    destruct (ZMap.get i conf) as [T C] eqn:Hconfi.
    destruct Hi2 as [Hiin [Hifinish Hipri]].
    pose proof (priEntry_lt_neg_tf _ _ _ _ Hipri) as H1.
    clear Hipri.

    simpl in Hiin.
    destruct Hiin; try congruence.

    destruct Hsound as [Hsort _].
    inv Hsort.
    rewrite Forall_forall in H4.

    assert (H6: priEntry_lt (PriEntry dj j) (PriEntry (t / T * T + T) i) = true).
    {
      rewrite <- priEntry_id_diff_lt; auto.
      apply H4.
      auto.
    }
    congruence.
  Qed.

  Lemma id_is_blocked_queue_shrink:
    forall t conf id did dj j Tj Cj queue TAC,
      ZMap.get j conf = mkPrioConfigValid Tj Cj ->
      vt_map_not_finished t Tj Cj (ZMap.get j TAC) = false ->
      id_is_blocked_queue id did t conf TAC (PriEntry dj j :: queue) ->
      id_is_blocked_queue id did t conf TAC queue.
  Proof.
    intros * Hconfj Hfinish Hblock.
    unfold id_is_blocked_queue in *.
    destruct Hblock as [i [Hi Hi']].

    exists i.
    split; auto.
    destruct (zeq i j) as [Heq | Hneq].
    {
      rewrite Heq in *.
      rewrite Hconfj in *.
      destruct Hi' as [_ [H _ ]].
      congruence.
    }
    {
      destruct (ZMap.get i conf).
      destruct Hi' as [? [? ? ]].
      split; auto.
      simpl in H.
      destruct H; try congruence.
    }
  Qed.
  
  Lemma id_not_blocked_queue_shrink:
    forall id did t conf TAC a q,
      id_not_blocked_queue id did t conf TAC (a :: q) ->
      id_not_blocked_queue id did t conf TAC q.
  Proof.
    intros * Hblock.
    unfold id_not_blocked_queue in *.
    intros * Hi Hconf Hin.
    apply Hblock; auto.
    simpl.
    auto.
  Qed.

  Lemma id_not_blocked_implies_queue:
    forall id t Nnat conf T C,
      id_not_blocked id t (Z.of_nat Nnat) conf (TAC_pool_at_t t Nnat conf) ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      id_not_blocked_queue id (t / T * T + T) t conf (TAC_pool_at_t t Nnat conf)
                           (real_priqueue Nnat t conf).
  Proof.
    intros * Hblock Hconfid.
    unfold id_not_blocked_queue.
    unfold id_not_blocked in Hblock.
    intros * Hi Hconfi Hin.

    assert (Hsound: priqueue_sound (Z.of_nat Nnat) t conf (real_priqueue Nnat t conf)).
    {
      apply real_priqueue_sound.
    }
    destruct Hsound as [_ Hval].
    specialize (Hval _ _ _ _ Hin Hconfi).
    destruct Hval as [Hii Hdd].
    rewrite Hdd.
    apply Hblock with (Cid := C); auto.
  Qed.

  Lemma id_is_blocked_implies_queue:
    forall id t Nnat conf T C,
      id_is_blocked id t (Z.of_nat Nnat) conf (TAC_pool_at_t t Nnat conf) ->
      ZMap.get id conf = mkPrioConfigValid T C ->
      id_is_blocked_queue id (t / T * T + T) t conf (TAC_pool_at_t t Nnat conf)
                           (real_priqueue Nnat t conf).
  Proof.
    intros * Hblock Hconfid.
    unfold id_is_blocked_queue.
    unfold id_is_blocked in Hblock.
    destruct Hblock as [i [Hi [Hi' H]]].
    
    exists i.
    split; auto.
    destruct (ZMap.get i conf) as [Ti Ci] eqn:Hconfi.
    rewrite Hconfid in H.
    destruct H.

    split; auto.
    - apply real_priqueue_complete with (C:=Ci); auto.
  Qed.
  
  Lemma vt_increases_id_not_blocked_queue:
    forall id did t N conf TAC queue,
      0 <= t ->
      0 <= id < N ->
      In (PriEntry did id) queue ->
      priqueue_sound N t conf queue ->
      TAC_constant_starting_t t TAC ->
      ZMap.get id (time_map_tick t TAC conf queue) (t + 1) =
      ZMap.get id TAC t + 1 ->
      id_not_blocked_queue id did t conf TAC queue.
  Proof.
    intros * Ht Hid Hin Hsound HTAC Hvt.
    induction queue.
    {
      simpl in Hin.
      omega.
    }
    {
      destruct a as [dj j].
      destruct (zeq j id) as [Heq | Hneq].
      {
        rewrite Heq in *.
        unfold id_not_blocked_queue.
        intros * Hi Hconfi Hini.
        right.
        
        assert (Hd: dj = did).
        {
          eapply priqueue_sound_equal_id_equal_entry; eauto.
          simpl.
          auto.
        }
        rewrite Hd in *. clear Hd.

        simpl in Hini.
        destruct Hini; try congruence.
        
        rewrite <- priEntry_id_diff_lt; auto.
        destruct Hsound as [Hsort _].
        inv Hsort.
        rewrite Forall_forall in H3.
        apply H3; auto.
      }
      {
        unfold id_not_blocked_queue.
        intros * Hi Hconfi Hini.
        
        destruct (zeq i j) as [Hj | Hj].
        {
          rewrite Hj in *. clear Hj.
          left.
          simpl in Hvt.
          rewrite Hconfi in Hvt.

          unfold TAC_constant_starting_t in HTAC.
          assert (Htt: t <= t + 1). omega.
          specialize (HTAC id _ Htt).
          rewrite <- HTAC in Hvt.
          
          destruct (vt_map_not_finished t T C (ZMap.get j TAC)); auto.
          exfalso.          
          rewrite ZMap.gso in Hvt; auto.
          assert (forall a, a = a + 1 -> False).
          {
            intros.
            omega.
          }
          apply H with (a := ZMap.get id TAC (t + 1)).
          auto.
        }
        {
          assert (Hin': In (PriEntry did id) queue).
          {
            simpl in Hin.
            destruct Hin; try congruence.
          }

          assert (Hsound': priqueue_sound N t conf queue).
          {
            eapply priqueue_sound_shrink; eauto.
          }

          assert (Hvt': ZMap.get id (time_map_tick t TAC conf queue) (t + 1) = ZMap.get id TAC t + 1).
          {
            simpl in Hvt.
            destruct (ZMap.get j conf) eqn:Hconfj.
            assert (Hfinish: vt_map_not_finished t z z0 (ZMap.get j TAC) = false).
            {
              unfold TAC_constant_starting_t in HTAC.
              assert (Htt: t <= t + 1). omega.
              specialize (HTAC id _ Htt).
              rewrite <- HTAC in Hvt.
              
              destruct (vt_map_not_finished t z z0 (ZMap.get j TAC)); auto.
              rewrite ZMap.gso in Hvt; auto.
              assert (forall a, a = a + 1 -> False).
              {
                intros.
                omega.
              }
              exfalso.
              apply H with (a := ZMap.get id TAC (t + 1)).
              auto.
            }
            rewrite Hfinish in Hvt.
            rewrite ZMap.gso in Hvt; auto.
          }

          specialize (IHqueue Hin' Hsound' Hvt').
          apply IHqueue; auto.
          - simpl in Hini.
            destruct Hini; try congruence.
        }
      }
    }
  Qed.

  Lemma vt_increases_id_not_blocked:
    forall id t N conf TAC,
      0 <= t ->
      0 <= id < N ->
      TAC_constant_starting_t t TAC ->
      ZMap.get id (time_map_tick t TAC conf (real_priqueue (Z.to_nat N) t conf)) (t + 1) =
      ZMap.get id TAC t + 1 ->
      id_not_blocked id t N conf TAC.
  Proof.
    intros.
    unfold id_not_blocked.
    intros.
    pose proof vt_increases_id_not_blocked_queue.
    unfold id_not_blocked_queue in H7.

    remember (Z.to_nat N) as Nnat.
    assert (HNN: N = Z.of_nat Nnat).
    {
      rewrite HeqNnat.
      rewrite Z2Nat.id; omega.
    }

    assert (Hsound: priqueue_sound N t conf (real_priqueue Nnat t conf)).
    {
      rewrite HNN.
      apply real_priqueue_sound.
    }
    
    eapply H7; eauto.
    - eapply real_priqueue_complete; eauto.
      omega.
    - eapply real_priqueue_complete; eauto.
      omega.
  Qed.

  Lemma TAC_increases_id_not_blocked:
    forall id t N conf,
      0 <= t ->
      0 <= id < N ->
      ZMap.get id (TAC_pool_at_t (t + 1) (Z.to_nat N) conf) (t + 1) =
      ZMap.get id (TAC_pool_at_t t (Z.to_nat N) conf) t + 1 ->
      id_not_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf).
  Proof.
    intros * Ht Hid Hvt.
    unfold TAC_pool_at_t in Hvt.
    remember (Z.to_nat t) as nt.
    assert (Htt: t + 1 = Z.of_nat (S nt)).
    {
      rewrite Heqnt.
      rewrite Nat2Z.inj_succ.
      rewrite Z2Nat.id; omega.
    }
    rewrite Htt in Hvt.
    rewrite Nat2Z.id in Hvt.
    assert (H1: Z.of_nat (S nt) = t + 1). omega.
    rewrite H1 in Hvt.
    simpl in Hvt.
    apply vt_increases_id_not_blocked; auto.
    - apply TAC_pool_at_t_constant_starting_t; auto.
    - unfold TAC_pool_at_t.
      rewrite <- Heqnt.
      assert (Htt': Z.of_nat nt = t).
      {
        rewrite Heqnt.
        rewrite Z2Nat.id; omega.
      }
      rewrite Htt' in Hvt.
      auto.
  Qed.    
  
  Lemma vt_stagnates_id_is_blocked_queue:
    forall id did t N conf TAC queue,
      0 <= t ->
      0 <= id < N ->
      In (PriEntry did id) queue ->
      priqueue_sound N t conf queue ->
      TAC_constant_starting_t t TAC ->
      ZMap.get id (time_map_tick t TAC conf queue) (t + 1) = ZMap.get id TAC t ->
      id_is_blocked_queue id did t conf TAC queue.
  Proof.
    intros * Ht Hid Hin Hsound HTAC Hvt.
    induction queue.
    {
      simpl in Hin.
      omega.
    }
    {
      unfold TAC_constant_starting_t in HTAC.
      assert (Htt: t <= t + 1). omega.
      specialize (HTAC id _ Htt).
      
      destruct a as [dj j].
      destruct (zeq j id) as [Heq | Hneq].
      {
        rewrite Heq in *.
        
        simpl in Hvt.
        destruct (ZMap.get id conf) as [T C] eqn:Hconfid.
        destruct (vt_map_not_finished t T C (ZMap.get id TAC)).
        - rewrite ZMap.gss in Hvt.
          rewrite zle_true in Hvt; try omega.
        - rewrite ZMap.gss in Hvt.
          rewrite zle_true in Hvt; try omega.
      }
      {        
        unfold id_is_blocked_queue.
        simpl in Hvt.

        destruct (ZMap.get j conf) as [Tj Cj] eqn:Hconfj.
        destruct (vt_map_not_finished t Tj Cj (ZMap.get j TAC)) eqn:Hfinish.
        {
          rewrite ZMap.gso in Hvt; auto.
          exists j.
          split; auto.
          rewrite Hconfj.

          destruct Hsound as [Hsort Hval].

          assert (Hin': In (PriEntry dj j) (PriEntry dj j :: queue)).
          {
            simpl. auto.
          }
          assert (Hdj: dj = t /Tj*Tj + Tj).
          {            
            specialize (Hval _ _ _ _ Hin' Hconfj).
            omega.
          }
          rewrite Hdj.
          split; auto.
          - simpl. auto.
          - split; auto.
            rewrite <- Hdj.
            inv Hsort.
            rewrite Forall_forall in H2.
            rewrite <- priEntry_id_diff_lt; auto.
            apply H2.
            simpl in Hin.
            destruct Hin; try congruence.
        }
        {
          rewrite ZMap.gso in Hvt; auto.
          assert (H1: In (PriEntry did id) queue).
          {
            simpl in Hin.
            destruct Hin; try congruence.
          }
          assert (H2: priqueue_sound N t conf queue).
          {
            eapply priqueue_sound_shrink; eauto.
          }
          specialize (IHqueue H1 H2 Hvt).
          
          destruct IHqueue as [x [Hx Hx2]].
          exists x.
          split; auto.
          destruct (ZMap.get x conf) as [T C] eqn:Hconf.
          destruct Hx2 as [Hx3 [Hx4 Hx5]].
          split; auto.
          simpl.
          auto.
        }
      }
    }
  Qed.
    
  Lemma vt_stagnates_id_is_blocked:
    forall id t N conf TAC,
      0 <= t ->
      0 <= id < N ->
      TAC_constant_starting_t t TAC ->
      ZMap.get id (time_map_tick t TAC conf (real_priqueue (Z.to_nat N) t conf)) (t + 1) =
      ZMap.get id TAC t ->
      id_is_blocked id t N conf TAC.
  Proof.
    intros.
    pose proof vt_stagnates_id_is_blocked_queue.

    destruct (ZMap.get id conf) as [Tid Cid] eqn:Hconfid.
    remember (Z.to_nat N) as Nnat.
    assert (HNN: N = Z.of_nat Nnat).
    {
      rewrite HeqNnat.
      rewrite Z2Nat.id; omega.
    }
    
    assert (Hblock: id_is_blocked_queue id (t/Tid*Tid+Tid) t conf
             TAC
             (real_priqueue (Z.to_nat N) t conf)).
    {
      apply vt_stagnates_id_is_blocked_queue with (N:=N); auto.
      - eapply real_priqueue_complete; eauto.
        rewrite Z2Nat.id; omega.
      - rewrite HNN.
        rewrite Nat2Z.id.
        apply real_priqueue_sound.
      - rewrite HNN.
        rewrite Nat2Z.id.
        auto.
    }

    unfold id_is_blocked_queue in Hblock.
    unfold id_is_blocked.
    destruct Hblock as [i [Hi Hi']].
    exists i.
    destruct (ZMap.get i conf) as [T C] eqn:Hconfi.
    destruct Hi' as [Hi1 [Hi2 Hi3]].

    rewrite Hconfid.
    split; auto.

    rewrite HNN in Hi1.
    rewrite Nat2Z.id in Hi1.
    assert (Hsound: priqueue_sound N t conf (real_priqueue Nnat t conf)).
    {
      rewrite HNN.
      apply real_priqueue_sound.
    }

    destruct Hsound as [_ Hval].
    specialize (Hval _ _ _ _ Hi1 Hconfi).
    omega.
  Qed.

  Lemma TAC_stagnates_id_is_blocked:
    forall id t N conf,
      0 <= t ->
      0 <= id < N ->
      ZMap.get id (TAC_pool_at_t (t + 1) (Z.to_nat N) conf) (t + 1) =
      ZMap.get id (TAC_pool_at_t t (Z.to_nat N) conf) t ->
      id_is_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf).
  Proof.
    intros * Ht Hid Hvt.
    unfold TAC_pool_at_t in Hvt.
    remember (Z.to_nat t) as nt.
    assert (Htt: t + 1 = Z.of_nat (S nt)).
    {
      rewrite Heqnt.
      rewrite Nat2Z.inj_succ.
      rewrite Z2Nat.id; omega.
    }
    rewrite Htt in Hvt.
    rewrite Nat2Z.id in Hvt.
    assert (H1: Z.of_nat (S nt) = t + 1). omega.
    rewrite H1 in Hvt.
    simpl in Hvt.
    apply vt_stagnates_id_is_blocked; auto.
    - apply TAC_pool_at_t_constant_starting_t; auto.
    - unfold TAC_pool_at_t.
      rewrite <- Heqnt.
      assert (Htt': Z.of_nat nt = t).
      {
        rewrite Heqnt.
        rewrite Z2Nat.id; omega.
      }
      rewrite Htt' in Hvt.
      auto.
  Qed.

  Lemma id_is_blocked_TAC_stagnates:
    forall id t N conf,
      0 <= t ->
      0 <= id < N ->
      id_is_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf) ->      
      ZMap.get id (TAC_pool_at_t (t + 1) (Z.to_nat N) conf) (t + 1) =
      ZMap.get id (TAC_pool_at_t t (Z.to_nat N) conf) t.
  Proof.
    intros * Ht Hid Hblock.
    
    assert (H1: 0 <= t + 1). omega.
    assert (H2: 0 <= id < Z.of_nat (Z.to_nat N)).
    {
      rewrite Z2Nat.id; omega.
    }
    
    pose proof (TAC_pool_all_monotonic_continuous _ _ conf _ H1 H2).
    unfold monotonic_continuous in H.
    assert (H3: t >= 0). omega.
    specialize (H _ H3).

    assert (Hvt: ZMap.get id (TAC_pool_at_t (t + 1) (Z.to_nat N) conf) t = ZMap.get id (TAC_pool_at_t t (Z.to_nat N) conf) t).
    {
      rewrite vertical_map_canonical; omega.
    }
    rewrite Hvt in H.
    destruct H; try omega.
    
    exfalso.
    
    assert (Hblock': id_not_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf)).
    {
      apply TAC_increases_id_not_blocked; auto.
    }
    eapply id_blocked_absurd; eauto.
    omega.
  Qed.

  Lemma id_not_blocked_TAC_increases:
    forall id t N conf,
      0 <= t ->
      0 <= id < N ->
      id_not_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf) ->      
      ZMap.get id (TAC_pool_at_t (t + 1) (Z.to_nat N) conf) (t + 1) =
      ZMap.get id (TAC_pool_at_t t (Z.to_nat N) conf) t + 1.
  Proof.
    intros * Ht Hid Hblock.
    
    assert (H1: 0 <= t + 1). omega.
    assert (H2: 0 <= id < Z.of_nat (Z.to_nat N)).
    {
      rewrite Z2Nat.id; omega.
    }
    
    pose proof (TAC_pool_all_monotonic_continuous _ _ conf _ H1 H2).
    unfold monotonic_continuous in H.
    assert (H3: t >= 0). omega.
    specialize (H _ H3).

    assert (Hvt: ZMap.get id (TAC_pool_at_t (t + 1) (Z.to_nat N) conf) t = ZMap.get id (TAC_pool_at_t t (Z.to_nat N) conf) t).
    {
      rewrite vertical_map_canonical; omega.
    }
    rewrite Hvt in H.
    destruct H; try omega.
    
    exfalso.
    
    assert (Hblock': id_is_blocked id t N conf (TAC_pool_at_t t (Z.to_nat N) conf)).
    {
      apply TAC_stagnates_id_is_blocked; auto.
    }
    eapply id_blocked_absurd; eauto.
    omega.
  Qed.
  
End VT_AND_PRIORITY.



Section DEBUGGING.
  
  Fixpoint output_f_aux (s: Z) (nt: nat) (f: Z -> Z) :=
    match nt with
    | O => nil
    | S n =>
      let x := s + Z.of_nat n in
      (x, f x) :: output_f_aux s n f
    end.

  Function output_f_range (xstart xend: Z) (f: Z -> Z) :=    
    rev (output_f_aux xstart (Z.to_nat (xend - xstart)) f).
  
End DEBUGGING.



Section QUANTUMUTICK.

  Definition quantum_utick_correct Q t conf N:=
    forall i, 0 <= i < N ->
      cal_quantum_valid_at_i t i
      (ZMap.get i Q) conf (TAC_pool_at_t (t+1) (Z.to_nat N) conf).

  (*
  Lemma Quantum_utick_correct_real_quantum:
    forall Q,
      quantum_utick_correct (real_quantum Q) (hyper_period - 1) real_prioconfig_pool num_prio_proc.
  Proof.
    intros.
    unfold quantum_utick_correct, real_quantum.
    intros i Hi.
    rewrite init_zmap_inside; auto.
    unfold cal_quantum_valid_at_i, real_prioconfig_pool, real_time_container_pool.
    destruct (zeq i 0).
    rewrite e.
    rewrite ZMap.gss.
    auto.
    destruct (zeq i 1).
    rewrite e.
    rewrite ZMap.gso; [ | congruence].
    rewrite ZMap.gss.
    auto.
    destruct (zeq i 2).
    rewrite e.
    rewrite ZMap.gso; [ | congruence].
    rewrite ZMap.gso; [ | congruence].      
    rewrite ZMap.gss.
    auto.
    omega.
  Qed.
*)

End QUANTUMUTICK.


